﻿--[[
Name: LibStatLogic-1.2
Description: A Library for stat conversion, calculation and summarization.
Revision: $Revision: 124 $
Author: Whitetooth
Email: hotdogee [at] gmail [dot] com
Last Update: $Date: 2012-09-13 11:58:36 +0000 (Thu, 13 Sep 2012) $
Website:
Documentation:
SVN: $URL $
Dependencies: UTF8
License: LGPL v3
Features:
  StatConversion -
    Ratings -> Effect
    Str -> AP, Block
    Agi -> Crit, Dodge, AP, RAP, Armor
    Sta -> Health, SpellDmg(Talent)
    Int -> Mana, SpellCrit
    Spi -> MP5, HP5
    and more!
  StatMods - Get stat mods from talents and buffs for every class
  BaseStats - for all classes and levels
  ItemStatParser - Fast multi level indexing algorithm instead of calling strfind for every stat

Debug:
  /sldebug

  /slwarning
	Allows libstatlogic to show warnings if an item is not recognized.
	This is a compromise between nothing and full /sldebug mode
]]

local MAJOR = "LibStatLogic-1.2";
local MINOR = "$Revision: 124 $";

local StatLogic = LibStub:NewLibrary(MAJOR, MINOR)
if not StatLogic then return end

-- For Mikk's findglobals script
local _G = _G
local tostring,tonumber,gsub,select,next,pairs,ipairs,type,unpack,strsub,strlen =
       tostring,tonumber,gsub,select,next,pairs,ipairs,type,unpack,strsub,strlen
local ceil,floor = 
       ceil,floor
local CR_DEFENSE_SKILL, CR_DODGE, CR_PARRY = 
       CR_DEFENSE_SKILL, CR_DODGE, CR_PARRY
local GetCombatRating, GetCombatRatingBonus, GetParryChance, GetDodgeChance =
       GetCombatRating, GetCombatRatingBonus, GetParryChance, GetDodgeChance
local UnitBuff, UnitDebuff, GetItemInfo, GetItemGem, GetTime = 
       UnitBuff, UnitDebuff, GetItemInfo, GetItemGem, GetTime
-- GLOBALS: error, print, setmetatable, getmetatable, debugstack
-- GLOBALS: ItemRefTooltip, UIParent, ShowUIPanel
-- GLOBALS: DEFAULT_CHAT_FRAME
       
----------------------
-- Version Checking --
----------------------
StatLogic.Major = MAJOR;
StatLogic.Minor = MINOR;
local wowBuildNo = tonumber((select(2, GetBuildInfo()))); --e.g. "5.0.4", "16016", "Aug 21 2012", 50001
StatLogic.wowBuildNo = wowBuildNo --e.g. 16016
local toc = tonumber((select(4, GetBuildInfo()))); --e.g. 50001
StatLogic.ItemsNotRecognized = {}; --list of all text that we were unable to parse

--[[
	Version		Build
	==========  =====
	4.0.1		13164
	4.0.1a		13205
				13221
	4.0.3		13287
	4.0.3a		13329
	4.0.6		13596  --
	4.0.6a		13623  --The Shattering/Cataclysm
	4.1.0		13914  --Rise of the Zandalari
	4.1.0a		14007  --Rise of the Zandalari
	4.2.0		14333  --Rage of the Firelands
	4.2.0a		14480  --Rage of the Firelands
	4.2.2		14545  --Rage of the Firelands
	4.3.0		15005  --Hour of Twilight
	4.3.0a		15055  --Hour of Twilight
	4.3.2		15211  --Hour of Twilight
	4.3.3		15354  --Hour of Twilight
	4.3.4		15595  --Hour of Twilight	GetBuildInfo() -> "4.3.4", "15595", "Apr 10 2012", 40300
	
	--Mists of Pandaria Beta
	5.0.4		15929  --Mists of Pandaria Beta	GetBuildInfo() -> "5.0.4", "15929", "Jul 31 2012", 50001
	5.0.4		15952  --Mists of Pandaria Beta	GetBuildInfo() -> "5.0.4", "15952", "Aug 5 2012",  50001
	5.0.4		15961  --Mista of Pandaria Beta	GetBuildInfo() -> "5.0.4", "15961", "Aug 7 2012",  50001
	5.0.4		15983  --Mists of Pandaria Beta	GetBuildInfo() -> "5.0.4", "15983", "Aug 14 2012", 50001
	
	--Mists of Pandaria 
	5.0.4		16016  --Mists of Pandaria	GetBuildInfo() -> "5.0.4", "16016", "Aug 21 2012", 50001  Released to live 8/28/2012
	5.0.5		16048  --Mists of Panderia  GetBuildInfo() -> "5.0.4", "16048", "Sep 5 2012", 50001   Released to live 9/11/2012
--]]

-- Our localization information
local L = nil;

-------------------
-- Set Debugging --
-------------------
local DEBUG = false
function ToggleDebug()
	DEBUG = not DEBUG
  	if DEBUG then
		print(ORANGE_FONT_COLOR_CODE.."LibStatLogic"..FONT_COLOR_CODE_CLOSE..": Debugging enabled");
	else
		print(ORANGE_FONT_COLOR_CODE.."LibStatLogic"..FONT_COLOR_CODE_CLOSE..": Debugging disabled");
	end;
end
SlashCmdList["STATLOGIC_DEBUG"] = ToggleDebug;
SLASH_STATLOGIC_DEBUG1 = "/sldebug"


--Show warnings by default for enUS for few weeks after each launch - until we've caught everything.
--Once the warnings have settled down, they can be disabled
local ItemNotRecognizedWarningDefault = (GetLocale() == "enUS") and true;
--Note: We won't show warnings, by default, for any other locales, because their translations are 
--      so broken and far behind that the game will never be usable for anyone besides enUS.

local _itemNotRecognizedWarning = ItemNotRecognizedWarningDefault;
function ToggleWarnings()
	_itemNotRecognizedWarning = not _itemNotRecognizedWarning;
	if _itemNotRecognizedWarning then
		print(ORANGE_FONT_COLOR_CODE.."LibStatLogic"..FONT_COLOR_CODE_CLOSE..": Warnings "..GREEN_FONT_COLOR_CODE.."enabled"..FONT_COLOR_CODE_CLOSE ..". A warning will be displayed if an item's attribute is not recognized");
	else
		print(ORANGE_FONT_COLOR_CODE.."LibStatLogic"..GRAY_FONT_COLOR_CODE..": Warnings disabled");
	end;
end;
SlashCmdList["STATLOGIC_WARNING"] = ToggleWarnings
SLASH_STATLOGIC_WARNING1 = "/slwarning"

function PrintCommandHelp()
	print(ORANGE_FONT_COLOR_CODE.."LibStatLogic"..FONT_COLOR_CODE_CLOSE..": Arguments to /tp:");
	print("   "..LIGHTYELLOW_FONT_COLOR_CODE.."debug"..FONT_COLOR_CODE_CLOSE.." - toggle debugging message");
	print("   "..LIGHTYELLOW_FONT_COLOR_CODE.."warning"..FONT_COLOR_CODE_CLOSE.." - toggle display of warning messages of unknown item text");
	print("   "..LIGHTYELLOW_FONT_COLOR_CODE.."locale"..FONT_COLOR_CODE_CLOSE.." - display core localization info used by LibStatLogic");
end;

function DumpLocaleInfo()
	local blueColorCode = "|cff88ccff";

	print(string.format(ORANGE_FONT_COLOR_CODE.."LibStatLogic"..FONT_COLOR_CODE_CLOSE.." version %s.%s", StatLogic.Major, StatLogic.Minor));
	print(string.format("   "..blueColorCode.."GetLocale()"..FONT_COLOR_CODE_CLOSE.."=\"%s\"", GetLocale() ));
	print(string.format("   "..blueColorCode.."L.LOCALE_STHOUSAND"..FONT_COLOR_CODE_CLOSE.."=\"%s\" "..LIGHTYELLOW_FONT_COLOR_CODE.."Character used to separate groups of digits e.g. \",\"", L.LOCALE_STHOUSAND));
	print(string.format("   "..blueColorCode.."L.LOCALE_SDECIMAL"..FONT_COLOR_CODE_CLOSE.."=\"%s\" "..LIGHTYELLOW_FONT_COLOR_CODE.."Character(s) used for the decimal separator  e.g. \".\"", L.LOCALE_SDECIMAL));
	
	print(string.format("   "..blueColorCode.."L.patNumber"..FONT_COLOR_CODE_CLOSE.."=\"%s\" "..LIGHTYELLOW_FONT_COLOR_CODE.."regular expression to find a localized number e.g. \%d+[,\%d]*", L.patNumber));
	print(string.format("   "..blueColorCode.."L.patDecimal"..FONT_COLOR_CODE_CLOSE.."=\"%s\" "..LIGHTYELLOW_FONT_COLOR_CODE.."regex to find a localized decimal number e.g. \%d+[,\%d]*.?\%d", L.patDecimal));
end;

SlashCmdList["STATLOGIC_SLASHCMD"] = function(msg)
	if (msg == "debug") then
		ToggleDebug();
	elseif (msg == "warning") or (msg == "warnings") then
		ToggleWarnings();
	elseif (msg == "?") or (msg == "help") or (msg == "") then
		PrintCommandHelp();
	elseif (msg == "locale") or (msg == "dumplocaleinfo") then
		DumpLocaleInfo();
	else
		print("/sl "..msg.."   huh?");
	end;
end;
SLASH_STATLOGIC_SLASHCMD1 = "/sl";
	


-----------------
-- Debug Tools --
-----------------
local function debugPrint(text)
	if DEBUG == true then
		print(text)
	end
end

--5.0.4 (build 16016) GetPrimaryTalentTree() removed and replaced with GetSpecialization()
--Create a compatibility function for pre-Mists
local GetSpecializationBase = GetSpecialization;
local function GetSpecialization(isInspect, isPet)
--	local spec;
	if (wowBuildNo >= 16016) then --v5.0.4
		return GetSpecializationBase(isInspect, isPet);
	else
		return GetPrimaryTalentTree(isInspect, isPet);
	end;
	
	--print(spec);
	--return spec;
end;

--[[---------------------------------
  :SetTip(item)
-------------------------------------
Notes:
  * This is a debugging tool for localizers
  * Displays item in ItemRefTooltip
  * item:
  :;itemId : number - The numeric ID of the item. ie. 12345
  :;"itemString" : string - The full item ID in string format, e.g. "item:12345:0:0:0:0:0:0:0".
  :::Also supports partial itemStrings, by filling up any missing ":x" value with ":0", e.g. "item:12345:0:0:0"
  :;"itemName" : string - The Name of the Item, ex: "Hearthstone"
  :::The item must have been equiped, in your bags or in your bank once in this session for this to work.
  :;"itemLink" : string - The itemLink, when Shift-Clicking items.
Arguments:
  number or string - itemId or "itemString" or "itemName" or "itemLink"
Returns:
  None
Example:
  StatLogic:SetTip("item:3185:0:0:0:0:0:1957")
-----------------------------------]]
function StatLogic:SetTip(item)
	local name, link, _, _, reqLv, _, _, _, itemType = GetItemInfo(item)
	if not link then
		DEFAULT_CHAT_FRAME:AddMessage("|c00ff0000Item not in local cache. Run '/item itemid' to quary server(requires Sniff addon).|r")
		return
	end
	ItemRefTooltip:ClearLines()
	ShowUIPanel(ItemRefTooltip)
	if ( not ItemRefTooltip:IsShown() ) then
		ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
	end
	ItemRefTooltip:SetHyperlink(link)
end

-------------------------
-- Localization Tables --
-------------------------
--[[
Localization tips
1. Enable debugging in game with /sldebug
2. There are often ItemIDs in comments for you to check if the items works correctly in game
3. Use the addon Sniff(http://www.wowinterface.com/downloads/info6259/) to get a link in game from an ItemID, Usage: /item 19316
4. Atlas + AtlasLoot is also a good source of items to check
5. Red colored text output from debug means that line does not have a match
6. Building your own ItemStrings(ex: "item:28484:1503:0:2946:2945:0:0:0"): http://www.wowwiki.com/ItemString
   linkType, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId
7. Getting a visual on your ItemString:
   a. Display the ItemRefTooltip by clicking on a link in chat
   b. /dump StatLogic:SetTip("item:23344:2746")
6. Testing Enchants:
   a. Obtain the enchantId from wowhead.
      EX: Find the enchantId for [Golden Spellthread]:
      Find the spell page for the enchant: http://www.wowhead.com/?spell=31370
      Under Spell Details, look for "Enchant Item Permanent (2746)"
      2746 is the enchantId
   b. We need an item to attach the enchant, I like to use [Scout's Pants] ID:23344. (/item 23344 if you don't have it in your cache)
      ItemString: "item:23344:2746"
   c. /dump StatLogic:GetSum("item:23344:2746")
--]]
--[[
-- Item Stat Scanning Procedure
-- Trim spaces using strtrim(text)
-- Strip color codes
-- 1. Fast Exclude - Exclude obvious lines that do not need to be checked
--    Exclude a string by matching the whole string, these strings are indexed in L.Exclude.
--    Exclude a string by looking at the first X chars, these strings are indexed in L.Exclude.
--    Exclude lines starting with '"'. (Flavor text)
--    Exclude lines that are not white and green and normal (normal for Frozen Wrath bonus)
-- 2. Whole Text Lookup - Mainly used for enchants or stuff without numbers
--    Whole strings are indexed in L.WholeTextLookup
-- 3. Single Plus Stat Check - "+10 Spirit"
--    String is matched with pattern L.SinglePlusStatCheck, 2 captures are returned.
--    If a match is found, the non-number capture is looked up in L.StatIDLookup.
-- 4. Single Equip Stat Check - "Equip: Increases attack power by 81."
--    String is matched with pattern L.SingleEquipStatCheck, 2 captures are returned.
--    If a match is found, the non-number capture is looked up in L.StatIDLookup.
-- 5. Pre Scan - Short list of patterns that will be checked before going into Deep Scan.
-- 6. Deep Scan - When all the above checks fail, we will use the Deep Scan, this is slow but only about 10% of the lines need it.
--    Strip leading "Equip: ", "Socket Bonus: ".
--    Strip trailing ".".
--    Separate the string using L.DeepScanSeparators.
--    Check if the separated strings are found in L.WholeTextLookup.
--    Try to match each separated string to patterns in L.DeepScanPatterns in order, patterns in L.DeepScanPatterns should have less inclusive patterns listed first and more inclusive patterns listed last.
--    If no match then separate the string using L.DeepScanWordSeparators, then check again.
--]]
--[[DEBUG stuff, no need to translate
textTable = {
  "Spell Damage +6 and Spell Hit Rating +5",
  "+3 Stamina, +4 Critical Strike Rating",
  "+26 Healing Spells & 2% Reduced Threat",
  "+3 Stamina/+4 Critical Strike Rating",
  "Socket Bonus: 2 mana per 5 sec.",
  "Equip: Increases damage and healing done by magical spells and effects by up to 150.",
  "Equip: Increases the spell critical strike rating of all party members within 30 yards by 28.",
  "Equip: Increases damage and healing done by magical spells and effects of all party members within 30 yards by up to 33.",
  "Equip: Increases healing done by magical spells and effects of all party members within 30 yards by up to 62.",
  "Equip: Increases your spell damage by up to 120 and your healing by up to 300.",
  "Equip: Restores 11 mana per 5 seconds to all party members within 30 yards.",
  "Equip: Increases healing done by spells and effects by up to 300.",
  "Equip: Increases attack power by 420 in Cat, Bear, Dire Bear, and Moonkin forms only.",
  "+10 Defense Rating/+10 Stamina/+15 Block Value", -- ZG Enchant
  "+26 Attack Power and +14 Critical Strike Rating", -- Swift Windfire Diamond ID:28556
  "+26 Healing Spells & 2% Reduced Threat", -- Bracing Earthstorm Diamond ID:25897
  "+6 Spell Damage, +5 Spell Crit Rating", -- Potent Ornate Topaz ID: 28123
  ----
  "Critical Rating +6 and Dodge Rating +5", -- Assassin's Fire Opal ID:30565
  "Healing +11 and 2 mana per 5 sec.", -- Royal Tanzanite ID: 30603
}
--]]
--local PatternLocale = {};
--local DisplayLocale = {};

local locale = GetLocale(); --BlizzardAPI  e.g. "enUS"
local noPatternLocale --flag is set if we have no matching PatternLocale, and are falling back to enUS
L = PatternLocale[locale]
if not L then
	noPatternLocale = true;
	L = PatternLocale.enUS;
	
	print(ORANGE_FONT_COLOR_CODE.."LibStatLogic"..FONT_COLOR_CODE_CLOSE..": Unknown locale "..locale..". Falling back to enUS."..FONT_COLOR_CODE_CLOSE);
	assert(L, "No PatternLocale information loaded!");
end

-- Setup any values from current DisplayLocale into enUSDisplayLocale, since that's the gold version
local function SetupDisplayLocale(locale, enUS)
	for k in pairs(enUS) do
		--If the k-th item is a table, the recursively call SetupDisplayLocale
		if type(enUS[k]) == "table" and type(locale[k]) == "table" then
			SetupDisplayLocale(locale[k], enUS[k])
		elseif locale[k] then
			enUS[k] = locale[k]
		end
	end
end

--copy DisplayLocale[locale] items into enUS
if DisplayLocale[locale] then
	SetupDisplayLocale(DisplayLocale[locale], DisplayLocale.enUS)
end
local D = DisplayLocale.enUS;

--Dispose of the unused locale variables
PatternLocale = nil; 
DisplayLocale = nil;

local function EscapeRegularExpression(pattern)
	-- "." ==> "%."
	local s = string.gsub(pattern, "%.", "%%%.");

	return s;
end;

function string.replace(s, oldValue, newValue)
	oldValue = EscapeRegularExpression(oldValue);
	newValue = EscapeRegularExpression(newValue);

	return string.gsub(s, oldValue, newValue);
end;

--[[
	20120805
	With Mista of Panderia (5.0.1), item stat numbers are now given as localized numbers. e.g.:
		Before: "2632 Armor"
		After:  "2,632 Armor"
		
	This means that regular expressions of (%d+) to find a number, e.g.:
		String: 2632 Armor
		Find:   ^(%d+) Armor$
		
	Will no longer work. All references to (%d+) need to be replaced with (%d+[,%d]*)
	
		String: 2,632 Armor
		Find:   ^(%d+[,%d]*) Armor$
		
	For this purpose there are 4 localization (L) entries
		L.LOCALE_STHOUSAND  The current locale's thousand's separator (e.g. enUS: ","  deDE: "." )
		L.LOCALE_SDECIMAL   The current locale's decimal separator (e.g. enUS: "."   deDE: "," )
		L.patNumber         The regular expression to find an Integer number (e.g. 1.234.567)
		L.patDecimal        The regular expression to find a Decimal number  (e.g. 1.234.567,80)

	The problem then is to remove the comma and parse the remaining number.	
	
	lua's tonumber(string) will not handle commas:
	
		tonumber("123")    -> 123
		tonumber("123.45") -> 123.45
		tonumber("1,234")  -> nil
	
	So we need to first remove the commas:
	
		string.gsub("1,234", ",", "") -> "1234"
		
	But, we also have to handle the nonsense part that in Germany, number's are *displayed* as:
			1.234,56
			
	but Lua's tonumber() funtion, even on deDE client, can only understand:
			1234.56
			
	e.g.
			/dump tonumber("5.50")
			5.5
			/dump tonumber("5,50")
			nil
			/dump tonumber("1,234.56")
			nil
			/dump tonumber("1.234,56")
			nil

			/dump GetLocale()
			"deDE"	
			
	So we have to replace the DecimalSeparator with a period.
--]]
function CNumberEx(value, base, localization)
	--[[
	if (value == nil) then
		error("Cannot convert nil to a number", 2);
	end;
	--]]
	assert(localization, "CNumberEx: Argument localization is nil");
	
	local t = type(value);
	if (t == "string") then
		local n = tonumber(value, base);
		if n then
			--print("tonumber("..value..") worked, returning "..tostring(n));
			return n;
		else
			--print("Stripping thousands \""..(localization.LOCALE_STHOUSAND).."\"");
			
			--strip out potential thousand's separators
			assert(localization.LOCALE_STHOUSAND);
			local s = string.replace(value, localization.LOCALE_STHOUSAND, "");
			
			--print("New string: \""..s.."\"");
			
			--convert any decimal markers into a "." (period)
			assert(localization.LOCALE_SDECIMAL);
			if (localization.LOCALE_SDECIMAL ~= ".") then
				s = string.replace(s, localization.LOCALE_SDECIMAL, ".");
			end;
			
			--print("decimal's fixed: \""..s.."\"");
			
			n = tonumber(s, base);
			return n;
			--if n then
			--	return n;
			--else
			--	error(string.format("Could not convert value \"%s\" to a number.", value), 2);
			--end;
		end;
	else
		local n = tonumber(value, base);
		return n;
		--if n then
		--	return n;
		--else
		--	error(string.format("Could not convert value \"%s\" to a number.", value), 2);
		--end;
	end;
end;

function CNumber(value, base)
	return CNumberEx(value, base, L);
end;

--Create a tonumber function inside our localization table that is the same as CNumber
L["tonumber"] = CNumber;
--[[	
	["tonumber"] = function(s)
		local n = tonumber(s);
		if n then
			return n;
		else
			--strip thousands separators
			return tonumber((gsub(s, LOCALE_STHOUSAND, "")));
		end
	end,
--]]	

-- Add all lower case strings to ["StatIDLookup"]
local strutf8lower = string.utf8lower
local temp = {}
for k, v in pairs(L.StatIDLookup) do
	temp[strutf8lower(k)] = v
end
for k, v in pairs(temp) do
	L.StatIDLookup[k] = v
end
temp = nil

--[[---------------------------------
  :GetStatNameFromID(stat)
-------------------------------------
Notes:
  * Returns localized names for stat
Arguments:
  string - "StatID". ex: "DODGE", "DODGE_RATING"
Returns:
  ; "longName" : string - The full name for stat.
  ; "shortName" : string - The short name for stat.
Example:
  local longName, shortName = StatLogic:GetStatNameFromID("FIRE_RES") -- "Fire Resistance", "FR"
-----------------------------------]]
function StatLogic:GetStatNameFromID(stat)
	local name = D.StatIDToName[stat]
	if not name then return end
	return unpack(name)
end


-----------
-- Cache --
-----------
local cache = {}
setmetatable(cache, {__mode = "kv"}) -- weak table to enable garbage collection


--------------
-- Activate --
--------------
-- When a newer version is registered
local tip = StatLogic.tip
if not tip then
	-- Create a custom tooltip for scanning
	tip = CreateFrame("GameTooltip", MAJOR.."Tooltip", nil, "GameTooltipTemplate")
	StatLogic.tip = tip
	tip:SetOwner(UIParent, "ANCHOR_NONE")
	for i = 1, 40 do
		tip[i] = _G[MAJOR.."TooltipTextLeft"..i]
		if not tip[i] then
			tip[i] = tip:CreateFontString()
			tip:AddFontStrings(tip[i], tip:CreateFontString())
			_G[MAJOR.."TooltipTextLeft"..i] = tip[i]
		end
	end
elseif not _G[MAJOR.."TooltipTextLeft40"] then
	for i = 1, 40 do
		_G[MAJOR.."TooltipTextLeft"..i] = tip[i]
	end
end
local tipMiner = StatLogic.tipMiner
if not tipMiner then
	-- Create a custom tooltip for data mining
	tipMiner = CreateFrame("GameTooltip", MAJOR.."MinerTooltip", nil, "GameTooltipTemplate")
	StatLogic.tipMiner = tipMiner
	tipMiner:SetOwner(UIParent, "ANCHOR_NONE")
	for i = 1, 40 do
		tipMiner[i] = _G[MAJOR.."MinerTooltipTextLeft"..i]
		if not tipMiner[i] then
			tipMiner[i] = tipMiner:CreateFontString()
			tipMiner:AddFontStrings(tipMiner[i], tipMiner:CreateFontString())
			_G[MAJOR.."MinerTooltipTextLeft"..i] = tipMiner[i]
		end
	end
elseif not _G[MAJOR.."MinerTooltipTextLeft40"] then
	for i = 1, 40 do
		_G[MAJOR.."MinerTooltipTextLeft"..i] = tipMiner[i]
	end
end
local StatLogicMinerTooltipTextLeft5 = _G[MAJOR.."MinerTooltipTextLeft5"]
local StatLogicMinerTooltipTextLeft6 = _G[MAJOR.."MinerTooltipTextLeft6"]

---------------------
-- Local Variables --
---------------------
-- Player info
local _, playerClass = UnitClass("player")
local _, playerRace = UnitRace("player")

-- Localize globals
local tonumber = L.tonumber; --Use the localized version of tonumber from localization table L
local _G = getfenv(0)
local strfind = strfind
local strsub = strsub
local strupper = strupper
local strmatch = strmatch
local strtrim = strtrim
local strsplit = strsplit
local strjoin = strjoin
local gmatch = gmatch
local gsub = gsub
local wipe = wipe
local pairs = pairs
local ipairs = ipairs
local type = type
local unpack = unpack
local strutf8lower = string.utf8lower
local strutf8sub = string.utf8sub
local GetInventoryItemLink = GetInventoryItemLink
local IsUsableSpell = IsUsableSpell
local UnitLevel = UnitLevel
local UnitStat = UnitStat
local GetShapeshiftForm = GetShapeshiftForm
local GetShapeshiftFormInfo = GetShapeshiftFormInfo
local GetTalentInfo = GetTalentInfo
local GetInventoryItemID = GetInventoryItemID
local GetSpellInfo = GetSpellInfo
local GetCVarBool = GetCVarBool

-- Cached GetItemInfo
local GetItemInfoCached = setmetatable({}, { __index = function(self, n)
		local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemCount, itemEquipLoc, itemTexture, itemSellValue = GetItemInfo(n)
		if itemName then
			-- store in cache only if it exists in the local cache
			self[n] = {itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemCount, itemEquipLoc, itemTexture, itemSellValue}
			return self[n] -- return result
		end
	end })
local GetItemInfo = function(item)
	local info = GetItemInfoCached[item]
	if info then
		return unpack(info)
	end
end
StatLogic.GetItemInfo = GetItemInfo

-- taken from lua programming gems
local function memoize(f)
	local mem = {} -- memoizing table
	return function(x)
		if not mem[x] then
			mem[x] = f(x)
		end
		return mem[x]
	end
end
local loadstring = memoize(loadstring)

---------------
-- Lua Tools --
---------------
-- metatable for stat tables
local statTableMetatable = {
	__add = function(op1, op2)
		if type(op2) == "table" then
			for k, v in pairs(op2) do
				if type(v) == "number" then
					op1[k] = (op1[k] or 0) + v
					if op1[k] == 0 then
						op1[k] = nil
					end
				end
			end
		end
		return op1
	end,
	__sub = function(op1, op2)
		if type(op2) == "table" then
			for k, v in pairs(op2) do
				if type(v) == "number" then
					op1[k] = (op1[k] or 0) - v
					if op1[k] == 0 then
						op1[k] = nil
					end
				elseif k == "itemType" then
					local i = 1
					while op1["diffItemType"..i] do
						i = i + 1
					end
					op1["diffItemType"..i] = op2.itemType
				end
			end
		end
		return op1
	end,
}

-- Table pool borrowed from Tablet-2.0 (ckknight) --
local pool = {}

-- Delete table and return to pool
local function del(t)
	if t then
		for k in pairs(t) do
			t[k] = nil
		end
		setmetatable(t, nil)
		pool[t] = true
	end
end

local function delMulti(t)
	if t then
		for k in pairs(t) do
			if type(t[k]) == "table" then
				del(t[k])
			else
				t[k] = nil
			end
		end
		setmetatable(t, nil)
		pool[t] = true
	end
end

-- Copy table
local function copy(parent)
	local t = next(pool) or {}
	pool[t] = nil
	if parent then
		for k,v in pairs(parent) do
			t[k] = v
		end
		setmetatable(t, getmetatable(parent))
	end
	return t
end

-- New table
local function new(...)
	local t = next(pool) or {}
	pool[t] = nil

	for i = 1, select('#', ...), 2 do
		local k = select(i, ...)
		if k then
			t[k] = select(i+1, ...)
		else
			break
		end
	end
	return t
end

-- New stat table
local function newStatTable(...)
	local t = next(pool) or {}
	pool[t] = nil
	setmetatable(t, statTableMetatable)

	for i = 1, select('#', ...), 2 do
		local k = select(i, ...)
		if k then
			t[k] = select(i+1, ...)
		else
			break
		end
	end
	return t
end

StatLogic.StatTable = {}
StatLogic.StatTable.new = newStatTable
StatLogic.StatTable.del = del
StatLogic.StatTable.copy = copy

-- End of Table pool --

-- deletes the contents of a table, then returns itself
local function clearTable(t)
	if t then
		for k in pairs(t) do
			if type(t[k]) == "table" then
				del(t[k]) -- child tables get put into the pool
			else
				t[k] = nil
			end
		end
		setmetatable(t, nil)
	end
	return t
end

-- copyTable
local function copyTable(to, from)
	if not clearTable(to) then
		to = new()
	end
	for k,v in pairs(from) do
		if type(k) == "table" then
			k = copyTable(new(), k)
		end
		if type(v) == "table" then
			v = copyTable(new(), v)
		end
		to[k] = v
	end
	setmetatable(to, getmetatable(from))
	return to
end

-- Taken from AceLibrary
function StatLogic:argCheck(arg, num, kind, kind2, kind3, kind4, kind5)
	if type(num) ~= "number" then
		return error(string.format("Bad argument #3 to `argCheck' (number expected, got %s)", type(num)), 2);
	elseif type(kind) ~= "string" then
		return error(string.format("Bad argument #4 to `argCheck' (string expected, got %s)", type(kind)), 2);
	end
	arg = type(arg)
	if arg ~= kind and arg ~= kind2 and arg ~= kind3 and arg ~= kind4 and arg ~= kind5 then
		local stack = debugstack()
		local func = stack:match("`argCheck'.-([`<].-['>])")
		if not func then
			func = stack:match("([`<].-['>])")
		end
		if kind5 then
			return error(string.format("Bad argument #%s to %s (%s, %s, %s, %s, or %s expected, got %s)", tonumber(num) or 0/0, func, kind, kind2, kind3, kind4, kind5, arg), 2);
		elseif kind4 then
			return error(string.format("Bad argument #%s to %s (%s, %s, %s, or %s expected, got %s)", tonumber(num) or 0/0, func, kind, kind2, kind3, kind4, arg), 2);
		elseif kind3 then
			return error(string.format("Bad argument #%s to %s (%s, %s, or %s expected, got %s)", tonumber(num) or 0/0, func, kind, kind2, kind3, arg), 2);
		elseif kind2 then
			return error(string.format("Bad argument #%s to %s (%s or %s expected, got %s)", tonumber(num) or 0/0, func, kind, kind2, arg), 2);
		else
			return error(string.format("Bad argument #%s to %s (%s expected, got %s)", tonumber(num) or 0/0, func, kind, arg), 2);
		end
	end
end

--Starting with 5.0.4, Blizzard's UnitClass() function now also returns a "classID" number.
--But since they don't define the constants themselves, and we've always used numeric constants
--we'll adopt their numbers, and define the matching constants
local CLASSID_WARRIOR = 1;
local CLASSID_PALADIN = 2;
local CLASSID_HUNTER = 3;
local CLASSID_ROGUE = 4;
local CLASSID_PRIEST = 5;
local CLASSID_DEATHKNIGHT = 6;
local CLASSID_SHAMAN = 7;
local CLASSID_MAGE = 8;
local CLASSID_WARLOCK = 9;
local CLASSID_MONK = 10;
local CLASSID_DRUID = 11;
local CLASSID_MAX = CLASSID_DRUID; --the highest valid classID

local CR_PVP_POWER = 27; --copied from Constants.lua until Mists launches
local CR_MAX = CR_PVP_POWER; --Maximum possible CombatRating constant. Some code checks for passed combat rating value being too high.

--[[---------------------------------
  :GetClassIdOrName(class)
-------------------------------------
Notes:
  * Converts ClassID to and from "ClassName"
  * class:
  :{| class="wikitable"
  !ClassID!!"ClassName"
  |-
  |1||"WARRIOR"
  |-
  |2||"PALADIN"
  |-
  |3||"HUNTER"
  |-
  |4||"ROGUE"
  |-
  |5||"PRIEST"
  |-
  |6||"DEATHKNIGHT"
  |-
  |7||"SHAMAN"
  |-
  |8||"MAGE"
  |-
  |9||"WARLOCK"
  |-
  |10||"MONK"
  |-
  |11||"DRUID"
  |}
Arguments:
  number or string - ClassID or "ClassName"
Returns:
  None
Example:
  StatLogic:GetClassIdOrName("WARRIOR") -- 1
  StatLogic:GetClassIdOrName(10) -- "DRUID"
-----------------------------------]]
local ClassNameToID = {
	"WARRIOR", --1
	"PALADIN", --2
	"HUNTER",  --3
	"ROGUE",   --4
	"PRIEST",  --5
	"DEATHKNIGHT", --6
	"SHAMAN", --7
	"MAGE", --8
	"WARLOCK", --9
	"MONK", --10
	"DRUID", --11
	["WARRIOR"] = 1,
	["PALADIN"] = 2,
	["HUNTER"] = 3,
	["ROGUE"] = 4,
	["PRIEST"] = 5,
	["DEATHKNIGHT"] = 6,
	["SHAMAN"] = 7,
	["MAGE"] = 8,
	["WARLOCK"] = 9,
	["MONK"] = 10,
	["DRUID"] = 11,
}

function StatLogic:GetClassIdOrName(class)
  return ClassNameToID[class]
end

--[[ Interface\FrameXML\PaperDollFrame.lua
CR_WEAPON_SKILL = 1;
CR_DEFENSE_SKILL = 2;
CR_DODGE = 3;
CR_PARRY = 4;
CR_BLOCK = 5;
CR_HIT_MELEE = 6;
CR_HIT_RANGED = 7;
CR_HIT_SPELL = 8;
CR_CRIT_MELEE = 9;
CR_CRIT_RANGED = 10;
CR_CRIT_SPELL = 11;
CR_HIT_TAKEN_MELEE = 12;
CR_HIT_TAKEN_RANGED = 13;
CR_HIT_TAKEN_SPELL = 14;
COMBAT_RATING_RESILIENCE_CRIT_TAKEN = 15; formerly CR_CRIT_TAKEN_MELEE
COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN = 16;  formerly CR_CRIT_TAKEN_RANGED
CR_CRIT_TAKEN_SPELL = 17;
CR_HASTE_MELEE = 18;
CR_HASTE_RANGED = 19;
CR_HASTE_SPELL = 20;
CR_WEAPON_SKILL_MAINHAND = 21;
CR_WEAPON_SKILL_OFFHAND = 22;
CR_WEAPON_SKILL_RANGED = 23;
CR_EXPERTISE = 24;
CR_ARMOR_PENETRATION = 25;
CR_MASTERY = 26;
CR_PVP_POWER = 27;
--]]

local RatingNameToID = {
	[CR_WEAPON_SKILL] = "WEAPON_RATING",
	[CR_DEFENSE_SKILL] = "DEFENSE_RATING",
	[CR_DODGE] = "DODGE_RATING",
	[CR_PARRY] = "PARRY_RATING",
	[CR_BLOCK] = "BLOCK_RATING",
	[CR_HIT_MELEE] = "MELEE_HIT_RATING",
	[CR_HIT_RANGED] = "RANGED_HIT_RATING",
	[CR_HIT_SPELL] = "SPELL_HIT_RATING",
	[CR_CRIT_MELEE] = "MELEE_CRIT_RATING",
	[CR_CRIT_RANGED] = "RANGED_CRIT_RATING",
	[CR_CRIT_SPELL] = "SPELL_CRIT_RATING",
	[CR_HIT_TAKEN_MELEE] = "MELEE_HIT_AVOID_RATING",
	[CR_HIT_TAKEN_RANGED] = "RANGED_HIT_AVOID_RATING",
	[CR_HIT_TAKEN_SPELL] = "SPELL_HIT_AVOID_RATING",
	[COMBAT_RATING_RESILIENCE_CRIT_TAKEN] = "MELEE_CRIT_AVOID_RATING",
	[COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN] = "RANGED_CRIT_AVOID_RATING",
	[CR_CRIT_TAKEN_SPELL] = "SPELL_CRIT_AVOID_RATING",
	[CR_HASTE_MELEE] = "MELEE_HASTE_RATING",
	[CR_HASTE_RANGED] = "RANGED_HASTE_RATING",
	[CR_HASTE_SPELL] = "SPELL_HASTE_RATING",
	[CR_WEAPON_SKILL_MAINHAND] = "MAINHAND_WEAPON_RATING",
	[CR_WEAPON_SKILL_OFFHAND] = "OFFHAND_WEAPON_RATING",
	[CR_WEAPON_SKILL_RANGED] = "RANGED_WEAPON_RATING",
	[CR_EXPERTISE] = "EXPERTISE_RATING",
	[CR_ARMOR_PENETRATION] = "ARMOR_PENETRATION_RATING",
	[CR_MASTERY] = "MASTERY_RATING",
	[CR_PVP_POWER] = "PVP_POWER", --27;
	["DEFENSE_RATING"] = CR_DEFENSE_SKILL,
	["DODGE_RATING"] = CR_DODGE,
	["PARRY_RATING"] = CR_PARRY,
	["BLOCK_RATING"] = CR_BLOCK,
	["MELEE_HIT_RATING"] = CR_HIT_MELEE,
	["RANGED_HIT_RATING"] = CR_HIT_RANGED,
	["SPELL_HIT_RATING"] = CR_HIT_SPELL,
	["MELEE_CRIT_RATING"] = CR_CRIT_MELEE,
	["RANGED_CRIT_RATING"] = CR_CRIT_RANGED,
	["SPELL_CRIT_RATING"] = CR_CRIT_SPELL,
	["MELEE_HIT_AVOID_RATING"] = CR_HIT_TAKEN_MELEE,
	["RANGED_HIT_AVOID_RATING"] = CR_HIT_TAKEN_RANGED,
	["SPELL_HIT_AVOID_RATING"] = CR_HIT_TAKEN_SPELL,
	["MELEE_CRIT_AVOID_RATING"] = COMBAT_RATING_RESILIENCE_CRIT_TAKEN,
	["RANGED_CRIT_AVOID_RATING"] = COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN,
	["SPELL_CRIT_AVOID_RATING"] = CR_CRIT_TAKEN_SPELL,
	["RESILIENCE_RATING"] = COMBAT_RATING_RESILIENCE_CRIT_TAKEN,
	["MELEE_HASTE_RATING"] = CR_HASTE_MELEE,
	["RANGED_HASTE_RATING"] = CR_HASTE_RANGED,
	["SPELL_HASTE_RATING"] = CR_HASTE_SPELL,
	["DAGGER_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["SWORD_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["2H_SWORD_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["AXE_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["2H_AXE_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["MACE_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["2H_MACE_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["GUN_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["CROSSBOW_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["BOW_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["FERAL_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["FIST_WEAPON_RATING"] = CR_WEAPON_SKILL,
	["WEAPON_RATING"] = CR_WEAPON_SKILL,
	["MAINHAND_WEAPON_RATING"] = CR_WEAPON_SKILL_MAINHAND,
	["OFFHAND_WEAPON_RATING"] = CR_WEAPON_SKILL_OFFHAND,
	["RANGED_WEAPON_RATING"] = CR_WEAPON_SKILL_RANGED,
	["EXPERTISE_RATING"] = CR_EXPERTISE,
	["ARMOR_PENETRATION_RATING"] = CR_ARMOR_PENETRATION,
	["MASTERY_RATING"] = CR_MASTERY, --26
	["PVP_POWER"] = CR_PVP_POWER, --27
}

--[[---------------------------------
  :GetRatingIdOrName(rating)
-------------------------------------
Notes:
  * Converts RatingID to and from "StatID"
  * rating:
  :;RatingID : number - As defined in PaperDollFrame.lua of Blizzard default ui
  :;"StatID" : string - The the key values of the DisplayLocale table in StatLogic
  :{| class="wikitable"
  !RatingID!!"StatID"
  |-
  |CR_WEAPON_SKILL||"WEAPON_RATING"
  |-
  |CR_DEFENSE_SKILL||"DEFENSE_RATING"
  |-
  |CR_DODGE||"DODGE_RATING"
  |-
  |CR_PARRY||"PARRY_RATING"
  |-
  |CR_BLOCK||"BLOCK_RATING"
  |-
  |CR_HIT_MELEE||"MELEE_HIT_RATING"
  |-
  |CR_HIT_RANGED||"RANGED_HIT_RATING"
  |-
  |CR_HIT_SPELL||"SPELL_HIT_RATING"
  |-
  |CR_CRIT_MELEE||"MELEE_CRIT_RATING"
  |-
  |CR_CRIT_RANGED||"RANGED_CRIT_RATING"
  |-
  |CR_CRIT_SPELL||"SPELL_CRIT_RATING"
  |-
  |CR_HIT_TAKEN_MELEE||"MELEE_HIT_AVOID_RATING"
  |-
  |CR_HIT_TAKEN_RANGED||"RANGED_HIT_AVOID_RATING"
  |-
  |CR_HIT_TAKEN_SPELL||"SPELL_HIT_AVOID_RATING"
  |-
  |COMBAT_RATING_RESILIENCE_CRIT_TAKEN||"MELEE_CRIT_AVOID_RATING"
  |-
  |COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN||"RANGED_CRIT_AVOID_RATING"
  |-
  |CR_CRIT_TAKEN_SPELL||"SPELL_CRIT_AVOID_RATING"
  |-
  |CR_HASTE_MELEE||"MELEE_HASTE_RATING"
  |-
  |CR_HASTE_RANGED||"RANGED_HASTE_RATING"
  |-
  |CR_HASTE_SPELL||"SPELL_HASTE_RATING"
  |-
  |CR_WEAPON_SKILL_MAINHAND||"MAINHAND_WEAPON_RATING"
  |-
  |CR_WEAPON_SKILL_OFFHAND||"OFFHAND_WEAPON_RATING"
  |-
  |CR_WEAPON_SKILL_RANGED||"RANGED_WEAPON_RATING"
  |-
  |CR_EXPERTISE||"EXPERTISE_RATING"
  |-
  |CR_ARMOR_PENETRATION||"ARMOR_PENETRATION_RATING"
  |}
Arguments:
  number or string - RatingID or "StatID"
Returns:
  None
Example:
  StatLogic:GetRatingIdOrStatId("CR_WEAPON_SKILL") -- 1
  StatLogic:GetRatingIdOrStatId("DEFENSE_RATING") -- 2
  StatLogic:GetRatingIdOrStatId("DODGE_RATING") -- 3
  StatLogic:GetRatingIdOrStatId(CR_PARRY) -- "PARRY_RATING"
-----------------------------------]]
function StatLogic:GetRatingIdOrStatId(rating)
	return RatingNameToID[rating]
end

local RatingIDToConvertedStat = {
	"WEAPON_SKILL",			--CR_WEAPON_SKILL	1
	"DEFENSE",				--CR_DEFENSE_SKILL	2
	"DODGE",				--CR_DODGE	3
	"PARRY",				--CR_PARRY	4
	"BLOCK",				--CR_BLOCK	5
	"MELEE_HIT",			--CR_HIT_MELEE	6
	"RANGED_HIT",			--CR_HIT_RANGED	7
	"SPELL_HIT",			--CR_HIT_SPELL	8
	"MELEE_CRIT",			--CR_CRIT_MELEE	9
	"RANGED_CRIT",			--CR_CRIT_RANGED	10
	"SPELL_CRIT",			--CR_CRIT_SPELL	11
	"MELEE_HIT_AVOID",		--CR_HIT_TAKEN_MELEE	12
	"RANGED_HIT_AVOID",		--CR_HIT_TAKEN_RANGED	13
	"SPELL_HIT_AVOID",		--CR_HIT_TAKEN_SPELL	14
	"MELEE_CRIT_AVOID",		--COMBAT_RATING_RESILIENCE_CRIT_TAKEN	15
	"PLAYER_DAMAGE_TAKEN",	--COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN	16
	"SPELL_CRIT_AVOID",		--CR_CRIT_TAKEN_SPELL	18
	"MELEE_HASTE",			--CR_HASTE_MELEE	18
	"RANGED_HASTE",			--CR_HASTE_RANGED	19
	"SPELL_HASTE",			--CR_HASTE_SPELL	20
	"WEAPON_SKILL",			--CR_WEAPON_SKILL_MAINHAND	21
	"WEAPON_SKILL",			--CR_WEAPON_SKILL_OFFHAND	22
	"WEAPON_SKILL",			--CR_WEAPON_SKILL_RANGED	23
	"EXPERTISE",			--CR_EXPERTISE	24
	"ARMOR_PENETRATION",	--CR_ARMOR_PENETRATION	25
	"MASTERY",				--CR_MASTERY	26
	"PVP_POWER",			--CR_PVP_POWER	27
}


----------------
-- Stat Tools --
----------------
local function StripGlobalStrings(text)
	-- ITEM_SOCKET_BONUS = "Socket Bonus: %s"; -- Tooltip tag for socketed item matched socket bonuses
	text = gsub(text, "%%%%", "%%") -- "%%" -> "%"
	text = gsub(text, " ?%%%d?%.?%d?%$?[cdsgf]", "") -- delete "%d", "%s", "%c", "%g", "%2$d", "%.2f" and a space in front of it if found
	-- So StripGlobalStrings(ITEM_SOCKET_BONUS) = "Socket Bonus:"
	return text
end

local function GetStanceIcon()
	local currentStance = GetShapeshiftForm()
	if currentStance ~= 0 then
		return GetShapeshiftFormInfo(currentStance)
	end
end
StatLogic.GetStanceIcon = GetStanceIcon

local function PlayerHasAura(aura)
	return UnitBuff("player", aura or "") or UnitDebuff("player", aura or "")
end
StatLogic.PlayerHasAura = PlayerHasAura


local function GetPlayerAuraRankStack(buff)
	--name, rank, icon, stack, debuffType, duration, expirationTime, isMine, isStealable = UnitAura("player", buff)
	local name, rank, _, stack = UnitBuff("player", buff)
	if not name then -- if not a buff, check for debuff
		name, rank, _, stack = UnitDebuff("player", buff)
	end
	if name then
		if not stack or stack == 0 then
			stack = 1
		end
		return tonumber(strmatch(rank, "(%d+)") or 1), stack
	end
end
StatLogic.GetPlayerAuraRankStack = GetPlayerAuraRankStack

--[[
	20120812 - Removed, as defense no longer exists in the game
local function GetTotalDefense(unit)
	local base, modifier = UnitDefense(unit);
	return base + modifier
end
--]]

local function PlayerHasGlyph(glyph, talentGroup)
	for i = 1, 9 do
		local _, _, _, glyphSpellID = GetGlyphSocketInfo(i, talentGroup)
		if glyphSpellID == glyph then
			return true
		end
	end
end
StatLogic.PlayerHasGlyph = PlayerHasGlyph

---------------
-- Item Sets --
---------------
local SetToItem  = {
	-- SetID comes from ItemSet.dbc
	[571] = {24264, 24261}, -- Whitemend Wisdom - Increases spell power by 10% of your total Intellect.
	[552] = {21848, 21847, 21846}, -- Wrath of Spellfire - Increases spell power by 7% of your total Intellect.
}
-- Build ItemToSet from SetToItem
local ItemToSet = {}
for set, items in pairs(SetToItem) do
	for _, item in pairs(items) do
		ItemToSet[item] = set
	end
end

-- Create a frame
local ItemSetFrame = StatLogic.ItemSetFrame
if not ItemSetFrame then
	ItemSetFrame = CreateFrame("Frame", "StatLogicItemSetFrame")
	StatLogic.ItemSetFrame = ItemSetFrame
else
	ItemSetFrame:UnregisterAllEvents()
end

--[[
PlayerItemSets = {
  [844] = 4,
}
--]]
local PlayerItemSets = {}
-- API
function StatLogic:PlayerHasItemSet(itemset)
	return PlayerItemSets[itemset]
end
-- Don't set any scripts if the class doesn't have any sets to check
if table.maxn(ItemToSet) ~= 0 then
	local WatchInventoryID = {
		(GetInventorySlotInfo("HeadSlot")),
		(GetInventorySlotInfo("ShoulderSlot")),
		(GetInventorySlotInfo("ChestSlot")),
		(GetInventorySlotInfo("HandsSlot")),
		(GetInventorySlotInfo("LegsSlot")),
	}
	local function UpdatePlayerItemSets()
		wipe(PlayerItemSets)
		for _, slot in pairs(WatchInventoryID) do
			local set = ItemToSet[GetInventoryItemID('player', slot)]
			if set then
				PlayerItemSets[set] = (PlayerItemSets[set] or 0) + 1
			end
		end
	end
	-- we will schedule this since PLAYER_EQUIPMENT_CHANGED fires multiple times when you switch whole sets
	ItemSetFrame:SetScript("OnUpdate", function(self, elapsed)
		if self.updateTime and (GetTime() >= self.updateTime) then
			self.updateTime = nil
			UpdatePlayerItemSets()
		end
	end)
	ItemSetFrame:SetScript("OnEvent", function(self, event, ...) if self[event] then return self[event](self, event, ...) end end)
	ItemSetFrame:RegisterEvent("PLAYER_EQUIPMENT_CHANGED")
	function ItemSetFrame:PLAYER_EQUIPMENT_CHANGED()
		self.updateTime = GetTime() + 0.1 -- 0.1 sec delay
	end
	-- Initialize on PLAYER_LOGIN
	if (IsLoggedIn()) then -- LOD
		UpdatePlayerItemSets()
	else
		ItemSetFrame:RegisterEvent("PLAYER_LOGIN")
		function ItemSetFrame:PLAYER_LOGIN()
			UpdatePlayerItemSets()
		end
	end
end

--================================--
-- Armor Specialization Detection --
--================================--
--Returns the kind of armor that the class must have equipped in order to get the Armor Specialization bonus
local ArmorSpecClasses = {
	["WARRIOR"] =     L["Plate"],
	["PALADIN"] =     L["Plate"],
	["HUNTER"]  =     L["Mail"],
	["ROGUE"]   =     L["Leather"],
	--["PRIEST"] =    nil, --Priests don't get armor specialization since they can only *wear* cloth
	["DEATHKNIGHT"] = L["Plate"],
	["SHAMAN"] =      L["Mail"],
	--["MAGE"] =      nil, --Mage's don't get armor specialization since they can only *wear* cloth
	--["WARLOCK"] =   nil, --Mage's don't get armor specialization since they can only *wear* cloth
	["MONK"] =        L["Leather"], --20120809: Sure, add Monks. Cloth->Leather
	["DRUID"] =       L["Leather"],
}

-- Create a frame
local ArmorSpecFrame = StatLogic.ArmorSpecFrame;
if not ArmorSpecFrame then
	ArmorSpecFrame = CreateFrame("Frame", "StatLogicItemSetFrame")
	StatLogic.ArmorSpecFrame = ArmorSpecFrame
else
	ArmorSpecFrame:UnregisterAllEvents()
end

-- ArmorSpecActive =
-- nil: You have an empty or non-spec armor slot
-- 0: Appropreate armor in all slots but unspeced
-- 1, 2, 3: Appropreate armor in all slots, then its your GetSpecialization()
local ArmorSpecActive = nil;
local playerArmorType = ArmorSpecClasses[playerClass];
-- Don't set any scripts if cloth class
if playerArmorType then
	-- All 8 slots needs to have something equipped in order to receive the bonus
	local WatchInventoryID = {
		(GetInventorySlotInfo("HeadSlot")),
		(GetInventorySlotInfo("ShoulderSlot")),
		(GetInventorySlotInfo("ChestSlot")),
		(GetInventorySlotInfo("WristSlot")),
		(GetInventorySlotInfo("HandsSlot")),
		(GetInventorySlotInfo("WaistSlot")),
		(GetInventorySlotInfo("LegsSlot")),
		(GetInventorySlotInfo("FeetSlot")),
	}
	local function UpdateArmorSpecActive()
		ArmorSpecActive = nil
		for _, slot in pairs(WatchInventoryID) do
			local item = GetInventoryItemID('player', slot)
			if not item then return end
			if (select(7, GetItemInfo(item))) ~= playerArmorType then return end
		end
		-- all pass
		ArmorSpecActive = GetSpecialization() or 0  --5.0.1  GetPrimaryTalentTree replace with GetSpecialization
		--debugPrint(ArmorSpecActive);
	end
	-- we will schedule this since PLAYER_EQUIPMENT_CHANGED fires multiple times when you switch whole sets
	ArmorSpecFrame:SetScript("OnUpdate", function(self, elapsed)
		if self.updateTime and (GetTime() >= self.updateTime) then
			self.updateTime = nil
			UpdateArmorSpecActive()
		end
	end)
	ArmorSpecFrame:SetScript("OnEvent", function(self, event, ...) if self[event] then return self[event](self, event, ...) end end)
	ArmorSpecFrame:RegisterEvent("PLAYER_EQUIPMENT_CHANGED")
	ArmorSpecFrame:RegisterEvent("PLAYER_TALENT_UPDATE")
	function ArmorSpecFrame:PLAYER_EQUIPMENT_CHANGED()
		self.updateTime = GetTime() + 0.1 -- 0.1 sec delay
	end
	function ArmorSpecFrame:PLAYER_TALENT_UPDATE()
		self.updateTime = GetTime() + 0.1 -- 0.1 sec delay
	end
	-- Initialize on PLAYER_LOGIN
	if (IsLoggedIn()) then -- LOD
		UpdateArmorSpecActive()
	else
		ArmorSpecFrame:RegisterEvent("PLAYER_LOGIN")
		function ArmorSpecFrame:PLAYER_LOGIN()
			UpdateArmorSpecActive()
		end
	end
end

--============--
-- Base Stats --
--============--
--[[
local RaceClassStatBase = {
  -- The Human Spirit - Increase Spirit by 5%
  Human = { --{20, 20, 20, 20, 21}
    WARRIOR = { --{3, 0, 2, 0, 0}
      {23, 20, 22, 20, 22}
    },
    PALADIN = { --{2, 0, 2, 0, 1}
      {22, 20, 22, 20, 23}
    },
    ROGUE = { --{1, 3, 1, 0, 0}
      {21, 23, 21, 20, 22}
    },
    PRIEST = { --{0, 0, 0, 2, 3}
      {20, 20, 20, 22, 25}
    },
    MAGE = { --{0, 0, 0, 3, 2}
      {20, 20, 20, 23, 24}
    },
    WARLOCK = { --{0, 0, 1, 2, 2}
      {20, 20, 21, 22, 24}
    },
  },
  Dwarf = { --{22, 16, 23, 19, 19}
    WARRIOR = {
      {25, 16, 25, 19, 19}
    },
    PALADIN = {
      {24, 16, 25, 19, 20}
    },
    HUNTER = { --{0, 3, 1, 0, 1}
      {22, 19, 24, 19, 20}
    },
    ROGUE = {
      {23, 19, 24, 19, 19}
    },
    PRIEST = {
      {22, 16, 23, 21, 22}
    },
  },
  NightElf = { --{17, 25, 19, 20, 20}
    WARRIOR = {--{3, 0, 2, 0, 0}
      {20, 25, 21, 20, 20}
    },
    HUNTER = {
      {17, 28, 20, 20, 21}
    },
    ROGUE = {
      {18, 28, 20, 20, 20}
    },
    PRIEST = {
      {17, 25, 19, 22, 23}
    },
    DRUID = { --{1, 0, 0, 2, 2}
      {18, 25, 19, 22, 22}
    },
  },
  -- Expansive Mind - Increase Intelligence by 5%
  Gnome = { --{15, 23, 19, 24, 20}
    WARRIOR = {--{3, 0, 2, 0, 0}
      {18, 23, 21, 24, 20}
    },
    ROGUE = {
      {, , , , }
    },
    MAGE = {
      {, , , , }
    },
    WARLOCK = {
      {, , , , }
    },
  },
  Draenei = { --{21, 17, 19, 21, 22}
    WARRIOR = { --{3, 0, 2, 0, 0}
      {24, 17, 21, 21, 22}
    },
    PALADIN = { --{2, 0, 2, 0, 1}
      {23, 17, 21, 21, 23}
    },
    HUNTER = { --{0, 3, 1, 0, 1}
      {21, 20, 20, 21, 23}
    },
    PRIEST = { --{0, 0, 0, 2, 3}
      {21, 17, 19, 23, 25}
    },
    SHAMAN = { --{1, 0, 1, 1, 2}
      {26, 15, 23, 16, 24}
    },
    MAGE = { --{0, 0, 0, 3, 2}
      {21, 17, 19, 24, 24}
    },
  },
  Orc = { --{23, 17, 22, 17, 23}
    WARRIOR = {--{3, 0, 2, 0, 0}
      {26, 17, 24, 17, 23}
    },
    HUNTER = { --{0, 3, 1, 0, 1}
      {23, 20, 23, 17, 24}
    },
    ROGUE = { --{1, 3, 1, 0, 0}
      {, , , , }
    },
    SHAMAN = { --{1, 0, 1, 1, 2}
      {24, 17, 23, 18, 25}
    },
    WARLOCK = { --{0, 0, 1, 2, 2}
      {, , , , }
    },
  },
  Scourge = { --{19, 18, 21, 18, 25}
    WARRIOR = {--{3, 0, 2, 0, 0}
      {22, 18, 23, 18, 25}
    },
    ROGUE = {
      {, , , , }
    },
    PRIEST = {
      {, , , , }
    },
    MAGE = {
      {, , , , }
    },
    WARLOCK = {
      {, , , , }
    },
  },
  Tauren = { --{25, 15, 22, 15, 22}
    WARRIOR = {--{3, 0, 2, 0, 0}
      {28, 15, 24, 15, 22}
    },
    HUNTER = { --{0, 3, 1, 0, 1}
      {, , , , }
    },
    SHAMAN = {
      {, , , , }
    },
    DRUID = { --{1, 0, 0, 2, 2}
      {26, 15, 22, 17, 24}
    },
  },
  Troll = { --{21, 22, 21, 16, 21}
    WARRIOR = {--{3, 0, 2, 0, 0}
      {24, 22, 23, 16, 21}
    },
    HUNTER = { --{0, 3, 1, 0, 1}
      {, , , , }
    },
    ROGUE = {
      {, , , , }
    },
    PRIEST = {
      {, , , , }
    },
    SHAMAN = {
      {, , , , }
    },
    MAGE = {
      {, , , , }
    },
  },
  BloodElf = { --{17, 22, 18, 24, 19}
    PALADIN = {--{2, 0, 2, 0, 1}
      {24, 16, 25, 19, 20}
    },
    HUNTER = { --{0, 3, 1, 0, 1}
      {21, 25, 22, 16, 22}
    },
    ROGUE = {
      {, , , , }
    },
    PRIEST = {
      {, , , , }
    },
    MAGE = {
      {, , , , }
    },
    WARLOCK = {
      {, , , , }
    },
  },
}
--]]
--[[
20120810 There's not even any reference to this variable
local RaceBaseStat = {
	["Human"] = {20, 20, 20, 20, 21},
	["Dwarf"] = {22, 16, 23, 19, 19},
	["NightElf"] = {17, 25, 19, 20, 20},
	["Gnome"] = {15, 23, 19, 24, 20},
	["Draenei"] = {21, 17, 19, 21, 22},
	["Orc"] = {23, 17, 22, 17, 23},
	["Scourge"] = {19, 18, 21, 18, 25},
	["Tauren"] = {25, 15, 22, 15, 22},
	["Troll"] = {21, 22, 21, 16, 21},
	["BloodElf"] = {17, 22, 18, 24, 19},
}
--]]
--[[
20120810: There's not even any reference to this variable
local ClassBonusStat = {
	["DRUID"] = {1, 0, 0, 2, 2},
	["HUNTER"] = {0, 3, 1, 0, 1},
	["MAGE"] = {0, 0, 0, 3, 2},
	["PALADIN"] = {2, 0, 2, 0, 1},
	["PRIEST"] = {0, 0, 0, 2, 3},
	["ROGUE"] = {1, 3, 1, 0, 0},
	["SHAMAN"] = {1, 0, 1, 1, 2},
	["WARLOCK"] = {0, 0, 1, 2, 2},
	["WARRIOR"] = {3, 0, 2, 0, 0},
}
--]]
--[[
20120810: There's not even any references to this variable
local ClassBaseHealth = {
	["WARRIOR"] = 40,
	["PALADIN"] = 38,
	["HUNTER"] = 46,
	["ROGUE"] = 45,
	["PRIEST"] = 52,
	--20120809: Nobody ever added Death Knight. Is this variable even used for anything?
	["SHAMAN"] = 47,
	["MAGE"] = 52,
	["WARLOCK"] = 43,
--  ["MONK"] = 42, --Come up with real monk class base value
	["DRUID"] = 54,
}
--]]
--[[
20120810: There's not even any reference to this variable
local ClassBaseMana = {
	["WARRIOR"] = 0,
	["PALADIN"] = 80,
	["HUNTER"] = 85,
	["ROGUE"] = 0,
	["PRIEST"] = 130,
	--20120809: Nobody ever added Death Knight. Is this variable even used for anything?
	["SHAMAN"] = 75,
	["MAGE"] = 120,
	["WARLOCK"] = 110,
--	["MONK"] = 0, --Come up with real monk class base mana
	["DRUID"] = 70,
}
--]]
--http://wowvault.ign.com/View.php?view=Stats.List&category_select_id=9

--==================================--
-- Stat Mods from Talents and Buffs --
--==================================--
--[[ Aura mods from Thottbot
Apply Aura: Mod Total Stat % (All stats)
Apply Aura: Mod Total Stat % (Strength)
Apply Aura: Mod Total Stat % (Agility)
Apply Aura: Mod Total Stat % (Stamina)
Apply Aura: Mod Total Stat % (Intellect)
Apply Aura: Mod Total Stat % (Spirit)
Apply Aura: Mod Max Health %
Apply Aura: Reduces Attacker Chance to Hit with Melee
Apply Aura: Reduces Attacker Chance to Hit with Ranged
Apply Aura: Reduces Attacker Chance to Hit with Spells (Spells)
Apply Aura: Reduces Attacker Chance to Crit with Melee
Apply Aura: Reduces Attacker Chance to Crit with Ranged
Apply Aura: Reduces Attacker Critical Hit Damage with Melee by %
Apply Aura: Reduces Attacker Critical Hit Damage with Ranged by %
Apply Aura: Mod Dmg % (Spells)
Apply Aura: Mod Dmg % Taken (Fire, Frost)
Apply Aura: Mod Dmg % Taken (Spells)
Apply Aura: Mod Dmg % Taken (All)
Apply Aura: Mod Dmg % Taken (Physical)
Apply Aura: Mod Base Resistance % (Physical)
Apply Aura: Mod Block Percent
Apply Aura: Mod Parry Percent
Apply Aura: Mod Dodge Percent
Apply Aura: Mod Shield Block %
Apply Aura: Mod Detect
Apply Aura: Mod Skill Talent (Defense)
--]]
--[[ StatModAuras, mods not in use are commented out for now
"MOD_STR",
"MOD_AGI",
"MOD_STA",
"MOD_INT",
"MOD_SPI",
"MOD_HEALTH",
"MOD_MANA",
"MOD_ARMOR",
"MOD_BLOCK_VALUE",
--"MOD_DMG", school,
"MOD_DMG_TAKEN", school,
--"MOD_CRIT_DAMAGE", school,
"MOD_CRIT_DAMAGE_TAKEN", school,
--"MOD_THREAT", school,

"ADD_DODGE", -- Used in StatLogic:GetDodgePerAgi()
--"ADD_PARRY",
--"ADD_BLOCK",
--"ADD_STEALTH_DETECT",
--"ADD_STEALTH",
--"ADD_DEFENSE",
--"ADD_THREAT", school,
"ADD_HIT_TAKEN", school,
"ADD_CRIT_TAKEN", school,

--Talents
"ADD_AP_MOD_STA" -- Hunter: Hunter vs. Wild
"ADD_AP_MOD_ARMOR" -- Death Knight: Bladed Armor
"ADD_AP_MOD_INT" -- Shaman: Mental Dexterity
"ADD_PARRY_RATING_MOD_STR" -- Death Knight: Forceful Deflection - Passive
"ADD_PARRY_MOD_STR" --Warrior, Rogue, Death Knight, Paladin
"ADD_SPELL_CRIT_RATING_MOD_SPI" -- Mage: Molten Armor - 3.1.0
"ADD_COMBAT_MANA_REGEN_MOD_INT"
"ADD_RANGED_AP_MOD_INT"
"ADD_ARMOR_MOD_INT"
"ADD_SPELL_DMG_MOD_STR" -- Paladin: Touched by the Light
"ADD_SPELL_DMG_MOD_STA"
"ADD_SPELL_DMG_MOD_INT"
"ADD_SPELL_DMG_MOD_SPI"
"ADD_SPELL_DMG_MOD_AP" -- Shaman: Mental Quickness, Paladin: Sheath of Light
"ADD_HEAL_MOD_STR" -- Paladin: Touched by the Light
"ADD_HEAL_MOD_AGI"
"ADD_HEAL_MOD_STA"
"ADD_HEAL_MOD_INT"
"ADD_HEAL_MOD_SPI"
"ADD_HEAL_MOD_AP" -- Shaman: Mental Quickness
"ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN"
"MOD_AP"
"MOD_RANGED_AP"
"MOD_SPELL_DMG"
"MOD_HEAL"

--"ADD_CAST_TIME"
--"MOD_CAST_TIME"
--"ADD_MANA_COST"
--"MOD_MANA_COST"
--"ADD_RAGE_COST"
--"MOD_RAGE_COST"
--"ADD_ENERGY_COST"
--"MOD_ENERGY_COST"
--"ADD_COOLDOWN"
--"MOD_COOLDOWN"
--]]

local StatModInfo = {
  ------------------------------------------------------------------------------
  -- initialValue: sets the initial value for the stat mod
  -- if initialValue == 0, inter-mod operations are done with addition,
  -- if initialValue == 1, inter-mod operations are done with multiplication,
  ------------------------------------------------------------------------------
  -- finalAdjust: added to the final result before returning,
  -- so we can adjust the return value to be used in addition or multiplication
  -- for addition: initialValue + finalAdjust = 0
  -- for multiplication: initialValue + finalAdjust = 1
  ------------------------------------------------------------------------------
  -- school: school arg is required for these mods
  ------------------------------------------------------------------------------
	["ADD_CRIT_TAKEN"] = {
		initialValue = 0,
		finalAdjust = 0,
		school = true,
	},
	["ADD_HIT_TAKEN"] = {
		initialValue = 0,
		finalAdjust = 0,
		school = true,
	},
	["ADD_DODGE"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_BLOCK_REDUCTION"] = { -- Meta Gems
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_AP_MOD_INT"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_AP_MOD_STA"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_AP_MOD_ARMOR"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_PARRY_RATING_MOD_STR"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_PARRY_MOD_STR"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_SPELL_CRIT_RATING_MOD_SPI"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_SPELL_HIT_RATING_MOD_SPI"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_RANGED_AP_MOD_INT"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_ARMOR_MOD_INT"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_SPELL_DMG_MOD_AP"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_SPELL_DMG_MOD_STR"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_SPELL_DMG_MOD_STA"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_SPELL_DMG_MOD_INT"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_SPELL_DMG_MOD_SPI"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_SPELL_DMG_MOD_PET_STA"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_SPELL_DMG_MOD_PET_INT"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_SPELL_DMG_MOD_MANA"] = { -- Improved Mana Gem
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_HEAL_MOD_AP"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_HEAL_MOD_STR"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_HEAL_MOD_AGI"] = { -- Nurturing Instinct
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_HEAL_MOD_STA"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_HEAL_MOD_INT"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_HEAL_MOD_SPI"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_HEAL_MOD_MANA"] = { -- Improved Mana Gem
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_COMBAT_MANA_REGEN_MOD_INT"] = { -- deprecated
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_COMBAT_MANA_REGEN_MOD_MANA"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN"] = {
		initialValue = 0,
		finalAdjust = 0,
	},
	["ADD_PET_STA_MOD_STA"] = {
		initialValue = 1,
		finalAdjust = -1,
	},
	["ADD_PET_INT_MOD_INT"] = {
		initialValue = 1,
		finalAdjust = -1,
	},
	["MOD_CRIT_DAMAGE_TAKEN"] = {
		initialValue = 0,
		finalAdjust = 1,
		school = true,
	},
	["MOD_DMG_TAKEN"] = {
		initialValue = 0,
		finalAdjust = 1,
		school = true,
	},
	["MOD_CRIT_DAMAGE"] = {
		initialValue = 0,
		finalAdjust = 1,
		school = true,
	},
	["MOD_DMG"] = {
		initialValue = 0,
		finalAdjust = 1,
		school = true,
	},
	["MOD_ARMOR"] = {
		initialValue = 1,
		finalAdjust = 0,
	},
	["MOD_HEALTH"] = {
		initialValue = 1,
		finalAdjust = 0,
	},
	["MOD_MANA"] = {
		initialValue = 1,
		finalAdjust = 0,
	},
	["MOD_STR"] = {
		initialValue = 1,
		finalAdjust = 0,
	},
	["MOD_AGI"] = {
		initialValue = 1,
		finalAdjust = 0,
	},
	["MOD_STA"] = {
		initialValue = 1,
		finalAdjust = 0,
	},
	["MOD_INT"] = {
		initialValue = 1,
		finalAdjust = 0,
	},
	["MOD_SPI"] = {
		initialValue = 1,
		finalAdjust = 0,
	},
	["MOD_BLOCK_VALUE"] = {
		initialValue = 0,
		finalAdjust = 1,
	},
	["MOD_AP"] = {
		initialValue = 0,
		finalAdjust = 1,
	},
	["MOD_RANGED_AP"] = {
		initialValue = 0,
		finalAdjust = 1,
	},
	["MOD_SPELL_DMG"] = {
		initialValue = 0,
		finalAdjust = 1,
	},
	["MOD_HEAL"] = {
		initialValue = 0,
		finalAdjust = 1,
	},
	["MOD_MELEE_HASTE"] = {
		initialValue = 1,
		finalAdjust = 0,
	},
	["MOD_SPELL_HASTE"] = {
		initialValue = 1,
		finalAdjust = 0,
	},
}
StatLogic.StatModInfo = StatModInfo -- so other addons can use this directly

StatLogic.SpellSchools = {
	"MELEE",
	"RANGED",
	"HOLY",
	"FIRE",
	"NATURE",
	"FROST",
	"SHADOW",
	"ARCANE",
}

------------------
-- StatModTable --
------------------
--[[ How to add a glyph support?
1. Go to the glyph item on wowhead.
2. Click on the green Use: TEXT on the "tooltip" to go to a spell page.
3. Click on "See also" tab for a spell with a gear icon, the spell id for this page is what you put in here.
--]]
--[[ How can I get aura data?
1. /dump UnitAura("player",1)
--]]
local StatModTable = {}
StatLogic.StatModTable = StatModTable -- so other addons can use this directly
if playerClass == "DRUID" then
	StatModTable["DRUID"] = {
		-- Druid: Nurturing Instinct (Rank 2) - 2,14
		-- 4.0.1: 2,14: Increases your healing spells by up to 35%/70% of your Agility, and increases healing done to you by 10%/20% while in Cat form.
		-- 4.0.6: 2,14: Increases your healing spells by up to 50%/100% of your Agility, and increases healing done to you by 10%/20% while in Cat form.
		-- 5.0.4: TODO: There are no more talents; removing tab/num.
		-- 5.0.4: Nurturing Instinct (SpellID 33873) Increases your nature spell power by 100% of your Agility.
		["ADD_HEAL_MOD_AGI"] = {
			{
				["spellid"] = 33873,
				["rank"] = { 1, },
				--["tab"] = 2,
				--["num"] = 14,
			},
		},
		-- Healers: Meditation
		-- 4.0.1: Allows 50% of your mana regeneration from Spirit to continue while in combat.
		-- 5.0.4: Druid specialization SpellID 85101
		["ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN"] = {
			{
				["known"] = 85101, --Meditation: Druid Specialization Talent
				["rank"] = {0.50, },
			},
			{
				["known"] = 95862, --Meditation: Shaman Specialization Talent 
				["rank"] = {0.50, },
			},
		},
		-- Druid: Feral Swiftness (Rank 2) - 2,1
		-- 4.0.1: Increases your chance to dodge while in Cat Form or Bear Form by 2/4%
		-- Druid: Natural Reaction (Rank 2) - 2,18
		-- 4.0.1: Reduces damage taken while in Bear Form by 6/12%, increases your dodge while in Bear Form by 3/6%, and you generate 1/3 rage every time you dodge while in Bear Form.
		-- 5.0.4: TODO: There are no more talents; removing tab/num.
		["ADD_DODGE"] = {
			{-- Feral Swiftness
				--["tab"] = 2,
				--["num"] = 1,
				["rank"] = { 2, 4, },
				["buff"] = 5487,        -- ["Bear Form"],
			},
			{-- Feral Swiftness
				--["tab"] = 2,
				--["num"] = 1,
				["rank"] = { 2, 4, },
				["buff"] = 768,        -- ["Cat Form"],
			},
			{-- Natural Reaction
				--["tab"] = 2,
				--["num"] = 18,
				["rank"] = { 3, 6, },
				["buff"] = 5487,        -- ["Bear Form"],
			},
		},
		-- Druid: Glyph of Barkskin
		-- 4.0.1: Reduces the chance you'll be critically hit by melee attacks by 25% while Barkskin is active.
		-- Druid: Thick Hide (Rank 3) - 2,11
		-- 4.0.1: Increases your Armor contribution from cloth and leather items by 4/7/10%, increases armor while in Bear Form by an additional 11/22/33%, and reduces the chance you'll be critically hit by melee attacks by 2/4/6%.
		-- Druid: Survival of the Fittest (Rank 3) - 2,18
		-- 4.0.1: Increases all attributes by 2%/4%/6% and reduces the chance you'll be critically hit by melee attacks by 2%/4%/6%.
		-- 5.0.4: TODO: There are no more talents; removing tab/num.
		["ADD_CRIT_TAKEN"] = {
			{-- Glyph of Barkskin
				["MELEE"] = true,
				["rank"] = { -0.25, },
				["buff"] = 22812,        -- ["Barkskin"],
				["glyph"] = 63057,
			},
			{-- Thick Hide
				["MELEE"] = true,
				--["tab"] = 2,
				--["num"] = 11,
				["rank"] = { -0.02, -0.04, -0.06, },
			},
		},
		-- Druid: Balance of Power (Rank 2) - 1,6
		-- 4.0.1: Increases your spell hit rating by an additional amount equal to 50/100% of your Spirit.
		-- 5.0.4: TODO: There are no more talents; removing tab/num.
		["ADD_SPELL_HIT_RATING_MOD_SPI"] = {
		  {
			["spellid"] = 33596,
			--["tab"] = 1,
			--["num"] = 6,
			["rank"] = {  0.5, 1, },
		  },
		},
		-- Druid: Enrage - Buff: 5229
		-- 4.0.1: Physical damage taken increased by 10%.
		-- Druid: Perseverance - Rank 3/3 - Talent: 3,5
		-- 4.0.1: Reduces all spell damage taken by 2/4/6%.
		-- Druid: Barkskin - Buff
		-- 4.0.1: All damage taken is reduced by 20%.
		-- Druid: Improved Barkskin (Rank 2) - 3,25
		-- 4.0.1: Increases the damage reduction granted by your Barkskin spell by 5/10%
		-- Druid: Natural Reaction (Rank 2) - 2,18
		-- 4.0.1: Reduces damage taken while in Bear Form by 6/12%, increases your dodge while in Bear Form by 3/6%, and you generate 1/3 rage every time you dodge while in Bear Form.
		-- Druid: Survival Instincts - Buff
		-- 4.0.1: Damage taken reduced by 60% while in Bear Form or Cat Form.
		-- 4.0.6: Damage taken reduced by 50% while in Bear Form or Cat Form.
		-- Druid: Moonkin Form - Buff
		-- 4.0.1: Armor contribution from items is increased by 120%.
		-- 4.0.6: All damage reduced by 15%.
		-- 5.0.4: TODO: There are no more talents; removing tab/num.
		["MOD_DMG_TAKEN"] = {
		  {-- Enrage
			["MELEE"] = true,
			["RANGED"] = true,
			["rank"] = { 0.1, },
			["buff"] = 5229,        -- ["Enrage"],
		  },
		  {-- Perseverance
			["HOLY"] = true,
			["FIRE"] = true,
			["NATURE"] = true,
			["FROST"] = true,
			["SHADOW"] = true,
			["ARCANE"] = true,
			--["tab"] = 3,
			--["num"] = 5,
			["rank"] = { -0.02, -0.04, -0.06, },
		  },
		  {-- Natural Reaction
			["MELEE"] = true,
			["RANGED"] = true,
			["HOLY"] = true,
			["FIRE"] = true,
			["NATURE"] = true,
			["FROST"] = true,
			["SHADOW"] = true,
			["ARCANE"] = true,
			--["tab"] = 2,
			--["num"] = 18,
			["rank"] = { -0.06, -0.12, },
			["buff"] = 5487,        -- ["Bear Form"],
		  },
		  {-- Barkskin
			["MELEE"] = true,
			["RANGED"] = true,
			["HOLY"] = true,
			["FIRE"] = true,
			["NATURE"] = true,
			["FROST"] = true,
			["SHADOW"] = true,
			["ARCANE"] = true,
			["rank"] = { -0.2, },
			["buff"] = 22812,        -- ["Barkskin"],
		  },
		  {-- Survival Instincts - Bear
			["MELEE"] = true,
			["RANGED"] = true,
			["HOLY"] = true,
			["FIRE"] = true,
			["NATURE"] = true,
			["FROST"] = true,
			["SHADOW"] = true,
			["ARCANE"] = true,
			["rank"] = {
			  -0.5,
			},
			["buff"] = 61336,        -- ["Survival Instincts"],
			["buff2"] = 5487,        -- ["Bear Form"],
		  },
		  {-- Survival Instincts - Cat
			["MELEE"] = true,
			["RANGED"] = true,
			["HOLY"] = true,
			["FIRE"] = true,
			["NATURE"] = true,
			["FROST"] = true,
			["SHADOW"] = true,
			["ARCANE"] = true,
			["rank"] = {
			  -0.5,
			},
			["buff"] = 61336,        -- ["Survival Instincts"],
			["buff2"] = 768,        -- ["Cat Form"],
		  },
		  {-- Moonkin Form - Buff
			["MELEE"] = true,
			["RANGED"] = true,
			["HOLY"] = true,
			["FIRE"] = true,
			["NATURE"] = true,
			["FROST"] = true,
			["SHADOW"] = true,
			["ARCANE"] = true,
			["rank"] = {
			  -0.15,
			},
			["buff"] = 24858,        -- ["Moonkin Form"],
		  },
		},
		-- Druid: Thick Hide (Rank 3) - 2,11
		-- 4.0.1: Increases your Armor contribution from cloth and leather items by 4/7/10%, increases armor while in Bear Form by an additional 11/22/33%, and reduces the chance you'll be critically hit by melee attacks by 2/4/6%.
		-- Druid: Bear Form - Buff
		-- 4.0.1: Increases melee attack power by 30, armor contribution from cloth and leather items by 65%/120%, and Stamina by 25%.  Also causes agility to increase attack power.
		-- Druid: Tree of Life - Buff
		-- 4.0.1: Armor increased by 120%.
		-- 5.0.4: TODO: There are no more talents; removing tab/num.
		["MOD_ARMOR"] = {
		  {-- Thick Hide
			--["tab"] = 2,
			--["num"] = 11,
			["rank"] = { 0.04, 0.07, 0.1, },
		  },
		  {-- Thick Hide
			--["tab"] = 2,
			--["num"] = 11,
			["rank"] = { 0.11, 0.22, 0.33, },
			["buff"] = 5487,        -- ["Bear Form"],
		  },
		  {-- Bear Form < lv40
			["rank"] = { 0.65, },
			["buff"] = 5487,        -- ["Bear Form"],
		  },
		  {-- Bear Form >= lv40
			["rank"] = { 
			  0.333333333333333333, -- 1.65 * 1.3333 = 2.2
			},
			["buff"] = 5487,        -- ["Bear Form"],
			["condition"] = "UnitLevel('player') >= 40",
		  },
		  {
			["rank"] = {
			  1.2,
			},
			["buff"] = 33891,        -- ["Tree of Life"],
		  },
		},
		-- Druid: Furor (Rank 3) - 2,2
		-- 4.0.1: Increases your maximum mana by 5/10/15%.
		-- 5.0.4: TODO: There are no more talents; removing tab/num.
		["MOD_MANA"] = {
		  {
			--["tab"] = 2,
			--["num"] = 2,
			["rank"] = { 0.05, 0.1, 0.15, },
		  },
		},
		-- Druid: Leather Specialization - Passive: 86530
		-- 4.0.1: Increases your primary attribute by 5% while wearing Leather in all armor slots.  Balance specialization grants Intellect, Feral specialization grants Agility while in Cat Form and Stamina while in Bear Form, and Restoration specialization grants Intellect.
		-- Druid: Heart of the Wild (Rank 5) - 3,4
		-- 4.0.1: Increases your Intellect by 2/4/6%. In addition, while in Bear Form your Stamina is increased by 3/7/10% and while in Cat Form your attack power is increased by 3/7/10%.
		-- Druid: Bear Form - Buff
		-- 4.0.1: Increases melee attack power by 30, armor contribution from cloth and leather items by 65%/120%, and Stamina by 25%.  Also causes agility to increase attack power.
		-- 5.0.4: TODO: There are no more talents; removing tab/num.
		["MOD_STA"] = {
		  {-- Heart of the Wild: +2/4/6/8/10% stamina in bear / dire bear
			--["tab"] = 3,
			--["num"] = 4,
			["rank"] = { 0.03, 0.07, 0.1, },
			["buff"] = 5487,        -- ["Bear Form"],
		  },
		  {-- Bear Form
			["rank"] = {
			  0.25,
			},
			["buff"] = 5487,        -- ["Bear Form"],
		  },
		  {-- Leather Specialization
			["rank"] = {
			  0.05,
			},
			["armorspec"] = 2,
			["known"] = 86530,
			["buff"] = 5487,        -- ["Bear Form"],
		  },
		},
		-- Druid: Leather Specialization - Passive: 86530
		-- 4.0.1: Increases your primary attribute by 5% while wearing Leather in all armor slots.  Balance specialization grants Intellect, Feral specialization grants Agility while in Cat Form and Stamina while in Bear Form, and Restoration specialization grants Intellect.
		["MOD_AGI"] = {
		  {-- Leather Specialization
			["rank"] = {
			  0.05,
			},
			["armorspec"] = 2,
			["known"] = 86530,
			["buff"] = 768,        -- ["Cat Form"],
		  },
		},
		-- Druid: Heart of the Wild (Rank 5) - 3,4
		-- 4.0.1: Increases your Intellect by 2/4/6%. In addition, while in Bear Form your Stamina is increased by 3/7/10% and while in Cat Form your attack power is increased by 3/7/10%.
		-- Druid: Aggression - Passive: 84735
		-- 4.0.1: Increases your attack power by 25%.
		-- 5.0.4: TODO: There are no more talents; removing tab/num.
		["MOD_AP"] = {
		  {
			--["tab"] = 3,
			--["num"] = 4,
			["rank"] = { 0.03, 0.07, 0.1, },
			["buff"] = 768,        -- ["Cat Form"],
		  },
		  {
			["rank"] = { 0.25, },
			["known"] = 84735, -- ["Aggression"]
		  },
		},
		-- Druid: Leather Specialization - Passive: 86530
		-- 4.0.1: Increases your primary attribute by 5% while wearing Leather in all armor slots.  Balance specialization grants Intellect, Feral specialization grants Agility while in Cat Form and Stamina while in Bear Form, and Restoration specialization grants Intellect.
		-- Druid: Heart of the Wild (Rank 5) - 3,4
		-- 4.0.1: Increases your Intellect by 2/4/6%. In addition, while in Bear Form your Stamina is increased by 3/7/10% and while in Cat Form your attack power is increased by 3/7/10%.
		-- 5.0.4: TODO: There are no more talents; removing tab/num.
	["MOD_INT"] = {
		  {
			--["tab"] = 3,
			--["num"] = 4,
			["rank"] = { 0.02, 0.04, 0.06, },
		  },
		  {-- Leather Specialization
			["rank"] = {
			  0.05,
			},
			["known"] = 86530,
			["armorspec"] = 1,
		  },
		  {-- Leather Specialization
			["rank"] = {
			  0.05,
			},
			["known"] = 86530,
			["armorspec"] = 3,
		  },
		},
}
elseif playerClass == "DEATHKNIGHT" then
	StatModTable["DEATHKNIGHT"] = {
    -- Death Knight: Icy Talons - Passive: 50887
    -- 4.0.6: Your melee attack speed is increased by 20%.
    -- Death Knight: Improved Icy Talons - Rank 1/1 - 2,13
    -- 4.0.6: Increases the melee and ranged attack speed of all party and raid 
    --        members within 100 yards by 10%, and your own attack speed by an additional 5%.
    -- Death Knight: Unholy Presence - Stance
    -- 4.0.6: Attack speed and rune regeneration increased 10%.
    -- Death Knight: Improved Unholy Presence - Rank 2/2 - 3,16
    -- 4.0.6: Grants you an additional 2/5% haste while in Unholy Presence. 
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_MELEE_HASTE"] = {
      {-- Icy Talons
        ["rank"] = {
          0.2,
        },
        ["known"] = 50887,
      },
      {-- Improved Icy Talons
        --["tab"] = 2,
        --["num"] = 13,
        ["rank"] = { 0.05, },
      },
      {-- Unholy Presence and Improved Unholy Presence
        --["tab"] = 3,
        --["num"] = 16,
        ["rank"] = { 
			[0] = 0.1, 0.12, 0.15,
        },
        ["stance"] = "Interface\\Icons\\Spell_Deathknight_UnholyPresence",
      },
    },
    -- Tanks: Forceful Deflection - Passive
	-- Apply Aura: Mod Combat Rating by % of Stat (8) Value: 27
    -- 4.0.1: Increases your Parry Rating by 25% of your total Strength.
    -- 4.2.0: Increases your Parry Rating by 27% of your total Strength.
	-- 5.0.4: Forceful deflection spell is invisible, but strength still contributes to Parry 
	--        But Strength contributes to Parry%, not Parry Rating. 
	--        Bonus strength also suffers diminishing returns.
	--[[
    ["ADD_PARRY_RATING_MOD_STR"] = {
		{
			--Strength doesn't truly affect Parry Rating (parry rating stays constant if you simply increase strength)
			--rather it affects Parry Chance (%). But Strength *does* seem to suffer diminishing returns,
			--as though it were going through its own Str -> CR_PARRY -> ParryChance conversion.
			--For simplicity, it's easy to think of Strength as affecting Pary Rating.
			--Finally: there is no longer any visible talent, spell, ability or passive that says that strength will
			--affect your Parry Chance. So i'll make the (easy) assumption that *everyone* get's parry from strength.
			--And nobody besides tanks kill care; but why would they?
			["rank"] = { 1.0886011033071300, },
			["new"] = 15929,
		},
		{
			["spellid"] = 49410,
			["rank"] = { 0.27, },
			["old"] = 15929, -- 4.2.0
		},
		{
			["spellid"] = 49410,
			["rank"] = { 0.25, },
			["old"] = 14333, -- 4.2.0
		},
    },
	--]]
	-- 5.0.4: Strength adds to Parry %, with bonus strength suffering diminishing returns
	["ADD_PARRY_MOD_STR"] = {
		{
			["known"] = 82246, --SpellID 82246: Parry (Death Knight)
			["rank"] = {1}, 
		},
	},
    -- Death Knight: Bladed Armor - Rank 3/3 - 1,3
    -- 4.0.1: Increases your attack power by 2/4/6 for every 180 armor value you have.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_AP_MOD_ARMOR"] = {
      {
        ["spellid"] = 49391,
        --["tab"] = 1,
        --["num"] = 3,
        ["rank"] = { 2/180, 4/180, 6/180, },
      },
    },
    -- Death Knight: Improved Blood Presence - Rank 2/2 - Talent: 1,14 - Stance
    -- 4.0.1: Reduces the chance that you will be critically hit by melee attacks while in Blood Presence by 3/6%. In addition, while in Frost Presence or Unholy Presence, you retain 2/4% damage reduction from Blood Presence.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_CRIT_TAKEN"] = {
      {-- Improved Blood Presence
        ["MELEE"] = true,
        --["tab"] = 1,
        --["num"] = 14,
        ["rank"] = { -0.03, -0.06, },
        ["stance"] = "Interface\\Icons\\Spell_Deathknight_BloodPresence",
      },
    },
    -- Death Knight: Blade Barrier - Rank 3/3 - Talent: 1,2 - Buff: 64856
    -- 4.0.1: 2/4/6% less damage taken.
    -- Death Knight: Icebound Fortitude - Buff: 48792
    -- 4.0.1: Damage taken reduced by 30%.
    -- 4.0.3: Damage taken reduced by 20%.
    -- Death Knight: Sanguine Fortitude - Rank 2/2 - 1,12 - Buff: 48792
    -- 4.0.1: Icebound Fortitude reduces damage taken by an additional 15/30%
    -- Death Knight: Bone Shield - Buff: 49222
    -- 4.0.1: Damage reduced by 20%.
    -- Death Knight: Anti-Magic Shell - Buff: 48707
    -- 4.0.1: Spell damage reduced by 75%.
    -- Death Knight: Blood Presence - Stance
    -- 4.0.1: Stamina increased by 8%.
    --        Armor contribution from cloth, leather, mail and plate items increased by 60%.
    --        Damage taken reduced by 8%.
    -- Death Knight: Improved Blood Presence - Rank 2/2 - Talent: 1,14 - Stance
    -- 4.0.1: Reduces the chance that you will be critically hit by melee attacks while in Blood Presence by 3/6%. In addition, while in Frost Presence or Unholy Presence, you retain 2/4% damage reduction from Blood Presence.
    -- Death Knight: Will of the Necropolis - Rank 3/3 - 1,15 - Buff: 81162
    -- 4.0.1: Damage taken reduced by 8/16/25%
    -- Enchant: Rune of Spellshattering - EnchantID: 3367
    -- 4.0.1: Deflects 4% of all spell damage to 2h weapon
    -- Enchant: Rune of Spellbreaking - EnchantID: 3595
    -- 4.0.1: Deflects 2% of all spell damage to 1h weapon
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_DMG_TAKEN"] = {
      {-- Blade Barrier
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 1,
        --["num"] = 2,
        ["rank"] = { -0.02, -0.04, -0.06,
        },
        ["buff"] = 64856,        -- ["Blade Barrier"],
      },
      {-- Icebound Fortitude
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.20,
        },
        ["buff"] = 48792,        -- ["Icebound Fortitude"],
      },
      {-- Sanguine Fortitude
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 1,
        --["num"] = 12,
        ["rank"] = { -0.15, -0.30, },
        ["buff"] = 48792,        -- ["Icebound Fortitude"],
      },
      {-- Bone Shield
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.20,
        },
        ["buff"] = 49222,        -- ["Bone Shield"],
      },
      {-- Anti-Magic Shell
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.75,
        },
        ["buff"] = 48707,        -- ["Anti-Magic Shell"],
      },
      {-- Blood Presence
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.08,
        },
        ["stance"] = "Interface\\Icons\\Spell_Deathknight_BloodPresence",
      },
      {-- Improved Blood Presence
        --["tab"] = 1,
        --["num"] = 14,
        ["rank"] = { -0.02, -0.04,},
        ["stance"] = "Interface\\Icons\\Spell_Deathknight_FrostPresence",
      },
      {-- Improved Blood Presence
        --["tab"] = 1,
        --["num"] = 14,
        ["rank"] = {
          -0.02, -0.04,
        },
        ["stance"] = "Interface\\Icons\\Spell_Deathknight_UnholyPresence",
      },
      {-- Will of the Necropolis
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 1,
        --["num"] = 15,
        ["rank"] = {
          -0.08, -0.16, -0.25,
        },
        ["buff"] = 81162,        -- ["Will of the Necropolis"],
      },
      {-- Rune of Spellshattering
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.04,
        },
        ["slot"] = 16, -- main hand slot
        ["enchant"] = 3367,
      },
      {-- Rune of Spellbreaking
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.02,
        },
        ["slot"] = 16, -- main hand slot
        ["enchant"] = 3595,
      },
      {-- Rune of Spellbreaking
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.02,
        },
        ["slot"] = 17, -- off hand slot
        ["enchant"] = 3595,
      },
    },
    -- Death Knight: Toughness - Rank 3/3 - 1,10
    -- 4.0.1: Increases your armor value from items by 3/7/10%.
    -- Death Knight: Blood Presence - Stance
    -- 4.0.1: Stamina increased by 8%.
    --        Armor contribution from cloth, leather, mail and plate items increased by 60%.
    --        Damage taken reduced by 8%.
    -- Enchant: Rune of the Stoneskin Gargoyle - EnchantID: 3847
    -- 4.0.1: +4% Armor and +2% Stamina to 2h weapon
    -- Enchant: Rune of the Nerubian Carapace - EnchantID: 3883
    -- 4.0.1: +2% Armor and +1% Stamina to 1h weapon
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_ARMOR"] = {
      {-- Blood Presence
        ["rank"] = {
          0.6,
        },
        ["stance"] = "Interface\\Icons\\Spell_Deathknight_BloodPresence",
      },
      {-- Toughness
        --["tab"] = 1,
        --["num"] = 10,
        ["rank"] = {
          0.03, 0.07, 0.1,
        },
      },
    },
    -- Death Knight: Vampiric Blood - Buff: 55233
    -- 4.0.1: Maximum health increased by 15%
    -- Death Knight: Glyph of Vampiric Blood - Glyph: 58676
    -- 4.0.1: Vampiric Blood no longer grants you health
    ["MOD_HEALTH"] = {
      {-- Vampiric Blood
        ["rank"] = {
          0.15,
        },
        ["buff"] = 55233,        -- ["Vampiric Blood"],
        ["condition"] = "not LibStub('LibStatLogic-1.2').PlayerHasGlyph(58676)", -- ["Glyph of Vampiric Blood"]
      },
    },
    -- Death Knight: Plate Specialization - Passive: 86524
    -- 4.0.1: Increases your primary attribute by 5% while wearing Plate in all armor slots. Blood specialization grants Stamina, Frost specialization grants Strength, and Unholy specialization grants Strength.
    -- Death Knight: Blood Presence - Stance
    -- 4.0.1: Stamina increased by 8%.
    --        Armor contribution from cloth, leather, mail and plate items increased by 60%.
    --        Damage taken reduced by 8%.
    -- Death Knight: Veteran of the Third War - Passive: 50029
    -- 4.0.1: Increases your total Stamina by 9% and your expertise by 6.
    -- Enchant: Rune of the Stoneskin Gargoyle - EnchantID: 3847
    -- 4.0.1: +4% Armor and +2% Stamina to 2h weapon
    -- Enchant: Rune of the Nerubian Carapace - EnchantID: 3883
    -- 4.0.1: +2% Armor and +1% Stamina to 1h weapon
    ["MOD_STA"] = {
      {-- Plate Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86524,
        ["armorspec"] = 1,
      },
      {-- Blood Presence
        ["rank"] = {
          0.08,
        },
        ["stance"] = "Interface\\Icons\\Spell_Deathknight_BloodPresence",
      },
      {-- Veteran of the Third War
        ["rank"] = {
          0.09,
        },
        ["known"] = 50029, -- ["Veteran of the Third War"]
      },
      {-- Rune of the Stoneskin Gargoyle
        ["rank"] = {
          0.02,
        },
        ["slot"] = 16, -- main hand slot
        ["enchant"] = 3847,
      },
      {-- Rune of the Nerubian Carapace
        ["rank"] = {
          0.01,
        },
        ["slot"] = 16, -- main hand slot
        ["enchant"] = 3883,
      },
      {-- Rune of the Nerubian Carapace
        ["rank"] = {
          0.01,
        },
        ["slot"] = 17, -- off hand slot
        ["enchant"] = 3883,
      },
    },
    -- Death Knight: Plate Specialization - Passive: 86524
    -- 4.0.1: Increases your primary attribute by 5% while wearing Plate in all armor slots. Blood specialization grants Stamina, Frost specialization grants Strength, and Unholy specialization grants Strength.
    -- Death Knight: Unholy Might - Passive: 91107
    -- 4.0.1: Dark power courses through your limbs, increasing your Strength by 15%.
    -- 4.0.6: Dark power courses through your limbs, increasing your Strength by 5%.
    -- 4.2.0: Dark power courses through your limbs, increasing your Strength by 20%.
    -- Death Knight: Abomination's Might - 1,11
    -- 4.0.1: Also increases your total Strength by 1%/2%.
    -- Death Knight: Pillar of Frost - Buff: 51271
    -- 4.0.1: Strength increased by 20%.
    -- Death Knight: Brittle Bones - Rank 2/2 - 2,14
    -- 4.0.1: Your Strength is increased by 2/4%
    -- Death Knight: Unholy Strength - Buff: 53365
    -- 4.0.1: Strength increased by 15%
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_STR"] = {
      {-- Plate Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86524,
        ["armorspec"] = 2,
      },
      {-- Plate Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86524,
        ["armorspec"] = 3,
      },
      {-- Unholy Might
        ["rank"] = {
          0.05,
        },
        ["known"] = 91107, -- ["Unholy Might"]
        ["old"] = 14333, -- 4.2.0
      },
      {-- Unholy Might
        ["rank"] = {
          0.2,
        },
        ["known"] = 91107, -- ["Unholy Might"]
        ["new"] = 14333, -- 4.2.0
      },
      {-- Abomination's Might
        --["tab"] = 1,
        --["num"] = 11,
        ["rank"] = {
          0.01, 0.02,
        },
      },
      {-- Pillar of Frost
        ["rank"] = {
          0.2,
        },
        ["buff"] = 51271,        -- ["Pillar of Frost"],
      },
      {-- Brittle Bones
        --["tab"] = 2,
        --["num"] = 14,
        ["rank"] = {
          0.02, 0.04,
        },
      },
      {-- Unholy Strength
        ["rank"] = {
          0.15,
        },
        ["buff"] = 53365,        -- ["Unholy Strength"],
      },
    },
  }
elseif playerClass == "HUNTER" then
	StatModTable["HUNTER"] = {
		-- Hunter: Deterrence - Buff: 19263
		-- 4.0.6: Chance to be missed by melee attacks increased by 100%, chance for ranged attacks to miss you increased by 100%, 100% chance to deflect spells.
		["ADD_HIT_TAKEN"] = {
			{-- Deterrence
				["MELEE"] = true,
				["RANGED"] = true,
				["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -1,
        },
        ["buff"] = 19263,
      },
    },
    -- Hunter: Animal Handler - Passive: 87325
    -- 4.0.1: Attack Power increased by 15%.
    -- 4.0.6: Attack Power increased by 25%.
    ["MOD_AP"] = {
      {-- Animal Handler
        ["rank"] = {
          0.25,
        },
        ["known"] = 87325, -- ["Animal Handler"]
      },
    },
    -- Hunter: Hunter vs. Wild - Rank 3/3 - 3,1
    -- 4.0.1: Increases your total Stamina by 4/7/10%.
    -- 4.0.1: Increases your total Stamina by 5/10/15%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_STA"] = {
      {-- Hunter vs. Wild
        --["tab"] = 3,
        --["num"] = 1,
        ["rank"] = {
          0.05, 0.1, 0.15,
        },
      },
    },
    -- Hunter: Mail Specialization - Passive: 86528
    -- 4.0.1: Increases your Agility by 5% while wearing Mail in all armor slots
    -- Hunter: Into the Wilderness - Passive: 84729
    -- 4.0.1: Agility increased by 10%.
    -- Hunter: Hunting Party - Rank 1/1 - 3,17
    -- 4.0.1: Increases your total Agility by an additional 10%
    -- 4.0.6: Increases your total Agility by an additional 2%
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_AGI"] = {
      {-- Mail Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86528,
        ["armorspec"] = 0,
      },
      {-- Mail Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86528,
        ["armorspec"] = 1,
      },
      {-- Mail Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86528,
        ["armorspec"] = 2,
      },
      {-- Mail Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86528,
        ["armorspec"] = 3,
      },
      {-- Into the Wilderness
        ["rank"] = {
          0.1,
        },
        ["known"] = 84729, -- ["Into the Wilderness"]
      },
      {-- Hunting Party
        --["tab"] = 3,
        --["num"] = 17,
        ["rank"] = {
          0.02,
        },
      },
    },
  }
elseif playerClass == "MAGE" then
	StatModTable["MAGE"] = {
    -- Mage: Wizardry - Passive: 89744
    -- 4.0.1: Increases your Intellect by 5%
    ["MOD_INT"] = {
      {-- Wizardry
        ["rank"] = {
          0.05,
        },
        ["known"] = 89744, -- ["Wizardry"]
      },
    },
    -- Mage: Molten Armor - Buff: 30482
    -- 4.0.1: Reduces the chance you are critically hit by 5%.
    -- Mage: Firestarter - Rank 1/1 - 2,15
    -- 4.0.1: Your Molten Armor allows you to cast the Scorch spell while moving instead of reducing the chance you are critically hit.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_CRIT_TAKEN"] = {
      {-- Molten Armor
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.05,
        },
        ["buff"] = 30482,        -- ["Molten Armor"],
      },
      {-- Firestarter
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 2,
        --["num"] = 15,
        ["rank"] = {
          0.05,
        },
        ["buff"] = 30482,        -- ["Molten Armor"],
      },
    },
    -- Improved Mana Gem - Rank 2/2 - 1,19 - Buff: 83098
    -- 4.0.1: Increases your spell power by 1/2% of your maximum mana
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_SPELL_DMG_MOD_MANA"] = {
      {
        --["tab"] = 1,
        --["num"] = 19,
        ["rank"] = {
          0.01, 0.02,
        },
        ["buff"] = 83098,        -- ["Improved Mana Gem"],
      },
    },
    -- Improved Mana Gem - Rank 2/2 - 1,19 - Buff: 83098
    -- 4.0.1: Increases your spell power by 1/2% of your maximum mana
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_HEAL_MOD_MANA"] = {
      {
        --["tab"] = 1,
        --["num"] = 19,
        ["rank"] = {
          0.01, 0.02,
        },
        ["buff"] = 83098,        -- ["Improved Mana Gem"],
      },
    },
    -- Mage: Mage Armor - Buff: 6117
    -- 4.0.1: Regenerate 3% of your maximum mana every 5 sec.
    -- Mage: Glyph of Frost Armor - Glyph: 98397 - Buff: 7302
    -- 4.1.0: Your Frost Armor also causes you to regenerate 2% of your maximum mana every 5 sec.
    ["ADD_COMBAT_MANA_REGEN_MOD_MANA"] = {
      {-- Mage Armor
        ["rank"] = {
          0.03,
        },
        ["buff"] = 6117,        -- ["Mage Armor"],
      },
      {-- Glyph of Frost Armor
        ["rank"] = {
          0.02,
        },
        ["buff"] = 7302,        -- ["Frost Armor"],
        ["glyph"] = 98397,
        ["new"] = 13914, -- 4.1.0
      },
    },
    -- Mage: Frost Armor - Buff: 7302
    -- 4.0.1: Increases armor from items by 20%.
    ["MOD_ARMOR"] = {
      {-- Frost Armor
        ["rank"] = {
          0.2,
        },
        ["buff"] = 7302,        -- ["Frost Armor"],
        ["old"] = 13914, -- 4.1.0
      },
    },
    -- Mage: Prismatic Cloak - Rank 3/3 - 1,11
    -- 4.0.1: Reduces all damage taken by 2/4/6%.
    -- Mage: Frost Armor - Buff: 7302
    -- 4.1.0: Reduces physical damage taken by 15%
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_DMG_TAKEN"] = {
      {
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 1,
        --["num"] = 11,
        ["rank"] = {
          -0.02, -0.04, -0.06,
        },
      },
      {-- Frost Armor
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["rank"] = {
          -0.15,
        },
        ["buff"] = 7302,        -- ["Frost Armor"],
        ["new"] = 13914, -- 4.1.0
      },
    },
  }
elseif playerClass == "PALADIN" then
	StatModTable["PALADIN"] = {
    -- Healers: Meditation
    -- 4.0.1: Allows 50% of your mana regeneration from Spirit to continue while in combat.
    ["ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN"] = {
      {-- Meditation
        ["rank"] = {
          0.50,
        },
        ["known"] = 95859,
      },
    },
    -- Paladin: Enlightened Judgements - Rank 2/2 - 1,11
    -- 4.0.1: Grants hit rating equal to 50/100% of any Spirit gained from items or effects
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_SPELL_HIT_RATING_MOD_SPI"] = {
      {-- Enlightened Judgements
        ["spellid"] = 53557,
        --["tab"] = 1,
        --["num"] = 11,
        ["rank"] = {
          0.5, 1,
        },
      },
    },
    -- Tanks: Forceful Deflection - Passive
	-- Apply Aura: Mod Combat Rating by % of Stat (8) Value: 27
    -- 4.0.1: Increases your Parry Rating by 25% of your total Strength. (No spell, talent, ability or buff. Tanks just get it)
    -- 4.2.0: Increases your Parry Rating by 27% of your total Strength.
	-- 5.0.4: Forceful deflection spell is invisible, but strength still contributes to Parry 
	--        But Strength contributes to Parry%, not Parry Rating. 
	--        Bonus strength also suffers diminishing returns.
	--[[
    ["ADD_PARRY_RATING_MOD_STR"] = {
		{
			--todo: only apply this if they're in "protection" spec
			["rank"] = { 1.08847381108549, }
		},
		{
			["spellid"] = 49410,
			["rank"] = { 0.25, },
			["old"] = 14333, -- 4.2.0
		},
		{
			["spellid"] = 49410,
			["rank"] = { 0.27, },
			["old"] = 14333, -- 4.2.0
		},
    },
	--]]
	-- 5.0.4: Strength adds to Parry %, with bonus strength suffering diminishing returns
	["ADD_PARRY_MOD_STR"] = {
		{
			["known"] = 82242, --SpellID 82242: Parry (Paladin)
			["rank"] = {1}, 
		},
	},
    -- Paladin: Sheath of Light - Passive: 53503
    -- 4.0.1: Increases your spell power by an amount equal to 30% of your attack power
    ["ADD_SPELL_DMG_MOD_AP"] = {
      {-- Sheath of Light
        ["rank"] = {
          0.3,
        },
        ["known"] = 53503,
      },
    },
    -- Paladin: Sheath of Light - Passive: 53503
    -- 4.0.1: Increases your spell power by an amount equal to 30% of your attack power
    ["ADD_HEAL_MOD_AP"] = {
      {-- Sheath of Light
        ["rank"] = {
          0.3,
        },
        ["known"] = 53503,
      },
    },
    -- Paladin: Touched by the Light - Passive: 53592
    -- 4.0.1: Increases your total Stamina by 15%, increases your spell hit by 6%, and increases your spell power by an amount equal to 60% of your Strength.
    ["ADD_SPELL_DMG_MOD_STR"] = {
      {-- Touched by the Light
        ["rank"] = {
          0.6,
        },
        ["known"] = 53592,
      },
    },
    -- Paladin: Touched by the Light - Passive: 53592
    -- 4.0.1: Increases your total Stamina by 15%, increases your spell hit by 6%, and increases your spell power by an amount equal to 60% of your Strength.
    ["ADD_HEAL_MOD_STR"] = {
      {-- Touched by the Light
        ["rank"] = {
          0.6,
        },
        ["known"] = 53592,
      },
    },
    -- Paladin: Sanctuary - Rank 3/3 - 2,8
    -- 4.0.1: Reduces the chance you'll be critically hit by melee attacks by 2/4/6% and reduces all damage taken by 3/7/10%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_CRIT_TAKEN"] = {
      {-- Sanctuary
        ["MELEE"] = true,
        --["tab"] = 2,
        --["num"] = 8,
        ["rank"] = {
          -0.02, -0.04, -0.06,
        },
      },
    },
    -- Paladin: Sanctuary - Rank 3/3 - 2,8
    -- 4.0.1: Reduces the chance you'll be critically hit by melee attacks by 2/4/6% and reduces all damage taken by 3/7/10%.
    -- Paladin: Ardent Defender - Buff: 31850
    -- 4.0.1: Damage taken reduced by 20%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_DMG_TAKEN"] = {
      {-- Sanctuary
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 2,
        --["num"] = 8,
        ["rank"] = {
          -0.03, -0.07, -0.1,
        },
      },
      {-- Ardent Defender
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.2,
        },
        ["buff"] = 31850,
      },
    },
    -- Paladin: Toughness - Rank 3/3 - 2,5
    -- 4.0.1: Increases your armor value from items by 3/6/10%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_ARMOR"] = {
      {
        --["tab"] = 2,
        --["num"] = 5,
        ["rank"] = {
          0.03, 0.06, 0.1,
        },
      },
    },
    -- Paladin: Plate Specialization - Passive: 86525
    -- 4.0.1: Increases your primary attribute by 5% while wearing Plate in all armor slots. Holy specialization grants Intellect, Protection specialization grants Stamina, and Retribution specialization grants Strength.
    -- Paladin: Touched by the Light - Passive: 53592
    -- 4.0.1: Increases your total Stamina by 15%, increases your spell hit by 6%, and increases your spell power by an amount equal to 60% of your Strength.
    ["MOD_STA"] = {
      {-- Plate Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86525,
        ["armorspec"] = 2,
      },
      {-- Touched by the Light
        ["rank"] = {
          0.15,
        },
        ["known"] = 53592,
      },
    },
    -- Paladin: Plate Specialization - Passive: 86525
    -- 4.0.1: Increases your primary attribute by 5% while wearing Plate in all armor slots. Holy specialization grants Intellect, Protection specialization grants Stamina, and Retribution specialization grants Strength.
    ["MOD_STR"] = {
      {-- Plate Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86525,
        ["armorspec"] = 3,
      },
    },
    -- Paladin: Plate Specialization - Passive: 86525
    -- 4.0.1: Increases your primary attribute by 5% while wearing Plate in all armor slots. Holy specialization grants Intellect, Protection specialization grants Stamina, and Retribution specialization grants Strength.
    ["MOD_INT"] = {
      {-- Plate Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86525,
        ["armorspec"] = 1,
      },
    },
  }
elseif playerClass == "PRIEST" then
	StatModTable["PRIEST"] = {
    -- Healers: Meditation
    -- 4.0.1: Allows 50% of your mana regeneration from Spirit to continue while in combat.
    -- Priest: Holy Concentration - Rank 2/2 - 2,8
    -- 4.0.1: Increases the amount of mana regeneration from Spirit while in combat by an additional 10/20%.
    -- 4.0.6: Increases the amount of mana regeneration from Spirit while in combat by an additional 15/30%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN"] = {
      {-- Meditation (Discipline)
        ["rank"] = {
          0.5,
        },
        ["known"] = 95860,
      },
      {-- Meditation (Holy)
        ["rank"] = {
          0.5,
        },
        ["known"] = 95861,
      },
      {-- Holy Concentration
        ["spellid"] = 34859,
        --["tab"] = 2,
        --["num"] = 8,
        ["rank"] = {
          0.15, 0.30,
        },
      },
    },
    -- Priest: Twisted Faith - Rank 2/2 - 3,7
    -- 4.0.1: Grants you spell hit rating equal to 50/100% of any Spirit gained from items or effects.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_SPELL_HIT_RATING_MOD_SPI"] = {
      {-- Twisted Faith
        ["spellid"] = 47577,
        --["tab"] = 3,
        --["num"] = 7,
        ["rank"] = {
          0.5, 1,
        },
      },
    },
    -- Priest: Inner Fire - Buff: 588
    -- 4.0.1: Increases armor from items by 60% and spell power by 1080.
    -- Priest: Glyph of Inner Fire - Glyph: 55686 - Buff: 588
    -- 4.0.1: Increases the armor from your Inner Fire spell by 50%.
    ["MOD_ARMOR"] = {
      {-- Inner Fire
        ["rank"] = {
          0.6,
        },
        ["buff"] = 588,        -- ["Inner Fire"],
      },
      {-- Glyph of Inner Fire
        ["rank"] = {
          0.1875, -- 1.9/1.6=1.1875
        },
        ["buff"] = 588,        -- ["Inner Fire"],
        ["glyph"] = 55686,        -- ["Glyph of Inner Fire"],
      },
    },
    -- Priest: Dispersion - Buff
    -- 4.0.1: Reduces all damage by 90%
    -- Priest: Shadowform - Buff
    -- 4.0.1: All damage you take reduced by 15%.
    -- Priest: Focused Will - Rank 2/2 - 1,19 - Buff: 45241. Stack 2
    -- 4.0.1: All damage taken reduced by 4/6%. Stacks up to 2 times.
    -- Priest: Inner Sanctum - Rank 3/3 - 1,6 - Buff: 588
    -- 4.0.1: Spell damage taken is reduced by 2/4/6% while within Inner Fire
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_DMG_TAKEN"] = {
      {-- Shadowform
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.15,
        },
        ["buff"] = 15473,        -- ["Shadowform"],
      },
      {-- Dispersion
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.9,
        },
        ["buff"] = 65544,        -- ["Dispersion"],
      },
      {-- Focused Will
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 1,
        --["num"] = 19,
        ["rank"] = {
          -0.04, -0.06,
        },
        ["buff"] = 45241,        -- ["Focused Will"],
        ["buffStack"] = true,
      },
      {-- Inner Sanctum
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 1,
        --["num"] = 6,
        ["rank"] = {
          -0.02, -0.04, -0.06,
        },
        ["buff"] = 588,        -- ["Inner Sanctum"],
      },
    },
    -- Priest: Enlightenment - Passive: 84732
    -- 4.0.1: Intellect increased by 15%.
    -- Priest: Mysticism - Passive: 89745
    -- 4.0.1: Increases your Intellect by 5%
    ["MOD_INT"] = {
      {-- Enlightenment
        ["rank"] = {
          0.15,
        },
        ["known"] = 84732,
      },
      {-- Mysticism
        ["rank"] = {
          0.05,
        },
        ["known"] = 89745,
      },
    },
  }
elseif playerClass == "ROGUE" then
	StatModTable["ROGUE"] = {
    -- Rogue: Savage Combat - Rank 2/2 - 2,16
    -- 4.0.1: Increases your total attack power by 2/4%.
    -- 4.2.0: Increases your total attack power by 3/6%.
    -- Rogue: Vitality - Passive: 61329
    -- 4.0.1: Increases your Attack Power by 15%.
    -- 4.0.1: Increases your Attack Power by 25%.
    -- 4.2.0: Increases your Attack Power by 30%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_AP"] = {
      {-- Savage Combat
        --["tab"] = 2,
        --["num"] = 16,
        ["rank"] = {
          0.02, 0.04,
        },
        ["old"] = 14333, -- 4.2.0
      },
      {-- Savage Combat
        --["tab"] = 2,
        --["num"] = 16,
        ["rank"] = {
          0.03, 0.06,
        },
        ["new"] = 14333, -- 4.2.0
      },
      {-- Vitality
        ["rank"] = {
          0.25,
        },
        ["known"] = 61329,
        ["old"] = 14333, -- 4.2.0
      },
      {-- Vitality
        ["rank"] = {
          0.3,
        },
        ["known"] = 61329,
        ["new"] = 14333, -- 4.2.0
      },
    },
    -- Rogue: Reinforced Leather - Rank 2/2 - 2,10
    -- 4.0.1: Increases your armor contribution from cloth and leather items by 25/50%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_ARMOR"] = {
      {-- Reinforced Leather
        --["tab"] = 2,
        --["num"] = 10,
        ["rank"] = {
          0.25, 0.5,
        },
      },
    },
    -- Rogue: Lightning Reflexes - Rank 3/3 - 2,8
    -- 4.0.1: Increases your chance to dodge enemy attacks by 3/6/9%
    -- Rogue: Evasion - Buff: 5277
    -- 4.0.1: Dodge chance increased by 50% and chance ranged attacks hit you reduced by 25%.
    -- Rogue: Ghostly Strike - Buff: 31022
    -- 4.0.1: Dodge chance increased by 15%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_DODGE"] = {
      {-- Lightning Reflexes
        --["tab"] = 2,
        --["num"] = 8,
        ["rank"] = {
          3, 6, 9,
        },
      },
      {-- Evasion
        ["rank"] = {
          50,
        },
        ["buff"] = 5277,        -- ["Evasion"],
      },
      {-- Ghostly Strike
        ["rank"] = {
          15,
        },
        ["buff"] = 31022,        -- ["Ghostly Strike"],
      },
    },
    -- Rogue: Evasion - Buff: 5277
    -- 4.0.1: Dodge chance increased by 50% and chance ranged attacks hit you reduced by 25%.
    ["ADD_HIT_TAKEN"] = {
      {-- Evasion
        ["RANGED"] = true,
        ["rank"] = {
          -0.25,
        },
        ["buff"] = 5277,        -- ["Evasion"],
      },
    },
    -- Rogue: Cloak of Shadows - Buff: 31224
    -- 4.0.1: Resisting all hostile spells.
    -- Rogue: Glyph of Cloak of Shadows - Buff: 31224 - Glyph: 63269
    -- 4.0.1: While Cloak of Shadows is active, you take 40% less physical damage.
    -- Rogue: Cheating Death - Buff: 45182
    -- 4.0.1: All damage taken reduced by 90%
    -- 4.1.0: All damage taken reduced by 80%
    -- Rogue: Deadened Nerves - Rank 3/3 - 1,11
    -- 4.0.1: Reduces all damage taken by 3/7/10%.%
    -- Rogue: Improved Recuperate - Rank 2/2 - 2,1 - Buff: 73651
    -- 4.0.1: Reduces all damage taken by 3/6% while your Recuperate ability is active.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_DMG_TAKEN"] = {
      {-- Cloak of Shadows
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -1,
        },
        ["buff"] = 31224,        -- ["Cloak of Shadows"],
      },
      {-- Glyph of Cloak of Shadows
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["rank"] = {
          -0.4,
        },
        ["buff"] = 31224,        -- ["Cloak of Shadows"],
        ["glyph"] = 63269,  -- ["Glyph of Cloak of Shadows"],
      },
      {-- Cheating Death
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.9,
        },
        ["buff"] = 45182,        -- ["Cheating Death"],
        ["old"] = 13914, -- 4.1.0
      },
      {-- Cheating Death
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.8,
        },
        ["buff"] = 45182,        -- ["Cheating Death"],
        ["new"] = 13914, -- 4.1.0
      },
      {-- Deadened Nerves
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 1,
        --["num"] = 11,
        ["rank"] = {
          -0.03, -0.07, -0.1,
        },
      },
      {-- Improved Recuperate
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 2,
        --["num"] = 1,
        ["rank"] = {
          -0.03, -0.06,
        },
        ["buff"] = 73651,        -- ["Recuperate"],
      },
    },
    -- Rogue: Leather Specialization - Passive: 86531
    -- 4.0.1: Increases your Agility by 5% while wearing Leather in all armor slots.
    -- Rogue: Sinister Calling - Passive: 31220
    -- 4.0.1: Increases your total Agility by 25%
    -- 4.0.6: Increases your total Agility by 30%
    ["MOD_AGI"] = {
      {-- Leather Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86531,
        ["armorspec"] = 0,
      },
      {-- Leather Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86531,
        ["armorspec"] = 1,
      },
      {-- Leather Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86531,
        ["armorspec"] = 2,
      },
      {-- Leather Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86531,
        ["armorspec"] = 3,
      },
      {-- Sinister Calling
        ["rank"] = {
          0.3,
        },
        ["known"] = 31220,
      },
    },
  }
elseif playerClass == "SHAMAN" then
	StatModTable["SHAMAN"] = {
    -- Druid: Elemental Precision - Rank 3/3 - 1,7
    -- 4.0.1: Grants you spell hit rating equal to 33/66/100% of any Spirit gained from items or effects.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_SPELL_HIT_RATING_MOD_SPI"] = {
      {
        ["spellid"] = 30674,
        --["tab"] = 1,
        --["num"] = 7,
        ["rank"] = {
          0.33, 0.66, 1,
        },
      },
    },
    -- Healers: Meditation
    -- 4.0.1: Allows 50% of your mana regeneration from Spirit to continue while in combat.
    ["ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN"] = {
      {
        ["rank"] = {
          0.50,
        },
        ["known"] = 95862,
      },
    },
    -- Shaman: Mental Quickness - Passive: 30814
    -- 4.0.1: Increases your spell power by an amount equal to 50% of your attack power
    ["ADD_SPELL_DMG_MOD_AP"] = {
      {-- Mental Quickness
        ["rank"] = {
          0.5,
        },
        ["known"] = 30814,
      },
    },
    -- Shaman: Mental Quickness - Passive: 30814
    -- 4.0.1: Increases your spell power by an amount equal to 50% of your attack power
    ["ADD_HEAL_MOD_AP"] = {
      {-- Mental Quickness
        ["rank"] = {
          0.5,
        },
        ["known"] = 30814,
      },
    },
    -- Shaman: Elemental Warding - Rank 3/3 - 1,5
    -- 4.0.1: Reduces magical damage taken by 4/8/12%.
    -- Shaman: Shamanistic Rage - Buff: 30823
    -- 4.0.1: All damage taken reduced by 30%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_DMG_TAKEN"] = {
      {-- Elemental Warding
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 1,
        --["num"] = 5,
        ["rank"] = {
          -0.04, -0.08, -0.12,
        },
      },
      {-- Shamanistic Rage
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.3,
        },
        ["buff"] = 30823,        -- ["Shamanistic Rage"],
      },
    },
    -- Shaman: Toughness - Rank 3/3 - 2,8
    -- 4.0.1: Increases your Stamina by 3/7/10%
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_STA"] = {
      {
        --["tab"] = 2,
        --["num"] = 8,
        ["rank"] = {
          0.03, 0.07, 0.1,
        },
      },
    },
    -- Shaman: Mail Specialization - Passive: 86529
    -- 4.0.1: Increases your primary attribute by 5% while wearing Mail in all armor slots. Elemental specialization grants Intellect, Enhancement specialization grants Agility, and Restoration specialization grants Intellect.
    ["MOD_AGI"] = {
      {-- Mail Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86529,
        ["armorspec"] = 2,
      },
    },
    -- Shaman: Mail Specialization - Passive: 86529
    -- 4.0.1: Increases your primary attribute by 5% while wearing Mail in all armor slots. Elemental specialization grants Intellect, Enhancement specialization grants Agility, and Restoration specialization grants Intellect.
    ["MOD_INT"] = {
      {-- Mail Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86529,
        ["armorspec"] = 1,
      },
      {-- Mail Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86529,
        ["armorspec"] = 3,
      },
    },
  }
elseif playerClass == "WARLOCK" then
	StatModTable["WARLOCK"] = {
    -- Warlock: Metamorphosis - Buff: 47241
    -- 4.0.1: Armor contribution from items increased by 600%. Chance to be critically hit by melee reduced by 6%.
    ["ADD_CRIT_TAKEN"] = {
      {-- Metamorphosis
        ["MELEE"] = true,
        ["rank"] = {
          -0.06,
        },
        ["buff"] = 47241,
      },
    },
    -- Warlock: Metamorphosis - Buff: 47241
    -- 4.0.1: Armor contribution from items increased by 600%. Chance to be critically hit by melee reduced by 6%.
    ["MOD_ARMOR"] = {
      {-- Metamorphosis
        ["rank"] = {
          6,
        },
        ["buff"] = 47241,
      },
    },
    -- 3.3.0 Imp stam total 233: pet base 118, player base 90, pet sta from player sta 0.75, pet kings 1.1, fel vitality 1.15
    -- /dump floor((118+floor(90*0.75))*1.1)*1.05 = 233.45 match
    -- /dump (118+floor(9f0*0.75))*1.1*1.05 = 224.025 wrong
	-- 5.0.3: Changed Mark of the Wild SpellID from 79061 to 1126
    ["ADD_PET_STA_MOD_STA"] = {
      { -- Base
        ["rank"] = {
          0.65,
        },
        ["condition"] = "UnitExists('pet')",
      },
      { -- Blessings on pet: floor() * 1.05
        ["rank"] = {
          0.05,
        }, -- BoK, MotW, EotSS
        ["condition"] = "UnitBuff('pet', GetSpellInfo(20217)) or UnitBuff('pet', GetSpellInfo(1126)) or UnitBuff('pet', GetSpellInfo(90363))",
      },
    },
    ["ADD_PET_INT_MOD_INT"] = {
      { -- Base
        ["rank"] = {
          0.5,
        },
        ["condition"] = "UnitExists('pet')",
      },
      { -- Blessings on pet: floor() * 1.05
        ["rank"] = {
          0.05,
        }, -- BoK, MotW, EotSS
        ["condition"] = "UnitBuff('pet', GetSpellInfo(20217)) or UnitBuff('pet', GetSpellInfo(1126)) or UnitBuff('pet', GetSpellInfo(90363))",
      },
    },
    -- Warlock: Soul Link - Buff: 25228
    -- 4.0.1: 20% of damage taken by master is taken by the demon instead.
    -- Warlock: Glyph of Soul Link - Buff: 25228 - Glyph: 63312
    -- 4.0.1: Increases the percentage of damage shared via your Soul Link by an additional 5%.
    ["MOD_DMG_TAKEN"] = {
      {-- Soul Link
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.2,
        },
        ["buff"] = 25228,        -- ["Soul Link"],
      },
      {-- Glyph of Soul Link
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.05,
        },
        ["buff"] = 25228,        -- ["Soul Link"],
        ["glyph"] = 63312,  -- ["Glyph of Soul Link"],
      },
    },
    -- Warlock: Demonic Embrace - Rank 3/3 - 2,1
    --          Increases your total Stamina by 4/7/10%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_STA"] = {
      {
        --["tab"] = 2,
        --["num"] = 1,
        ["rank"] = {
          0.04, 0.07, 0.1,
        },
      },
    },
    -- Warlock: Nethermancy - Passive: 86091
    -- 4.0.1: Increases your Intellect by 5%
    ["MOD_INT"] = {
      {-- Nethermancy
        ["rank"] = {
          0.05,
        },
        ["known"] = 86091,
      },
    },
  }
elseif playerClass == "WARRIOR" then
	StatModTable["WARRIOR"] = {
    -- Tanks: Forceful Deflection - Passive
	-- Apply Aura: Mod Combat Rating by % of Stat (8) Value: 27	
    --        Increases your Parry Rating by 25% of your total Strength.
    -- 4.2.0: Increases your Parry Rating by 27% of your total Strength.
	-- 5.0.4: Forceful deflection spell is invisible, but strength still contributes to Parry 
	--        But Strength contributes to Parry%, not Parry Rating. 
	--        Bonus strength also suffers diminishing returns.
	--[[
    ["ADD_PARRY_RATING_MOD_STR"] = {
		{
			--todo: only apply this if they're in "protection" spec
			["rank"] = { 1.08847381108549, }
		},
		{
			["spellid"] = 49410,
			["rank"] = { 0.25, },
			["old"] = 14333, -- 4.2.0
		},
		{
			["spellid"] = 49410,
			["rank"] = { 0.27, },
			["old"] = 14333, -- 4.2.0
		},
    },
	--]]
	-- 5.0.4: Strength adds to Parry %, with bonus strength suffering diminishing returns
	["ADD_PARRY_MOD_STR"] = {
		{
			["known"] = 3127, --SpellID 3127: Parry (Warrior)
			["rank"] = {1}, 
		},
	},
    -- Warrior: Bastion of Defense - Rank 2/2 - 3,10 - Stance: "Interface\\Icons\\Ability_Warrior_DefensiveStance"
    -- 4.0.1: Reduces the chance you'll be critically hit by melee attacks by 3/6% while in Defensive Stance.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["ADD_CRIT_TAKEN"] = {
      {-- Bastion of Defense
        ["MELEE"] = true,
        --["tab"] = 3,
        --["num"] = 10,
        ["rank"] = {
          -0.03, -0.06,
        },
        ["stance"] = "Interface\\Icons\\Ability_Warrior_DefensiveStance",
      },
    },
    -- Warrior: Shield Wall - Buff: 871
    -- 4.0.1: All damage taken reduced by 40%
    -- Warrior: Glyph of Shield Wall - Buff: 871 - Glyph: 63329
    -- 4.0.1: Shield Wall now reduces damage taken by 60%
    -- Warrior: Defensive Stance - Stance: "Interface\\Icons\\Ability_Warrior_DefensiveStance"
    -- 4.0.1: Decreases damage taken by 10%
    -- Warrior: Battle Stance - Stance: "Interface\\Icons\\Ability_Warrior_OffensiveStance"
    -- 4.0.1: Decreases damage taken by 5%
    -- Warrior: Death Wish - Buff: 12292
    -- 4.0.1: Increases all damage taken by 5%.
    -- Warrior: Glyph of Death Wish - Buff: 12292 - Glyph: 94374
    -- 4.0.1: Death Wish no longer increases damage taken.
    -- Warrior: Recklessness - Buff: 1719
    -- 4.0.1: All damage taken is increased by 20%.
    -- Warrior: Shield Mastery - Rank 3/3 - 3,5 - Spell Block - Buff: 97954
    -- 4.0.1: Magic damage reduced by 7/14/20%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_DMG_TAKEN"] = {
      {-- Shield Wall
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.4,
        },
        ["buff"] = 871,        -- ["Shield Wall"],
      },
      {-- Glyph of Shield Wall
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.2,
        },
        ["buff"] = 871,        -- ["Shield Wall"],
        ["glyph"] = 63329, -- Glyph of Shield Wall,
      },
      {-- Defensive Stance
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.1,
        },
        ["stance"] = "Interface\\Icons\\Ability_Warrior_DefensiveStance",
      },
      {-- Battle Stance
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.05,
        },
        ["stance"] = "Interface\\Icons\\Ability_Warrior_OffensiveStance",
      },
      {-- Death Wish
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          0.05,
        },
        ["buff"] = 12292,        -- ["Death Wish"],
      },
      {-- Glyph of Death Wish
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.05,
        },
        ["buff"] = 12292,        -- ["Death Wish"],
        ["glyph"] = 94374,        -- Glyph of Death Wish
      },
      {-- Recklessness
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          0.2,
        },
        ["buff"] = 1719,        -- ["Recklessness"],
      },
      {-- Shield Mastery
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        --["tab"] = 3,
        --["num"] = 5,
        ["rank"] = {
          -0.07, -0.14, -0.2,
        },
        ["buff"] = 97954,        -- ["Spell Block"],
        ["new"] = 13914, -- 4.1.0
      },
    },
    -- Warrior: Last Stand - Buff: 12976
    -- 4.0.1: Health increased by 30% of maximum.
    ["MOD_HEALTH"] = {
      {-- Last Stand
        ["rank"] = {
          0.3,
        },
        ["buff"] = 12975,        -- ["Last Stand"],
      },
    },
    -- Warrior: Toughness - Rank 3/3 - 3,2
    -- 4.0.1: Increases your armor value from items by 3/6/10%.
	-- 5.0.4: TODO: There are no more talents; removing tab/num.
    ["MOD_ARMOR"] = {
      {-- Toughness
        --["tab"] = 3,
        --["num"] = 2,
        ["rank"] = {
          0.03, 0.06, 0.1,
        },
      },
    },
    -- Warrior: Plate Specialization - Passive: 86526
    -- 4.0.1: Increases your primary attribute by 5% while wearing Plate in all armor slots. Arms specialization grants Strength, Fury specialization grants Strength, and Protection specialization grants Stamina.
    -- Warrior: Sentinel - Passive: 29144
    -- 4.0.1: Increases your total Stamina by 15%
    ["MOD_STA"] = {
      {-- Plate Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86526,
        ["armorspec"] = 3,
      },
      {-- Sentinel
        ["rank"] = {
          0.15,
        },
        ["known"] = 29144,
      },
    },
    -- Warrior: Plate Specialization - Passive: 86526
    -- 4.0.1: Increases your primary attribute by 5% while wearing Plate in all armor slots. Arms specialization grants Strength, Fury specialization grants Strength, and Protection specialization grants Stamina.
    ["MOD_STA"] = {
      {-- Plate Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86526,
        ["armorspec"] = 1,
      },
      {-- Plate Specialization
        ["rank"] = {
          0.05,
        },
        ["known"] = 86526,
        ["armorspec"] = 2,
      },
    },
  }
elseif playerClass == "MONK" then
	print("TODO: LibStatLogic:Building StatModTable. Player is MONK; i have no idea what stats that means for their stats. Someone needs to figure out all their numbers and add them to LibStatLogc's StatModTable");

	StatModTable["MONK"] = {
		--20120810: TODO: Add all the stuff a monk gets that affects tankpoints
	}
else
	print("A new "..playerClass.." class was added to the game! :D i have no idea what their stats are. :(  Adding empty entry to LibStatLogic's StatModTable");
	StatModTable[playerClass] = {
		--A new class was added to the game! (e.g. today it's MONK. Who 
	}
end


  StatModTable["ALL"] = {
    -- Death Knight: Improved Icy Talons - Buff: 55610
    -- 4.0.6: Melee and Ranged attack speed increased by 10%.
    -- Hunter: Hunting Party - Buff: 53290
    -- 4.0.6: Melee and Ranged attack speed increased by 10%.
    -- Hunter: Windfury Totem - Buff: 8512
    -- 4.0.6: Melee and Ranged attack speed increased by 10%.
    ["MOD_MELEE_HASTE"] = {
      {-- Improved Icy Talons
        ["rank"] = {
          0.1,
        },
        ["buff"] = 55610,
        ["group"] = D["10% Melee/Ranged Attack Speed"],
      },
      {-- Hunting Party
        ["rank"] = {
          0.1,
        },
        ["buff"] = 53290,
        ["group"] = D["10% Melee/Ranged Attack Speed"],
      },
      {-- Windfury Totem
        ["rank"] = {
          0.1,
        },
        ["buff"] = 8512,
        ["group"] = D["10% Melee/Ranged Attack Speed"],
      },
    },
    -- Druid: Moonkin Aura - Buff: 24907
    -- 4.0.6: Increases spell haste by 5%.
    -- Priest: Mind Quickening - Buff: 49868
    -- 4.0.6: Increases spell haste by 5%.
    -- Shaman: Wrath of Air Totem - Buff: 3738
    -- 4.0.6: Increases spell haste by 5%.
    ["MOD_SPELL_HASTE"] = {
      {-- Improved Icy Talons
        ["rank"] = {
          0.05,
        },
        ["buff"] = 24907,
        ["group"] = D["5% Spell Haste"],
      },
      {-- Mind Quickening
        ["rank"] = {
          0.05,
        },
        ["buff"] = 49868,
        ["group"] = D["5% Spell Haste"],
      },
      {-- Wrath of Air Totem
        ["rank"] = {
          0.05,
        },
        ["buff"] = 3738,
        ["group"] = D["5% Spell Haste"],
      },
    },
    -- ICC: Chill of the Throne
    --      Chance to dodge reduced by 20%.
    -- 4.0.1: Removed
    ["ADD_DODGE"] = {
      {
        ["rank"] = {
          -20,
        },
        ["buff"] = 69127,        -- ["Chill of the Throne"],
      },
    },
    -- Replenishment - Buff
    -- 4.0.1: Replenishes 1% of maximum mana per 10 sec.
    -- 4.0.1: Shadow Priest, Frost Mage, Survival Hunter, Destro Lock, Ret Paladin
    ["ADD_COMBAT_MANA_REGEN_MOD_MANA"] = {
      {
        ["rank"] = {
          0.005,
        },
        ["buff"] = 57669,        -- ["Replenishment"],
      },
    },
    -- Priest: Inspiration - Rank 2/2 - Buff: 15357
    -- 4.0.1: Reduces physical damage taken by 5/10%.
    -- Shaman: Ancestral Fortitude - Rank 2/2 - Buff: 16236
    -- 4.0.1: Reduces physical damage taken by 5/10%.
    -- Priest: Pain Suppression - Buff: 33206
    -- 4.0.1: All damage taken reduced by 40%
    -- Paladin: Divine Guardian - Buff: 70940
    -- 4.0.1: Damage taken reduced by 20%
    -- Death Knight: Anti-Magic Zone - Buff: 50461
    -- 4.0.1: Absorbs 75% of spell damage.
    -- Warrior: Safeguard - Buff: 46947
    -- 4.0.1: All damage taken reduced by 15/30%
    -- MetaGem: Shielded Skyflare Diamond - 41377
    -- 4.0.1: +32 Stamina and Reduce Spell Damage Taken by 2%
    -- Dwarf: Stoneform - Buff: 65116
    -- 4.0.1: Armor increased by 10%.
    -- 4.0.6: Damage taken reduced by 10%.
    ["MOD_DMG_TAKEN"] = {
      {-- Stoneform
        ["rank"] = {
          -0.1,
        },
        ["buff"] = 65116,        -- ["Stoneform"],
      },
      {-- Inspiration
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["rank"] = {
          -0.05, -0.1,
        },
        ["buff"] = 15357,        -- ["Inspiration"],
        ["group"] = D["Reduced Physical Damage Taken"],
      },
      {-- Ancestral Fortitude
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["rank"] = {
          -0.05, -0.1,
        },
        ["buff"] = 16236,        -- ["Ancestral Fortitude"],
        ["group"] = D["Reduced Physical Damage Taken"],
      },
      {-- Pain Suppression
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.4,
        },
        ["buff"] = 33206,        -- ["Pain Suppression"],
      },
      {-- Divine Guardian
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.2,
        },
        ["buff"] = 70940,        -- ["Divine Guardian"],
      },
      {-- Safeguard
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.15, -0.3,
        },
        ["buff"] = 46947,        -- ["Safeguard"],
      },
      {-- Anti-Magic Zone
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.75,
        },
        ["buff"] = 50461,        -- ["Anti-Magic Zone"],
      },
      {-- Effulgent Skyflare Diamond
        ["HOLY"] = true,
        ["FIRE"] = true,
        ["NATURE"] = true,
        ["FROST"] = true,
        ["SHADOW"] = true,
        ["ARCANE"] = true,
        ["rank"] = {
          -0.02,
        },
        ["meta"] = 41377,
      },
    },
    -- Warlock: Demonic Pact - Buff: 53646
    -- 4.0.1: Spell Power increased by 10%.
    -- Shaman: Totemic Wrath - Buff: 77747
    -- 4.0.1: Spell Power increased by 10%.
    -- Shaman: Flametongue Totem - Buff: 52109
    -- 4.0.1: Spell Power increased by 6%.
    -- Mage: Arcane Brilliance - Buff: 79058
    -- 4.0.1: Spell Power increased by 6%.
    -- Mage: Dalaran Brilliance - Buff: 61316
    -- 4.0.1: Spell Power increased by 6%.
    ["MOD_SPELL_DMG"] = {
      {-- Demonic Pact
        ["rank"] = {
          0.1,
        },
        ["buff"] = 53646,        -- ["Demonic Pact"],
        ["group"] = D["Spell Power Multiplier"],
      },
      {-- Totemic Wrath
        ["rank"] = {
          0.1,
        },
        ["buff"] = 77747,        -- ["Totemic Wrath"],
        ["group"] = D["Spell Power Multiplier"],
      },
      {-- Flametongue Totem
        ["rank"] = {
          0.06,
        },
        ["buff"] = 52109,        -- ["Flametongue Totem"],
        ["group"] = D["Spell Power Multiplier"],
      },
      {-- Arcane Brilliance
        ["rank"] = {
          0.06,
        },
        ["buff"] = 79058,        -- ["Arcane Brilliance"],
        ["group"] = D["Spell Power Multiplier"],
      },
      {-- Dalaran Brilliance
        ["rank"] = {
          0.06,
        },
        ["buff"] = 61316,
        ["group"] = D["Spell Power Multiplier"],
      },
    },
    -- Warlock: Demonic Pact - Buff: 53646
    -- 4.0.1: Spell Power increased by 10%.
    -- Shaman: Totemic Wrath - Buff: 77747
    -- 4.0.1: Spell Power increased by 10%.
    -- Shaman: Flametongue Totem - Buff: 52109
    -- 4.0.1: Spell Power increased by 6%.
    -- Mage: Arcane Brilliance - Buff: 79058
    -- 4.0.1: Spell Power increased by 6%.
    ["MOD_HEAL"] = {
      {
        ["rank"] = {
          0.1,
        },
        ["buff"] = 53646,        -- ["Demonic Pact"],
        ["group"] = D["Spell Power Multiplier"],
        ["newtoc"] = 40000,
      },
      {
        ["rank"] = {
          0.1,
        },
        ["buff"] = 77747,        -- ["Totemic Wrath"],
        ["group"] = D["Spell Power Multiplier"],
        ["newtoc"] = 40000,
      },
      {
        ["rank"] = {
          0.06,
        },
        ["buff"] = 52109,        -- ["Flametongue Totem"],
        ["group"] = D["Spell Power Multiplier"],
        ["newtoc"] = 40000,
      },
      {
        ["rank"] = {
          0.06,
        },
        ["buff"] = 79058,        -- ["Arcane Brilliance"],
        ["group"] = D["Spell Power Multiplier"],
        ["newtoc"] = 40000,
      },
    },
    -- Night Elf : Quickness - Racial
    -- 4.0.1: Reduces the chance that melee and ranged attackers will hit you by 2%.
    ["ADD_HIT_TAKEN"] = {
      {
        ["MELEE"] = true,
        ["RANGED"] = true,
        ["rank"] = {
          -0.02,
        },
        ["race"] = "NightElf",
      },
    },
    -- MetaGem: Eternal Earthsiege Diamond - Meta: 41396
    -- 4.0.1: +1% Shield Block Value
    -- MetaGem: Eternal Earthstorm Diamond - Meta: 35501
    -- 4.0.1: +1% Shield Block Value
    -- MetaGem: Eternal Shadowspirit Diamond - Meta: 52293
    -- 4.0.1: +5% Shield Block Value
    ["ADD_BLOCK_REDUCTION"] = {
      {-- Eternal Earthsiege Diamond
        ["rank"] = {
          0.01,
        },
        ["meta"] = 41396,
      },
      {-- Eternal Earthstorm Diamond
        ["rank"] = {
          0.01,
        },
        ["meta"] = 35501,
      },
      {-- Eternal Shadowspirit Diamond
        ["rank"] = {
          0.05,
        },
        ["meta"] = 52293,
      },
    },
    -- MetaGem: Austere Earthsiege Diamond - Meta: 41380
    -- 4.0.1: 2% Increased Armor Value from Items
    -- MetaGem: Austere Shadowspirit Diamond - Meta: 52294
    -- 4.0.1: 2% Increased Armor Value from Items
    ["MOD_ARMOR"] = {
      {-- Austere Earthsiege Diamond
        ["rank"] = {
          0.02,
        },
        ["meta"] = 41380,
      },
      {-- Austere Shadowspirit Diamond
        ["rank"] = {
          0.02,
        },
        ["meta"] = 52294,
      },
    },
    -- Hunter: Trueshot Aura - Buff: 19506
    -- 4.0.1: Attack power increased by 10%.
    -- Death Knight: Abomination's Might - Buff: 55972
    -- 4.0.1: Attack power increased by 5/10%.
    -- Shaman: Unleashed Rage - Buff: 30809
    -- 4.0.1: Melee attack power increased by 4/7/10%.
    -- Paladin: Blessing of Might - Buff: 19740
    -- 4.0.1: Increasing attack power by 10%.
    ["MOD_AP"] = {
      {-- Trueshot Aura
        ["rank"] = {
          0.1,
        },
        ["buff"] = 19506,        -- ["Trueshot Aura"],
        ["group"] = D["Attack Power Multiplier"],
      },
      {-- Abominable Might
        ["rank"] = {
          0.05, 0.1,
        },
        ["buff"] = 55972,        -- ["Abominable Might"],
        ["group"] = D["Attack Power Multiplier"],
      },
      {-- Unleashed Rage
        ["rank"] = {
          0.04, 0.07, 0.1,
        },
        ["buff"] = 30809,        -- ["Unleashed Rage"],
        ["group"] = D["Attack Power Multiplier"],
      },
      {-- Blessing of Might
        ["rank"] = {
          0.1,
        },
        ["buff"] = 19740,        -- ["Blessing of Might"],
        ["group"] = D["Attack Power Multiplier"],
      },
    },
    -- Gnome: Expansive Mind - Racial
    -- 4.0.1: Mana pool increased by 5%.
    -- MetaGem: Ember Skyfire Diamond - Meta: 35503
    -- 4.0.1: +2% Maximum Mana
    -- MetaGem: Ember Skyflare Diamond - Meta: 41333
    -- 4.0.1: +2% Maximum Mana
    -- MetaGem: Beaming Earthsiege Diamond - Meta: 41389
    -- 4.0.1: +2% Mana
    -- MetaGem: Ember Shadowspirit Diamond - Meta: 52296
    -- 4.0.1: +2% Maximum Mana
    ["MOD_MANA"] = {
      {-- Expansive Mind
        ["rank"] = {
          0.05,
        },
        ["race"] = "Gnome",
      },
      {-- Beaming Earthsiege Diamond
        ["rank"] = {
          0.02,
        },
        ["meta"] = 41389,
      },
      {-- Ember Skyfire Diamond
        ["rank"] = {
          0.02,
        },
        ["meta"] = 35503,
      },
      {-- Ember Skyflare Diamond
        ["rank"] = {
          0.02,
        },
        ["meta"] = 41333,
      },
      {-- Ember Shadowspirit Diamond
        ["rank"] = {
          0.02,
        },
        ["meta"] = 52296,
      },
    },
    -- Paladin: Blessing of Kings - Buff: 20217
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Druid: Mark of the Wild - Buff: 1126
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Hunter: Embrace of the Shale Spider - Buff: 90363
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Leatherworking: Blessing of Forgotten Kings - Buff: 69378
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 4%.
	-- 5.0.3: Changed Mark of the Wild SpellID from 79061 to 1126
    ["MOD_STR"] = {
      {-- Blessing of Kings
        ["rank"] = {
          0.05,
        },
        ["buff"] = 20217,        -- ["Blessing of Kings"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Mark of the Wild
        ["rank"] = {
          0.05,
        },
        ["buff"] = 1126,        -- ["Mark of the Wild"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Embrace of the Shale Spider
        ["rank"] = {
          0.05,
        },
        ["buff"] = 90363,        -- ["Embrace of the Shale Spider"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Blessing of Forgotten Kings
        ["rank"] = {
          0.04,
        },
        ["buff"] = 69378,        -- ["Blessing of Forgotten Kings"],
        ["group"] = D["Stat Multiplier"],
      },
    },
    -- Paladin: Blessing of Kings - Buff: 20217
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Druid: Mark of the Wild - Buff: 1126
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Hunter: Embrace of the Shale Spider - Buff: 90363
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Leatherworking: Blessing of Forgotten Kings - Buff: 69378
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 4%.
	-- 5.0.3: Changed Mark of the Wild SpellID from 79061 to 1126
    ["MOD_AGI"] = {
      {-- Blessing of Kings
        ["rank"] = {
          0.05,
        },
        ["buff"] = 20217,        -- ["Blessing of Kings"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Mark of the Wild
        ["rank"] = {
          0.05,
        },
        ["buff"] = 1126,        -- ["Mark of the Wild"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Embrace of the Shale Spider
        ["rank"] = {
          0.05,
        },
        ["buff"] = 90363,        -- ["Embrace of the Shale Spider"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Blessing of Forgotten Kings
        ["rank"] = {
          0.04,
        },
        ["buff"] = 69378,        -- ["Blessing of Forgotten Kings"],
        ["group"] = D["Stat Multiplier"],
      },
    },
    -- Paladin: Blessing of Kings - Buff: 20217
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Druid: Mark of the Wild - Buff: 1126
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Hunter: Embrace of the Shale Spider - Buff: 90363
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Leatherworking: Blessing of Forgotten Kings - Buff: 69378
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 4%.
	-- 5.0.3: Changed Mark of the Wild SpellID from 79061 to 1126
    ["MOD_STA"] = {
      {-- Blessing of Kings
        ["rank"] = {
          0.05,
        },
        ["buff"] = 20217,        -- ["Blessing of Kings"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Mark of the Wild
        ["rank"] = {
          0.05,
        },
        ["buff"] = 1126,        -- ["Mark of the Wild"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Embrace of the Shale Spider
        ["rank"] = {
          0.05,
        },
        ["buff"] = 90363,        -- ["Embrace of the Shale Spider"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Blessing of Forgotten Kings
        ["rank"] = {
          0.04,
        },
        ["buff"] = 69378,        -- ["Blessing of Forgotten Kings"],
        ["group"] = D["Stat Multiplier"],
      },
    },
    --[[
        Up to level 80 1 stamina grants 10 health. But levels 81..85 that value increases. 
        It could be implemented as a GetHealthPerStamina(level) function (which i started to do), 
        but everyone would have to change their code to stop using the hard-coded "10 health per stamina".
        Instead we can code it as a set of conditional percentage modifiers for all classes in the "MOD_HEALTH" group
            Level    HP/sta    % increase
            =====    =======    ========
            80        10.0    0.00   (confirmed 10.0 12/12/2010, 4.0.3)
            81        10.8    0.08
            82        11.6    0.12   (confirmed 11.6 12/11/2010, 4.0.3)
            83        12.4    0.24   (confirmed 12.4 12/12/2010, 4.0.3)
            84        13.2    0.32
            85        14.0    0.40
            
        Question for Whitefang: Could the modifier be written as the following?:
                {
                    ["rank"] = { 0.08 * (UnitLevel('player') - 80) },
                    ["condition"] = "UnitLevel('player') >= 81"
                }
                
            It would be future proofed, but i don't know if it's right to code the rank based on a formula 
                (i.e. what happens if they ding?)
    --]]
    ["MOD_HEALTH"] = {
        -- Level 81..85
        {    
            ["rank"] = { 0.08, }, 
            ["condition"] = "UnitLevel('player') == 81", 
        },
        {    
            ["rank"] = { 0.12, }, 
            ["condition"] = "UnitLevel('player') == 82", 
        },
        {    
            ["rank"] = { 0.24, }, 
            ["condition"] = "UnitLevel('player') == 83", 
        },
        {    
            ["rank"] = { 0.32, }, 
            ["condition"] = "UnitLevel('player') == 84", 
        },
        {    
            ["rank"] = { 0.40, }, 
            ["condition"] = "UnitLevel('player') == 85", 
        },
    },
      
    -- Paladin: Blessing of Kings - Buff: 20217
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Druid: Mark of the Wild - Buff: 1126
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Hunter: Embrace of the Shale Spider - Buff: 90363
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 5%.
    -- Leatherworking: Blessing of Forgotten Kings - Buff: 69378
    -- 4.0.1: Strength, Agility, Stamina, and Intellect increased by 4%.
	-- 5.0.3: Changed Mark of the Wild SpellID from 79061 to 1126
    ["MOD_INT"] = {
      {-- Blessing of Kings
        ["rank"] = {
          0.05,
        },
        ["buff"] = 20217,        -- ["Blessing of Kings"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Mark of the Wild
        ["rank"] = {
          0.05,
        },
        ["buff"] = 1126,        -- ["Mark of the Wild"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Embrace of the Shale Spider
        ["rank"] = {
          0.05,
        },
        ["buff"] = 90363,        -- ["Embrace of the Shale Spider"],
        ["group"] = D["Stat Multiplier"],
      },
      {-- Blessing of Forgotten Kings
        ["rank"] = {
          0.04,
        },
        ["buff"] = 69378,        -- ["Blessing of Forgotten Kings"],
        ["group"] = D["Stat Multiplier"],
      },
    },
    -- Shaman: Mana Tide - Buff: 16191
    -- 4.0.1: Spirit increased by 350%.
    -- Human: The Human Spirit - Racial
    -- 4.0.1: Spirit increased by 3%.
    ["MOD_SPI"] = {
      {-- Mana Tide
        ["rank"] = {
          3.5,
        },
        ["buff"] = 16191,        -- ["Mana Tide"],
      },
      {-- The Human Spirit
        ["rank"] = {
          0.03,
        },
        ["race"] = "Human",
      },
    },
  }

-- Generate buff names
for class, tables in pairs(StatModTable) do
	for modName, mods in pairs(tables) do
		for key, mod in pairs(mods) do
			if mod.buff then
				mod.buffName = GetSpellInfo(mod.buff) or "nil"
			end
			if mod.buff2 then
				mod.buff2Name = GetSpellInfo(mod.buff2) or "nil"
			end
		end
	end
end

local function IsMetaGemActive(item)
	-- Check item
	if (type(item) == "string") or (type(item) == "number") then
	elseif type(item) == "table" and type(item.GetItem) == "function" then
		-- Get the link
		_, item = item:GetItem()
		if type(item) ~= "string" then return end
	else
		return
	end
	-- Check if item is in local cache
	local name, link, _, _, _, _, itemSubType = GetItemInfo(item)
	if not name or itemSubType ~= META_GEM then return end
	-- Start parsing
	tip:ClearLines() -- this is required or SetX won't work the second time its called
	tip:SetHyperlink(link)
	if not strfind(tip[3]:GetText(), "|cff808080") then
		-- Metagem requirements satisfied, check if metagem is equipped
		local headLink = GetInventoryItemLink("player", 1)
		if not headLink then return end
		local gemId = StatLogic:GetGemID(item)
		if not gemId then return end
		if strfind(headLink, ":"..gemId..":") then
			return true
		end
	end
end
StatLogic.IsMetaGemActive = IsMetaGemActive

local function SlotHasEnchant(enchantId, slotId)
	-- Check args
	if type(enchantId) ~= "number" then return end
	if (type(slotId) ~= "number") and (type(slotId) ~= "nil") then return end
	-- If slot is specified
	if type(slotId) == "number" then
		local slotLink = GetInventoryItemLink("player", slotId)
		if not slotLink then return end
		if strfind(slotLink, ":"..enchantId..":") then
			return 1
		end
	else
		-- check all slots 1 to 18 if slotId is nil
		local count = 0
		for slotId = 1, 18 do
			local slotLink = GetInventoryItemLink("player", slotId)
			if slotLink and strfind(slotLink, ":"..enchantId..":") then
				count = count + 1
			end
		end
		if count ~= 0 then
			return count
		end
	end
end
StatLogic.SlotHasEnchant = SlotHasEnchant

--[[---------------------------------
  :GetStatMod(stat, school)
-------------------------------------
Notes:
  * Calculates various stat mod values from talents and buffs.
  * initialValue: sets the initial value for the stat mod.
  ::if initialValue == 0, inter-mod operations are done with addition,
  ::if initialValue == 1, inter-mod operations are done with multiplication,
  * finalAdjust: added to the final result before returning, so we can adjust the return value to be used in addition or multiplication.
  :: for addition: initialValue + finalAdjust = 0
  :: for multiplication: initialValue + finalAdjust = 1
  * stat:
  :{| class="wikitable"
  !"StatMod"!!Initial value!!Final adjust!!schoo required
  |-
  |"ADD_CRIT_TAKEN"||0||0||Yes
  |-
  |"ADD_HIT_TAKEN"||0||0||Yes
  |-
  |"ADD_DODGE"||0||0||No
  |-
  |"ADD_AP_MOD_INT"||0||0||No
  |-
  |"ADD_AP_MOD_STA"||0||0||No
  |-
  |"ADD_AP_MOD_ARMOR"||0||0||No
  |-
  |"ADD_PARRY_RATING_MOD_STR"||0||0||No
  |-
  |"ADD_COMBAT_MANA_REGEN_MOD_INT"||0||0||No
  |-
  |"ADD_RANGED_AP_MOD_INT"||0||0||No
  |-
  |"ADD_ARMOR_MOD_INT"||0||0||No
  |-
  |"ADD_SPELL_DMG_MOD_AP"||0||0||No
  |-
  |"ADD_SPELL_DMG_MOD_STA"||0||0||No
  |-
  |"ADD_SPELL_DMG_MOD_INT"||0||0||No
  |-
  |"ADD_SPELL_DMG_MOD_SPI"||0||0||No
  |-
  |"ADD_HEAL_MOD_AP"||0||0||No
  |-
  |"ADD_HEAL_MOD_STR"||0||0||No
  |-
  |"ADD_HEAL_MOD_AGI"||0||0||No
  |-
  |"ADD_HEAL_MOD_STA"||0||0||No
  |-
  |"ADD_HEAL_MOD_INT"||0||0||No
  |-
  |"ADD_HEAL_MOD_SPI"||0||0||No
  |-
  |"ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN"||0||0||No
  |-
  |"MOD_CRIT_DAMAGE_TAKEN"||0||1||Yes
  |-
  |"MOD_DMG_TAKEN"||0||1||Yes
  |-
  |"MOD_CRIT_DAMAGE"||0||1||Yes
  |-
  |"MOD_DMG"||0||1||Yes
  |-
  |"MOD_ARMOR"||1||0||No
  |-
  |"MOD_HEALTH"||1||0||No
  |-
  |"MOD_MANA"||1||0||No
  |-
  |"MOD_STR"||0||1||No
  |-
  |"MOD_AGI"||0||1||No
  |-
  |"MOD_STA"||0||1||No
  |-
  |"MOD_INT"||0||1||No
  |-
  |"MOD_SPI"||0||1||No
  |-
  |"MOD_BLOCK_VALUE"||0||1||No
  |-
  |"MOD_AP"||0||1||No
  |-
  |"MOD_RANGED_AP"||0||1||No
  |-
  |"MOD_SPELL_DMG"||0||1||No
  |-
  |"MOD_HEAL"||0||1||No
  |}
Arguments:
  string - The type of stat mod you want to get
  [optional] string - Certain stat mods require an extra school argument
  [optional] string - Target spec to check
  [optional] string - modTable to check, "CLASS", "ALL", "BOTH"(default)
Returns:
  None
Example:
  StatLogic:GetStatMod("MOD_INT")
-----------------------------------]]
local buffGroup = {}
function StatLogic:GetStatMod(stat, school, talentGroup, modTable)
	local statModInfo = StatModInfo[stat]
  
    --Check that it wasn't an invalid stat
    if (statModInfo == nil) then
        local sError = "StatLogic:GetStatMod() Invalid stat mod requested \""..stat.."\""
        print(sError)
        error(sError)
    end
        
    local mod = statModInfo.initialValue
    -- if school is required for this statMod but not given
    if statModInfo.school and not school then
        --print("school is required for this stat, but none given. Returning finalAdjust value")
        return mod + statModInfo.finalAdjust 
    end
    -- disable for 4.0.1 until we get talent/buffs data implemented
    --if toc >= 40000 then return mod + statModInfo.finalAdjust end
    wipe(buffGroup)
  
	if not modTable then modTable = "BOTH" end
    
    -- Class specific mods
    if type(StatModTable[playerClass][stat]) == "table" and modTable ~= "ALL" then
    
        for _, case in ipairs(StatModTable[playerClass][stat]) do
        
            local ok = true
            if school and not case[school] then ok = nil end
            
            if ok and case.newtoc and toc < case.newtoc then ok = nil end
            if ok and case.oldtoc and toc >= case.oldtoc then ok = nil end
            if ok and case.new and wowBuildNo < case.new then ok = nil end
            if ok and case.old and wowBuildNo >= case.old then ok = nil end
            if ok and case.condition and not loadstring("return "..case.condition)() then ok = nil end
            if ok and case.buffName and not PlayerHasAura(case.buffName) then ok = nil end
            if ok and case.buff2Name and not PlayerHasAura(case.buff2Name) then ok = nil end
            if ok and case.stance and case.stance ~= GetStanceIcon() then ok = nil end
            if ok and case.glyph and not PlayerHasGlyph(case.glyph, talentGroup) then ok = nil end
            if ok and case.enchant and not SlotHasEnchant(case.enchant, case.slot) then ok = nil end
            if ok and case.itemset and ((not PlayerItemSets[case.itemset[1]]) or PlayerItemSets[case.itemset[1]] < case.itemset[2]) then ok = nil end
            if ok and case.armorspec and case.armorspec ~= ArmorSpecActive then ok = nil end
            if ok and case.known and not IsSpellKnown(case.known) then ok = nil end

	--if ok then
--		print(string.format("GetStatMod(%s) %s - OK", stat, case.rank[1]));
--	else
		--print(string.format("GetStatMod(%s) %s - not OK", stat, case.rank[1]));
	--end;

            
            if ok then
                local r, _
                local s = 1
                -- if talent field
                if case.tab and case.num then
                    _, _, _, _, r = GetTalentInfo(case.tab, case.num, nil, nil, talentGroup)
                    if case.buffName and case.buffStack then
                        _, s = GetPlayerAuraRankStack(case.buffName) -- Gets buff stack count, but use talent as rank
                    end
                elseif case.buffName then
					-- no talent but buff is given
                    r, s = GetPlayerAuraRankStack(case.buffName)
                    if not case.buffStack then
                        s = 1
                    end
                    -- no talent but all other given conditions are statisfied
                else--if case.condition or case.stance then
                    r = 1
                end
                if r and case.rank[r] then
                    if statModInfo.initialValue == 0 and not case.mul then
                        if not case.group then
                            mod = mod + case.rank[r] * s
                        elseif not buffGroup[case.group] then -- this mod is part of a group, but not seen before
                            mod = mod + case.rank[r] * s
                            buffGroup[case.group] = case.rank[r] * s
                        elseif (case.rank[r] * s) > buffGroup[case.group] then -- seen before and this one is better, do upgrade
                            mod = mod + case.rank[r] * s - buffGroup[case.group]
                            buffGroup[case.group] = case.rank[r] * s
                        else 
							-- seen before but not better, do nothing
                        end
                    else
                        if not case.group then
                            mod = mod * (case.rank[r] * s + 1)
                        elseif not buffGroup[case.group] then -- this mod is part of a group, but not seen before
                            mod = mod * (case.rank[r] * s + 1)
                            buffGroup[case.group] = (case.rank[r] * s + 1)
                        elseif (case.rank[r] * s + 1) > buffGroup[case.group] then -- seen before and this one is better, do upgrade
                            mod = mod * (case.rank[r] * s + 1) / buffGroup[case.group]
                            buffGroup[case.group] = (case.rank[r] * s + 1)
                        else 
							-- seen before but not better, do nothing
                        end
                    end
                end
            end
        end
    end
    
    -- Non class specific mods
    if type(StatModTable["ALL"][stat]) == "table" and modTable ~= "CLASS"  then
        for _, case in ipairs(StatModTable["ALL"][stat]) do
            local ok = true
            if school and not case[school] then ok = nil end
            if ok and case.newtoc and toc < case.newtoc then ok = nil end
            if ok and case.oldtoc and toc >= case.oldtoc then ok = nil end
            if ok and case.new and wowBuildNo < case.new then ok = nil end
            if ok and case.old and wowBuildNo >= case.old then ok = nil end
            if ok and case.condition and not loadstring("return "..case.condition)() then ok = nil end
            if ok and case.buffName and not PlayerHasAura(case.buffName) then ok = nil end
            if ok and case.stance and case.stance ~= GetStanceIcon() then ok = nil end
            if ok and case.race and case.race ~= playerRace then ok = nil end
            if ok and case.meta and not IsMetaGemActive(case.meta) then ok = nil end
            if ok then
                local r, _
                local s = 1
                -- if talent field
                if case.tab and case.num then
                    _, _, _, _, r = GetTalentInfo(case.tab, case.num, nil, nil, talentGroup)
                    if case.buffName and case.buffStack then
                        _, s = GetPlayerAuraRankStack(case.buffName) -- Gets buff rank and stack count
                    end
                    -- no talent but buff is given
                elseif case.buffName then
                    r, s = GetPlayerAuraRankStack(case.buffName)
                    if not case.buffStack then
                        s = 1
                    end
                    -- no talent but all other given conditions are statisfied
                else--if case.condition or case.stance then
                    r = 1
                end
                if r and case.rank[r] then
                    if statModInfo.initialValue == 0 then
                        if not case.group then
                            mod = mod + case.rank[r] * s
                        elseif not buffGroup[case.group] then -- this mod is part of a group, but not seen before
                            mod = mod + case.rank[r] * s
                            buffGroup[case.group] = case.rank[r] * s
                        elseif (case.rank[r] * s) > buffGroup[case.group] then -- seen before and this one is better, do upgrade
                            mod = mod + case.rank[r] * s - buffGroup[case.group]
                            buffGroup[case.group] = case.rank[r] * s
                        else -- seen before but not better, do nothing
                        end
                    else
                        if not case.group then
                            mod = mod * (case.rank[r] * s + 1)
                        elseif not buffGroup[case.group] then -- this mod is part of a group, but not seen before
                            mod = mod * (case.rank[r] * s + 1)
                            buffGroup[case.group] = (case.rank[r] * s + 1)
                        elseif (case.rank[r] * s + 1) > buffGroup[case.group] then -- seen before and this one is better, do upgrade
                            mod = mod * (case.rank[r] * s + 1) / buffGroup[case.group]
                            buffGroup[case.group] = (case.rank[r] * s + 1)
                        else -- seen before but not better, do nothing
                        end
                    end
                end
            end
        end
    end

    return mod + statModInfo.finalAdjust
end

--=====================================--
-- Avoidance stats diminishing returns --
--=====================================--
-- Formula reverse engineered by Whitetooth (hotdogee [at] gmail [dot] com)
--[[---------------------------------
This includes
1. Dodge from Dodge Rating, Agility.
2. Parry from Parry Rating.
3. Chance to be missed from level difference.

The following is the result of hours of work gathering data from beta servers and then spending even more time running 
multiple regression analysis on the data.

1. DR for Dodge, Parry, Missed are calculated separately.
2. Base avoidances are not affected by DR, (ex: Dodge from base Agility)
3. Death Knight's Parry from base Strength is affected by DR, base for parry is 5%.
4. Direct avoidance gains from talents and spells(ex: Evasion) are not affected by DR.
5. Indirect avoidance gains from talents and spells(ex: +Agility from Kings) are affected by DR
6. c and k values depend on class but does not change with level.

7. The DR formula:

1/x' = 1/c+k/x

x' is the diminished stat before converting to IEEE754.
x is the stat before diminishing returns.
c is the cap of the stat, and changes with class.
k is is a value that changes with class.
-----------------------------------]]
-- The following K, C_p, C_d are calculated by Whitetooth (hotdogee [at] gmail [dot] com)
local K = {
	0.956, -- 1: WARRIOR
	0.885970968828508, --0.956, -- 2: PALADIN
	0.988, -- 3: HUNTER
	0.988, -- 4: ROGUE
	0.983, -- 5: PRIEST
	0.956, -- 6: DEATHKNIGHT
	0.988, -- 7: SHAMAN
	0.983, -- 8: MAGE
	0.983, -- 9: WARLOCK
	0.975, --10: MONK  20120809: added monk entry. TODO: come up with a real number for "K". i just typed in 0.975 since it looks averageish. i don't even know what "K" is supposed to represent
	0.972, --11: DRUID
}

--C_p - Parry cap constant. In the limit of diminishing returns; Parry% will not be able to exceed C_p
local C_p = {
	1/0.0152366,		-- 1: WARRIOR
	237.164919575025,	-- 2: PALADIN   
	1/0.006870,			-- 3: HUNTER
	1/0.006870,			-- 4: ROGUE
	1/0.0152366,		-- 5: PRIEST
	1/0.0152366,		-- 6: DEATHKNIGHT
	1/0.006870,			-- 7: SHAMAN
	1/0.0152366,		-- 8: MAGE
	1/0.0152366,		-- 9: WARLOCK
	1/0.0153,			--10: MONK  20120809: added Monk entry. TODO: come up with a real number for "C_p". i just typed in 0.0153 since it looks averageish.
	1/0.0152366,		--11: DRUID
}

--C_d - Dodge cap constant
local C_d = {
	1/0.0152366,		-- 1: WARRIOR
	66.552213340240500,	-- 2: PALADIN   1/0.0152366
	1/0.006870,			-- 3: HUNTER
	1/0.006870,			-- 4: ROGUE
	1/0.006650,			-- 5: PRIEST
	1/0.0152366,		-- 6: DEATHKNIGHT
	1/0.006870,			-- 7: SHAMAN
	1/0.006650,			-- 8: MAGE
	1/0.006650,			-- 9: WARLOCK
	1/0.008,			--10: MONK  20120809: added Monk entry. TODO: come up with a real number for "C_d". i just typed in 0.008 since it looks averageish. i don't even know what "C_d" is supposed to represent.
	1/0.008555,			--11: DRUID
}

--C_m - Melee hit avoid cap constant
--[[
	I've done extensive tests that show the miss cap is 16% for warriors.
	Because the only tank I have with 150 pieces of epic gear required for the tests is a warrior,
	Until someone that has the will and gear to preform the tests for other classes, I'm going to assume the cap is the same(which most likely isn't)
--]]
local C_m = {
	16,	-- 1: WARRIOR
	16,	-- 2: PALADIN
	16,	-- 3: HUNTER
	16,	-- 4: ROGUE
	16,	-- 5: PRIEST
	16,	-- 6: DEATHKNIGHT
	16,	-- 7: SHAMAN
	16,	-- 8: MAGE
	16,	-- 9: WARLOCK
	16,	--10: MONK  20120809: added Monk entry. TODO: come up with a real number for "C_d". i just typed in 0.008 since it looks averageish. i don't even know what "C_d" is supposed to represent.
	16,	--11: DRUID
}

--[[
	Returns your 
--]]
function StatLogic:GetMissedChanceBeforeDR()
	--[[
	20120812 Your base chance to miss a mob used to be 5% starting at mobs your own level (see PaperDollFrame.lua:
			BASE_MISS_CHANCE_PHYSICAL = {
				[0] = 5.0;   mob your own level        |==========
				[1] = 5.5;   mob +1 level above you    |===========
				[2] = 6.0;   mob +2 levels above you   |============
				[3] = 8.0;   mob +3 levels above you   |================
			};
			
	With 5.0.4 (Mists of Panderia), the base chance is now 3%:
			BASE_MISS_CHANCE_PHYSICAL = {
				[0] = 3.0;   |======
				[1] = 4.5;   |=========
				[2] = 6.0;   |============
				[3] = 7.5;   |===============
			};			

	Defense, and all the math that went along with it, was removed in cataclysm.
	Defense was used to be a complicated method to essentially work out that i am more likely
	to miss mobs that are higher level than me.	
	In fact, in Cataclysm the scale was non-linear.
	In Mists it does look linear (over the +0..+3 range), but your chance to be missed still depends on
	how many levels above you the attacker is, not your defense.
	--]]
			
	--local baseDefense, additionalDefense = UnitDefense("player"); --400, 0 for level 85
	--local defenseFromDefenseRating = floor(self:GetEffectFromRating(GetCombatRating(CR_DEFENSE_SKILL), CR_DEFENSE_SKILL)); --GetCombatRating(CR_DEFENSE_SKILL) always returns zero
	--local modMissed = defenseFromDefenseRating * 0.04
	--local drFreeMissed = 5 + (baseDefense + additionalDefense - defenseFromDefenseRating) * 0.04
	--return modMissed, drFreeMissed
	return BASE_MISS_CHANCE_PHYSICAL[0], BASE_MISS_CHANCE_PHYSICAL[0]
end

--[[---------------------------------
  :GetDodgeChanceBeforeDR()
-------------------------------------
Notes:
  * Calculates your current Dodge% before diminishing returns.
  * Dodge% = modDodge + drFreeDodge
  * drFreeDodge includes:
  ** Base dodge
  ** Dodge from base agility
  ** Dodge modifier from base defense
  ** Dodge modifers from talents or spells
  * modDodge includes
  ** Dodge from dodge rating
  ** Dodge from additional defense
  ** Dodge from additional dodge
Arguments:
  None
Returns:
  ; modDodge : number - The part that is affected by diminishing returns.
  ; drFreeDodge : number - The part that isn't affected by diminishing returns.
Example:
  local modDodge, drFreeDodge = StatLogic:GetDodgeChanceBeforeDR()
	8.8917, 5.0257
	which when added gives 13.9174. 
	Then then has to suffer DR down to 13.
-----------------------------------]]
local BaseDodge
function StatLogic:GetDodgeChanceBeforeDR()

	--UseGetDodgeChance to do all the heavy lifting
	
	local dodgeRating = GetCombatRating(CR_DODGE);
	local agility = UnitStat("player", 2); --2=Strength
	local class = ClassNameToID[playerClass];
	
	local dodgeChance, dodgeBeforeDR, freeDodge = StatLogic:GetDodgeChance(dodgeRating, agility, class);

	return dodgeBeforeDR-freeDodge, freeDodge;

	--[[
	local class = ClassNameToID[playerClass];

	-- drFreeDodge
	local stat, effectiveStat, posBuff, negBuff = UnitStat("player", 2); -- 2 = Agility
	local baseAgi = stat - posBuff - negBuff;
	local dodgePerAgi = self:GetDodgePerAgi();
  
	--local drFreeDodge = BaseDodge[class] + dodgePerAgi * baseAgi
	--			+ self:GetStatMod("ADD_DODGE") + (baseDefense - UnitLevel("player") * 5) * 0.04
	-- modDodge
	local dodgeFromDodgeRating = self:GetEffectFromRating(GetCombatRating(CR_DODGE), CR_DODGE, UnitLevel("player"));
	--local dodgeFromDefenceRating = floor(self:GetEffectFromRating(GetCombatRating(CR_DEFENSE_SKILL), CR_DEFENSE_SKILL)) * 0.04;
	local dodgeFromAdditionalAgi = dodgePerAgi * (effectiveStat - baseAgi);
	local modDodge = dodgeFromDodgeRating + dodgeFromAdditionalAgi;

	local drFreeDodge = GetDodgeChance() - self:GetAvoidanceAfterDR("DODGE", modDodge, class);

	return modDodge, drFreeDodge
	--]]
end


--[[---------------------------------
  :GetParryChanceBeforeDR()
-------------------------------------
Notes:
  * Calculates your current Parry% before diminishing returns.
  * Parry% = modParry + drFreeParry
  * drFreeParry includes:
  ** Base parry
  ** Parry from base agility
  ** Parry modifers from talents or spells
  * modParry includes
  ** Parry from parry rating
  ** Parry from additional parry
Arguments:
  None
Returns:
  ; modParry : number - The part that is affected by diminishing returns.
  ; drFreeParry : number - The part that isn't affected by diminishing returns.
Example:
  local modParry, drFreeParry = StatLogic:GetParryChanceBeforeDR()
-----------------------------------]]
function StatLogic:GetParryChanceBeforeDR()

	--UseGetParryChance to do all the heavy lifting
	local parryRating = GetCombatRating(CR_PARRY);
	local strength = UnitStat("player", 1); --1=Strength
	local class = ClassNameToID[playerClass];
	
	local parryChance, parryBeforeDR, freeParry = StatLogic:GetParryChance(parryRating, strength, class);

	return parryBeforeDR-freeParry, freeParry;
	
	--[[	
	local parryFromParryRating = self:GetEffectFromRating(GetCombatRating(CR_PARRY), CR_PARRY)
	local modParry = parryFromParryRating;

	-- drFreeParry
	local drFreeParry = GetParryChance() - self:GetAvoidanceAfterDR("PARRY", modParry, class)

	return modParry, drFreeParry;
	--]]
end

--[[---------------------------------
  :GetAvoidanceAfterDR(avoidanceType, avoidanceBeforeDR[, class])
-------------------------------------
Notes:
  * Avoidance DR formula and k, C_p, C_d constants derived by Whitetooth (hotdogee [at] gmail [dot] com)
  * avoidanceBeforeDR is the part that is affected by diminishing returns.
  * See :GetClassIdOrName(class) for valid class values.
  * Calculates the avoidance after diminishing returns, this includes:
  *# Dodge from Dodge Rating, Agility.
  *# Parry from Parry Rating.
  *# Chance to be missed.
  * The DR formula: 1/x' = 1/c+k/x
  ** x' is the diminished stat before converting to IEEE754.
  ** x is the stat before diminishing returns.
  ** c is the cap of the stat, and changes with class.
  ** k is is a value that changes with class.
  * Formula details:
  *# DR for Dodge, Parry, Missed are calculated separately.
  *# Base avoidances are not affected by DR, (ex: Dodge from base Agility)
  *# Death Knight's Parry from base Strength is affected by DR, base for parry is 5%.
  *# Direct avoidance gains from talents and spells(ex: Evasion) are not affected by DR.
  *# Indirect avoidance gains from talents and spells(ex: +Agility from Kings) are affected by DR
  *# c and k values depend on class but does not change with level.
  :{| class="wikitable"
  ! !!k!!C_p!!1/C_p!!C_d!!1/C_d
  |-
  |Warrior||0.9560||47.003525||0.021275||88.129021||0.011347
  |-
  |Paladin||0.9560||47.003525||0.021275||88.129021||0.011347
  |-
  |Hunter||0.9880||145.560408||0.006870||145.560408||0.006870
  |-
  |Rogue||0.9880||145.560408||0.006870||145.560408||0.006870
  |-
  |Priest||0.9530||0||0||150.375940||0.006650
  |-
  |Deathknight||0.9560||47.003525||0.021275||88.129021||0.011347
  |-
  |Shaman||0.9880||145.560408||0.006870||145.560408||0.006870
  |-
  |Mage||0.9530||0||0||150.375940||0.006650
  |-
  |Warlock||0.9530||0||0||150.375940||0.006650
  |-
  |Monk|| ? || ? || ? || ? || ?
  |-
  |Druid||0.9720||0||0||116.890707||0.008555
  |}
Arguments:
  string - "DODGE", "PARRY", "MELEE_HIT_AVOID"(NYI)
  number - amount of avoidance before diminishing returns in percentages.
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
Returns:
  ; avoidanceAfterDR : number - avoidance after diminishing returns in percentages.
Example:
  local modParry, drFreeParry = StatLogic:GetParryChanceBeforeDR()
  local modParryAfterDR = StatLogic:GetAvoidanceAfterDR("PARRY", modParry)
  local parry = modParryAfterDR + drFreeParry

  local modParryAfterDR = StatLogic:GetAvoidanceAfterDR("PARRY", modParry, "WARRIOR")
  local parry = modParryAfterDR + drFreeParry
-----------------------------------]]
function StatLogic:GetAvoidanceAfterDR(avoidanceType, avoidanceBeforeDR, class)

	-- argCheck for invalid input
	self:argCheck(avoidanceType, 2, "string")
	self:argCheck(avoidanceBeforeDR, 3, "number")
	self:argCheck(class, 4, "nil", "string", "number")

	--[[
		20120810 - Strange quirk in LUA
			1/0 is +INF (division by zero)
		but
			1 / ( 1/0) is 0

		Some users are experiencing division by zero error
		Lets avoid the division by zero altogether.
	--]]
	if (avoidanceBeforeDR == 0) then
		return 0;
	end;
	
	-- if class is a class string, convert to class id
	if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
		class = ClassNameToID[strupper(class)]
	elseif type(class) ~= "number" or class < 1 or class > CLASSID_MAX then
		-- if class is invalid input, default to player class
		class = ClassNameToID[playerClass]
	end

	local C = C_d; --assume they want "DODGE"
	if avoidanceType == "PARRY" then
		C = C_p
	elseif avoidanceType == "MELEE_HIT_AVOID" then
		C = C_m
	end
	
	if (class == CLASSID_MONK) then --MONK
		print("TODO: LibStatLogic:GetAvoidanceAfterDR - table C_dodge[], C_parry[], C_meleeHitAvoid[] and K[] need to have their MONK values updated");
	end;
	
--	print("avoidanceType: "..avoidanceType);
--	print("avoidanceBeforeDR: "..avoidanceBeforeDR);
--	print("class: "..class);
	
--	print("C[class]: "..C[class]);
--	print("K[class]: "..K[class]);
	
	return 1 / (1 / C[class] + K[class] / avoidanceBeforeDR)
end

--[[---------------------------------
  :GetAvoidanceGainAfterDR(avoidanceType, gainBeforeDR)
-------------------------------------
Notes:
  * Calculates the avoidance gain after diminishing returns with player's current stats.
Arguments:
  string - "DODGE", "PARRY", "MELEE_HIT_AVOID"(NYI)
  number - Avoidance gain before diminishing returns in percentages.
Returns:
  ; gainAfterDR : number - Avoidance gain after diminishing returns in percentages.
Example:
  -- How much dodge will I gain with +30 Agi after DR?
  local gainAfterDR = StatLogic:GetAvoidanceGainAfterDR("DODGE", 30*StatLogic:GetDodgePerAgi())
  -- How much dodge will I gain with +20 Parry Rating after DR?
  local gainAfterDR = StatLogic:GetAvoidanceGainAfterDR("PARRY", StatLogic:GetEffectFromRating(20, CR_PARRY))
-----------------------------------]]
function StatLogic:GetAvoidanceGainAfterDR(avoidanceType, gainBeforeDR)
	-- argCheck for invalid input
	self:argCheck(gainBeforeDR, 2, "number")
	local class = ClassNameToID[playerClass]

	if avoidanceType == "PARRY" then
		local modAvoidance, drFreeAvoidance = self:GetParryChanceBeforeDR();
		local newAvoidanceChance = self:GetAvoidanceAfterDR(avoidanceType, modAvoidance + gainBeforeDR) + drFreeAvoidance;
		if newAvoidanceChance < 0 then newAvoidanceChance = 0 end;
		return newAvoidanceChance - GetParryChance();
	elseif avoidanceType == "DODGE" then
		local modAvoidance, drFreeAvoidance = self:GetDodgeChanceBeforeDR();
		local newAvoidanceChance = self:GetAvoidanceAfterDR(avoidanceType, modAvoidance + gainBeforeDR) + drFreeAvoidance;
		if newAvoidanceChance < 0 then newAvoidanceChance = 0 end; -- because GetDodgeChance() is 0 when negative
		return newAvoidanceChance - GetDodgeChance();
	elseif avoidanceType == "MELEE_HIT_AVOID" then
		local modAvoidance = self:GetMissedChanceBeforeDR();
		return self:GetAvoidanceAfterDR(avoidanceType, modAvoidance + gainBeforeDR) - self:GetAvoidanceAfterDR(avoidanceType, modAvoidance);
	end
end

--[[
	parryChance, parryBeforeDR, freeParry = StatLogic:GetParryChance([parryRating], [strength], [classId]);
	
	params
		parryRating (number, optional): The players Parry Rating. If nil will use the current player's Parry Rating
		strength (number, optional): The player's Strength. If nil will use the current player's Strength
		classId (number, string, optiona): The player's classId number or ClassName string. If empty will use the current player's class

	returns
		parryChance (number): the player's Parry Chance, a percentage number (e.g. 37.5432)
		parryBeforeDR (number): the player's Parry Chance before diminishing returns, a percentage number (e.g. 38.8832)
		freeParry (number): the portion of Parry Chance that does not suffer diminishing returns (e.g 5.01)
							Is a combination of base Parry Chance, as well as Parry Chance from base Strength
--]]
function StatLogic:GetParryChance(parryRating, strength, class)
--[[
	Thanks to Thack for figuring out how Strength affects Parry.
		http://sacredduty.net/2012/07/06/avoidance-diminishing-returns-in-mop-part-3/
		
	totalParry = (baseParry+baseStr/Q) +                      1
	                                     -----------------------------------------
										 1/Cp +                k
										        --------------------------------
												(totalStr-baseStr)/Q + preParry
												
	Where 
		baseParry: is your per-race base Parry amount
		baseStr:   is your Strength befor any +Strength (varies per race)
		preParry:  is your Parry from Parry Rating, before diminishing returns (=GetCombatRatingBonus(CR_PARRY), parryRating/Pc )
		Cp:        is a constant for Parry
		k:         is a constant
		Q:         is a constant that converts strength to Parry Chance
		
	The important points are:
		- if you're naked, =GetParryChance() returns the sum from (baseParry + baseStr/Q)
		- Parry from base Strength doesn't suffer diminishing returns
		- Parry from bonus Strength does suffer diminishing returns
		- Parry from Parry Rating does suffer diminishing returns
		
	baseParry
		HUMAN = 3%
		
	baseStrength   UnitStat("player", 1)[1]-[3]
		HUMAN = 164 
		
	Diminishing returns formulas has a strange effect that low parry values are boosted.
		Any pre-parry value (parryFromStrength + parryFromRating) above:
			
					Cp-Cp*k
					
		is diminshed, while values under that are increased.
		For level 85 paladin that works out to:
					Cp - Cp*k
					= Cp*(1-k)
					= 237.164919575071 * (1-0.88598405026478)
					= 237.164919575071 * (0.11401594973522)
					= 27.04058354922878897736270062
					
		Which means it can help to think of Cp as an upper-bound, and k defines the transition point on that curve
--]]
	--if Parry Rating is nil, then get the player's current Parry Rating
	if (parryRating == nil) then
		parryRating = GetCombatRating(CR_PARRY);
	end;

	--if strength is nil, then get the player's current Strength
	if (strength == nil) then
		strength = UnitStat("player", 1); --1=Strength
	end;

	-- if class is a class string, convert to class id
	if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
		class = ClassNameToID[strupper(class)]
	elseif type(class) ~= "number" or class < 1 or class > CLASSID_MAX then
		-- if class is invalid input, default to player class
		class = ClassNameToID[playerClass]
	end;

	local baseParry = 3; --e.g. 3%, Humans have base parry of 3%
	local baseStr = 164; --base strength
	
	--SpellID: 3127  Parry (Warrior)
	--SpellID: 82245 Parry (Rogue)
	--SpellID: 82246 Parry (Death Knight)
	--SpellID: 82247 Parry (Paladin)
	
	local base, stat, posBuff, negBuff = UnitStat("player", 1); --1=Strength
	local baseStr = stat-posBuff+negBuff; --e.g. 5125, 5125, 4953, 0
	
	local Cp = C_p[class]; --parry cap constant (e.g. 237.164919575071)
	local k = K[class]; --diminishing returns constant, e.g. 0.88598405026478 for Paladin
	local Q = 243.608552799873; --conversion factor of Strength to Parry%. E.g. 243.6 Strength --> +1% Parry (before DR)
	
	local Pc = 265.078335678336000; --Parry% from Parry Rating, before diminishing returns

	local freeParry = baseParry + baseStr/Q;
	local bonusParry = (strength-baseStr)/Q + parryRating/Pc;

	local parryBeforeDR = freeParry + bonusParry; --combination of free Parry, Parry from Parry Rating, and Parry from bonus Strength

	--print(string.format("Cp = %s, k=%s, Q=%s, strength=%s, baseStr=%s, parryRating=%s, freeParry=%s, bonusParry=%s, parryBeforeDR=%s", 
	--		Cp, k, Q, strength, baseStr, parryRating, freeParry, bonusParry, parryBeforeDR));
	
	local parryChance;
	if (bonusParry ~= 0) then
		parryChance = freeParry + 1/( 1/Cp + k/bonusParry );
	else
		parryChance = freeParry;
	end;

	return parryChance, parryBeforeDR, freeParry;
end;


--[[
	dodgeChance, dodgeBeforeDR, freeDodge = StatLogic:GetDodgeChance([dodgeRating], [agility], [classId]);
	
	params
		dodgeRating (number, optional): The players Dodge Rating. If nil will use the current player's Dodge Rating
		agility (number, optional): The player's Agility. If nil will use the current player's Agility
		classId (number, string, optiona): The player's classId number or ClassName string. If empty will use the current player's class

	returns
		dodgeChance (number): the player's Dodge Chance, a percentage number (e.g. 37.5432)
		dodgeBeforeDR (number): the player's Dodge Chance before diminishing returns, a percentage number (e.g. 38.8832)
		freeDodge (number): the portion of Dodge Chance that does not suffer diminishing returns (e.g 5.01)
							Is a combination of base Dodge Chance, as well as Dodge Chance from base Agility
--]]
function StatLogic:GetDodgeChance(dodgeRating, agility, class)
--[[
	Thanks to Thack for figuring out how Agility affects Dodge.
		http://sacredduty.net/2012/07/06/avoidance-diminishing-returns-in-mop-part-3/
		
	totalDodge = (baseDodge+baseAgi/Qd) +                      1
	                                     -----------------------------------------
										 1/Cp +                k
										        --------------------------------
												(totalStr-baseStr)/Qd + preDodge
												
	Where 
		baseDodge: is your per-race base Dodge amount
		baseAgi:   is your Agility befor any +Agility (varies per race) (e.g. 97)
		preDodge:  is your Dodge from Dodge Rating, before diminishing returns (=GetCombatRatingBonus(CR_DODGE), dodgeRating/Pd )
		Cd:        is a Dodge cap constant (e.g. 66.552213340240500)
		k:         is a constant (e.g. 0.885970968828508000)
		Qd:        is a constant that converts Agility to Dodge Chance (10025.4295440002000)
		
	The important points are:
		- if you're naked, =GetDodgeChance() returns the sum from (baseDodge + baseStr/Qd)
		- Dodge from base Agility doesn't suffer diminishing returns
		- Dodge from bonus Agility does suffer diminishing returns
		- Dodge from Dodge Rating does suffer diminishing returns
		
	baseDodge
		HUMAN = 5%
		
	baseAgi   UnitStat("player", 1)[1]-[3]
		HUMAN = 97 
		
	Diminishing returns formulas has a strange effect that low Dodge values are boosted.
		Any pre-Dodge value (dodgeFromAgility + dodgeFromRating) above:
			
					Cd-Cd*k
					
		is diminshed, while values under that are increased.
		For level 85 paladin that works out to:
					Cd - Cd*k
					= Cd*(1-k)
					= 66.552213340240500 * (1-0.88598405026478)
					= 66.552213340240500 * (0.11401594973522)
					= 7.5880138109684988
					
		Which means it can help to think of Cd as an upper-bound, and k defines the transition point on that curve
--]]
	--if Dodge Rating is nil, then get the player's current Dodge Rating
	if (dodgeRating == nil) then
		dodgeRating = GetCombatRating(CR_DODGE);
	end;

	--if Agility is nil, then get the player's current Agility
	if (agility == nil) then
		agility = UnitStat("player", 2); --2=Agility
	end;

	-- if class is a class string, convert to class id
	if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
		class = ClassNameToID[strupper(class)]
	elseif type(class) ~= "number" or class < 1 or class > CLASSID_MAX then
		-- if class is invalid input, default to player class
		class = ClassNameToID[playerClass]
	end;

	local baseDodge = 5; --e.g. 5%, Humans have base dodge of 5%
	local baseAgi = 97; --base Agility
	
	local base, stat, posBuff, negBuff = UnitStat("player", 2); --1=Agility
	local baseAgi = stat-posBuff+negBuff; --e.g. 5125, 5125, 4953, 0
	
	local Cd = C_d[class]; --dodge cap constant (e.g. 66.552213340240500)
	local k = K[class]; --diminishing returns constant, e.g. 0.88598405026478 for Paladin
	local Qd = 10025.42954400020000; --conversion factor of Agility to Dodge%. E.g. 10025.4 Strength --> +1% Dodge (before DR)
	
	local Pd = 265.078335678336000; --Dodge% from Dodge Rating, before diminishing returns (same value as for Parry)

	local freeDodge = baseDodge + baseAgi/Qd;
	local bonusDodge = (agility-baseAgi)/Qd + dodgeRating/Pd;

	local dodgeBeforeDR = freeDodge + bonusDodge; --combination of free Dodge, Dodge from Dodge Rating, and Dodge from bonus Agility

	--print(string.format("Cp = %s, k=%s, Q=%s, strength=%s, baseStr=%s, dodgeRating=%s, freeDodge=%s, bonusDodge=%s, dodgeBeforeDR=%s", 
	--		Cp, k, Q, strength, baseStr, dodgeRating, freeDodge, bonusDodge, dodgeBeforeDR));
	
	local dodgeChance;
	if (bonusDodge ~= 0) then
		dodgeChance = freeDodge + 1/( 1/Cd + k/bonusDodge );
	else
		dodgeChance = freeDodge;
	end;

	return dodgeChance, dodgeBeforeDR, freeDodge;
end;



function StatLogic:GetResilienceEffectAfterDR(damageReductionBeforeDR)
	-- argCheck for invalid input
	self:argCheck(damageReductionBeforeDR, 2, "number")
	return 100 - 100 * 0.99 ^ damageReductionBeforeDR
end

function StatLogic:GetResilienceEffectGainAfterDR(resAfter, resBefore)
	-- argCheck for invalid input
	self:argCheck(resAfter, 2, "number")
	self:argCheck(resBefore, 2, "nil", "number")
	local resCurrent = GetCombatRating(16)
	local drBefore
	if resBefore then
		drBefore = self:GetResilienceEffectAfterDR(self:GetEffectFromRating(resCurrent + resBefore, COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN))
	else
		drBefore = GetCombatRatingBonus(16)
	end
	return self:GetResilienceEffectAfterDR(self:GetEffectFromRating(resCurrent + resAfter, COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN)) - drBefore
end


--=================--
-- Stat Conversion --
--=================--
--[[---------------------------------
  :GetReductionFromArmor(armor, attackerLevel)
-------------------------------------
Notes:
  * Calculates the damage reduction from armor for given attacker level.
Arguments:
  [optional] number - Armor value. Default: player's armor value
  [optional] number - Attacker level. Default: player's level
Returns:
  ; damageRecudtion : number - Damage reduction value from 0 to 1. (not percentage)
Example:
  local damageRecudtion = StatLogic:GetReductionFromArmor(35000, 80) -- 0.69676006569452
-----------------------------------]]
function StatLogic:GetReductionFromArmor(armor, attackerLevel)
	self:argCheck(armor, 2, "nil", "number");
	self:argCheck(attackerLevel, 3, "nil", "number");
	if not armor then
		armor = select(2, UnitArmor("player"));
	end
	if not attackerLevel then
		attackerLevel = UnitLevel("player");
	end

	--Taken from PaperDollFrame.lua from Mists Beta (20120805 v5.0.4)
	--function PaperDollFrame_GetArmorReduction(armor, attackerLevel)  
	local levelModifier = attackerLevel;
	if ( levelModifier > 85 ) then
		levelModifier = levelModifier + (4.5 * (levelModifier-59)) + (20 * (levelModifier - 80)) + (22 * (levelModifier - 85));
	elseif ( levelModifier > 80 ) then
		levelModifier = levelModifier + (4.5 * (levelModifier - 59)) + (20 * (levelModifier - 80));
	elseif ( levelModifier > 59 ) then
		levelModifier = levelModifier + (4.5 * (levelModifier - 59));
	end
  	local temp = armor/(85*levelModifier + 400);
	local armorReduction = temp/(1+temp);
  
	-- caps at 0.75
	if armorReduction > 0.75 then
		armorReduction = 0.75
	end
	if armorReduction < 0 then
		armorReduction = 0
	end
	return armorReduction;
end

--[[---------------------------------
  :GetEffectFromDefense(defense, attackerLevel)
-------------------------------------
Notes:
  * Calculates the effective avoidance% from defense (before diminishing returns) for given attacker level
Arguments:
  [optional] string - Total defense value. Default: player's armor value
  [optional] number - Attacker level. Default: player's level
Returns:
  ; effect : number - 0.04% per effective defense.
Example:
  local effect = StatLogic:GetEffectFromDefense(415, 83) -- 0
-----------------------------------]]
--[[
	20120812 - Removed, as Defense was removed 2 years ago
	function StatLogic:GetEffectFromDefense(defense, attackerLevel)
	print("LibStatLogic:GetEffectFromDefense - Defense was removed from the game 2 years ago");
	
	self:argCheck(defense, 2, "nil", "number")
	self:argCheck(attackerLevel, 3, "nil", "number")
	if not defense then
		local base, add = UnitDefense("player")
		defense = base + add
	end
	if not attackerLevel then
		attackerLevel = UnitLevel("player")
	end
	return (defense - attackerLevel * 5) * 0.04
end
--]]

-- Build Mastery tables
-- MasterySpells[class][specIndex]
local MasterySpells = {
	{	--1. WARRIOR: 
		{ 76838}, --Arms(Damage)      Mastery: Strikes of Opportunity
		{ 76856}, --Fury(Damage)      Mastery: Unshackled Fury
		{ 76857}, --Protection(Tank)  Mastery: Critical Block
	}, 
	{	--2. Paladin
		{ 76669}, --Holy(Healer)         Mastery: Illuminated Healing
		{ 76671}, --Protection(Tank)     Mastery: Divine Bulwark
		{ 76672}, --Retribution(Damage)  Mastery: Hand of Light
	}, 
	{	--3. Hunter
		{ 76657}, --"Master of Beasts", 
		{ 76659}, --"Wild Quiver", 
		{ 76658}, --"Essence of the Viper"
	}, 
	{	--4. Rogue
		{ 76803}, 
		{ 76806}, 
		{ 76808},
	},
	{	--5. Priest
		{ 77484},
		{ 77485}, 
		{ 77486},
	},
	{	--6. Death Knight
		{ 77513},
		{ 77514},
		{ 77515},
	},
	{	--7. Shaman (verified 5.0.4, 16016)
		{ 77222}, --Elemental(Damage)    Mastery: Elemental Overload 
		{ 77223}, --Enhancement(Damage)  Mastery: Enhanced Elements
		{ 77226}, --Restoration(Healer)  Mastery: Deep Healing
	}, 
	{	--8. Mage
		{ 76547},
		{ 76595},
		{ 76613},
	},
	{
		--9. Warlock
		{ 77215},
		{ 77219},
		{ 77220},
	},
	{	--10. Monk
		{117906}, --Brewmaster: "Mastery: Elusive Brawler"      (Tank)
		{117907}, --Mistweaver: "Mastery: Gift of the Serpent"  (Healer)
		{115636}, --Windwalker: "Mastery: Combo Breaker"        (Damage)
	}, 
	{	--11. Druid (verified 5.0.4, build 15961)
		{77492}, --Balance(Damage)
		{77493}, --Feral(Damage)
		{77494}, --Guardian(Tank)
		{77495}, --Restoration(Healer)   
	},
}
StatLogic.MasterySpells = MasterySpells

-- MasteryEffect[class][specIndex]
local MasteryEffect = {
	{{1.00 }, {1.00}, {1.00,1.00}    }, --WARRIOR (1) (Strikes of Opportunity 2%, Unshackled Fury 4.70%, Block Chance 1.5% & Critical Block Chance 1.5%)
	{{1.00 }, {1.00}, {1.00}         }, --PALADIN (2) (Illuminated Healing 1%, Divine Bulwark 2.25%, Hand of Light 1%)
	{{1.00 }, {1.00}, {1.00}         }, --HUNTER (3)
	{{1.00 }, {1.00}, {1.00}         }, --ROGUE (4)
	{{1.00 }, {1.00}, {1.00}         }, --PRIEST (5)
	{{1.00 }, {1.00}, {1.00}         }, --DEATHKNIGHT (6)
	{{1.00 }, {1.00}, {1.00}         }, --SHAMAN (7)
	{{1.00 }, {1.00}, {1.00}         }, --MAGE (8)
	{{1.00 }, {1.00}, {1.00}         }, --WARLOCK (9)
	{{1.00 }, {1.00}, {1.00}         }, --MONK (10)
	{{1.00 }, {1.00}, {1.00}, {1.00} }, --DRUID (11)  Balance (Damage), Feral (Damage), Guardian (Tank), Restoration (Healer) 
}
if wowBuildNo < 16016 then --Patch 5.0.4 (Aug 21 2012)
	--Pre Mists of Panderia
	MasteryEffect = {
		{{2   }, {4.70}, {1.5,1.5}      }, --WARRIOR (1) (Strikes of Opportunity 2%, Unshackled Fury 4.70%, Block Chance 1.5% & Critical Block Chance 1.5%)
		{{1   }, {2.25}, {1   }         }, --PALADIN (2) (Illuminated Healing 1%, Divine Bulwark 2.25%, Hand of Light 1%)
		{{1.7 }, {2   }, {1   }         }, --HUNTER (3)
		{{3.5 }, {2   }, {2.5 }         }, --ROGUE (4)
		{{2.5 }, {1.25}, {4.3 }         }, --PRIEST (5)
		{{6.25}, {2   }, {4   }         }, --DEATHKNIGHT (6)
		{{2   }, {2.5 }, {2.5 }         }, --SHAMAN (7)
		{{1.5 }, {2.5 }, {2.5 }         }, --MAGE (8)
		{{1.63}, {1.5 }, {1.25}         }, --WARLOCK (9)
		{{1   }, {1   }, {1   }         }, --MONK (10)
		{{1.5 }, {4   }, {3.1 }, {1.25} }, --DRUID (11)  Balance (Damage), Feral (Damage), Guardian (Tank), Restoration (Healer) 
	}
end
StatLogic.MasteryEffect = MasteryEffect
-- Parse spell tooltip text and populate MasteryEffect, run this if mastery values change
-- function MakeMasteryEffect()
  -- for class, specs in ipairs(MasterySpells) do
    -- MasteryEffect[class] = {}
    -- print(strsub(ClassNameToID[class], 1, 1)..strsub(strlower(ClassNameToID[class]), 2))
    -- for spec, spellids in ipairs(specs) do
      -- MasteryEffect[class][spec] = {}
      -- for i, spellid in ipairs(spellids) do
        -- tipMiner:ClearLines() -- this is required or SetX won't work the second time its called
        -- tipMiner:AddSpellByID(spellid)
        -- print((GetSpellInfo(spellid)))
        -- print("Spec: "..(select(2, GetTalentTabInfo(spec))))
        -- print("SpellId: "..spellid)
        -- local descText = _G[MAJOR.."MinerTooltipTextLeft2"]:GetText()
        -- if not descText:find("([%.%d]+)%%") then
          -- descText = _G[MAJOR.."MinerTooltipTextLeft3"]:GetText()
          -- print(_G[MAJOR.."MinerTooltipTextLeft3"]:GetText() or "nil")
        -- else
          -- print(_G[MAJOR.."MinerTooltipTextLeft2"]:GetText() or "nil")
        -- end
        -- local effects = {}
        -- for e in descText:gmatch("([%.%d]+)%%") do
          -- tinsert(effects, tonumber(e))
        -- end
        -- if #effects == 2 then
          -- tinsert(MasteryEffect[class][spec], effects[2])
        -- elseif #effects == 3 then
          -- tinsert(MasteryEffect[class][spec], effects[3])
        -- elseif #effects == 4 then
          -- tinsert(MasteryEffect[class][spec], effects[3])
          -- tinsert(MasteryEffect[class][spec], effects[4])
        -- end
        -- --print(_G[MAJOR.."MinerTooltipTextLeft1"]:GetText() or "nil")
        -- --print(_G[MAJOR.."MinerTooltipTextLeft2"]:GetText() or "nil")
        -- --print(_G[MAJOR.."MinerTooltipTextLeft3"]:GetText() or "nil")
        -- --print("---------")
      -- end
    -- end
  -- end
  -- return MasteryEffect
-- end

--[[---------------------------------
  :GetEffectFromMastery(mastery[, specIndex][, class])
-------------------------------------
Notes:
  * Calculates the effect in percentage from mastery for given spec and class
Arguments:
  number - mastery value.
  [optional] number - talent spec to use. Default: player's talent spec
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
            1=first, 2=second, 3=third, 4=forth (e.g. Druid: 1=Balance, 2=Feral, 3=Guardian, 4=Restoration)
Returns:
  ; effect : number - 0.04% per effective defense.
Example:
  local effect = StatLogic:GetEffectFromMastery(415) -- 0
-----------------------------------]]
function StatLogic:GetEffectFromMastery(mastery, specIndex, class)
	--[[
		20120811  5.0.4 (build 15961) Mists of Panderia
		
		In Cataclysm the supplied "mastery" value is akin to Mists's "Mastery %" value. e.g.:

		Cataclysm
		=========
			Mastery 21.14 (8.00+13.14)
			Divine Bulwark
			Increases your chance to block melee attacks by 47%.
			Each point of Mastery increases block chance by an additional 2.25%
			Mastery rating 2356 (+13.14 mastery)
			
		So that calling GetEffectFromMastery(21.14) = 47.565 (i.e. 47%)
	
		In Mists on Panderia, Mastery has become a percentage itself:

		Mists of Panderia
		=================		
			Mastery 18.04% (8.00% + 10.04%)
			Mastery: Divine Bulwark
			Increases the damage reduction of your Shield of the Righteous by 18%,
				adds 18% to your Bastion of Glory, 
				and increases your chance to block melee attacks by 18%.
			Mastery 2290 (+10.04 mastery).
			
		So "Mastery Rating" now converts to a "Mastery %".
		And that "Mastery Percentage" now applies directly to whatever spell; one-to-one.
		
		There is no longer any notion of:
				"Each point of Mastery increases damage by an additional 2.5%".
				
		That means that in essence this function can be replaced with:
			return mastery;
			
		because a Mastery of 18.04% has an effect of 18.04:
		
			GetEffectFromMastery(18.04) = 18.04 (i.e. 18.04%)
				
		But i will keep the "MasteryEffect" table, filling it with all 1.0 factors.
		
		There is still the notion of having to convert "Mastery Rating" into "Mastery Percentage", 
		and subject to diminishing returns:
		
			GetEffectFromRating(2356, CR_MASTERY) = 13.141453 (4.3.4 build 15595)
	--]] 

	--TODO: convert this function to simply:
	--return 1.0, 1.0, 1.0;
	
	self:argCheck(mastery, 2, "number")
	self:argCheck(specIndex, 3, "nil", "number")

	if type(specIndex) ~= "number" or specIndex < 1 or specIndex > 4 then
	    specIndex = GetSpecialization(); --5.0.3 GetPrimaryTalentTree replaced with GetSpecialization
	    if not specIndex then return 0 end
	end
	
	-- argCheck for invalid input
	self:argCheck(class, 4, "nil", "string", "number")
	
	-- if class is a class string, convert to class id
	if type(class) == "string" then
	    class = ClassNameToID[strupper(class)] or ClassNameToID[playerClass]
	elseif type(class) ~= "number" or (class < 1) or (class > CLASSID_MAX) then
		-- if class is invalid input, default to player class
	    class = ClassNameToID[playerClass]
	end
	return 
			mastery *  MasteryEffect[class][specIndex][1], 
			mastery * (MasteryEffect[class][specIndex][2] or 0),
			mastery * (MasteryEffect[class][specIndex][3] or 0)  --some classes have even 3 effects
end


--[[---------------------------------
  :GetEffectFromRating(rating, id[, level][, class])
-------------------------------------
Notes:
  * Combat Rating formula and constants derived by Whitetooth (hotdogee [at] gmail [dot] com)
  * Calculates the stat effects from ratings for any level.
  * id: Rating ID as definded in PaperDollFrame.lua
  ::CR_WEAPON_SKILL = 1
  ::CR_DEFENSE_SKILL = 2
  ::CR_DODGE = 3
  ::CR_PARRY = 4
  ::CR_BLOCK = 5
  ::CR_HIT_MELEE = 6
  ::CR_HIT_RANGED = 7
  ::CR_HIT_SPELL = 8
  ::CR_CRIT_MELEE = 9
  ::CR_CRIT_RANGED = 10
  ::CR_CRIT_SPELL = 11
  ::CR_HIT_TAKEN_MELEE = 12
  ::CR_HIT_TAKEN_RANGED = 13
  ::CR_HIT_TAKEN_SPELL = 14
  ::COMBAT_RATING_RESILIENCE_CRIT_TAKEN = 15
  ::COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN = 16   --formerly CR_CRIT_TAKEN_RANGED
  ::CR_CRIT_TAKEN_SPELL = 17
  ::CR_HASTE_MELEE = 18
  ::CR_HASTE_RANGED = 19
  ::CR_HASTE_SPELL = 20
  ::CR_WEAPON_SKILL_MAINHAND = 21
  ::CR_WEAPON_SKILL_OFFHAND = 22
  ::CR_WEAPON_SKILL_RANGED = 23
  ::CR_EXPERTISE = 24
  ::CR_ARMOR_PENETRATION = 25
  * The Combat Rating formula:
  ** Percentage = Rating / RatingBase / H
  *** Level 1 to 10:  H = 2/52
  *** Level 10 to 60: H = (level-8)/52
  *** Level 60 to 70: H = 82/(262-3*level)
  *** Level 70 to 80: H = (82/52)*(131/63)^((level-70)/10)
  ::{| class="wikitable"
  !RatingID!!RatingBase
  |-
  |CR_WEAPON_SKILL||2.5
  |-
  |CR_DEFENSE_SKILL||1.5
  |-
  |CR_DODGE||12
  |-
  |CR_PARRY||15
  |-
  |CR_BLOCK||5
  |-
  |CR_HIT_MELEE||10
  |-
  |CR_HIT_RANGED||10
  |-
  |CR_HIT_SPELL||8
  |-
  |CR_CRIT_MELEE||14
  |-
  |CR_CRIT_RANGED||14
  |-
  |CR_CRIT_SPELL||14
  |-
  |CR_HIT_TAKEN_MELEE||10
  |-
  |CR_HIT_TAKEN_RANGED||10
  |-
  |CR_HIT_TAKEN_SPELL||8
  |-
  |COMBAT_RATING_RESILIENCE_CRIT_TAKEN||25
  |-
  |COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN||25   --formerly CR_CRIT_TAKEN_RANGED
  |-
  |CR_CRIT_TAKEN_SPELL||25
  |-
  |CR_HASTE_MELEE||10
  |-
  |CR_HASTE_RANGED||10
  |-
  |CR_HASTE_SPELL||10
  |-
  |CR_WEAPON_SKILL_MAINHAND||2.5
  |-
  |CR_WEAPON_SKILL_OFFHAND||2.5
  |-
  |CR_WEAPON_SKILL_RANGED||2.5
  |-
  |CR_EXPERTISE||2.5
  |-
  |CR_ARMOR_PENETRATION||4.69512176513672
  |}
  * Parry Rating, Defense Rating, Block Rating and Resilience: Low-level players will now convert these ratings into their corresponding defensive stats at the same rate as level 34 players.
Arguments:
  number - Rating value
  number - Rating ID as defined in PaperDollFrame.lua
  [optional] number - Level used in calculations. Default: player's level
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
Returns:
  ; effect : number - Effect value
  ; effect name : string - Stat ID of converted effect, ex: "DODGE", "PARRY"
Example:
  StatLogic:GetEffectFromRating(10, CR_DODGE)
  StatLogic:GetEffectFromRating(10, CR_DODGE, 70)
-----------------------------------]]

--[[ Rating ID as definded in PaperDollFrame.lua
CR_WEAPON_SKILL = 1;
CR_DEFENSE_SKILL = 2;
CR_DODGE = 3;
CR_PARRY = 4;
CR_BLOCK = 5;
CR_HIT_MELEE = 6;
CR_HIT_RANGED = 7;
CR_HIT_SPELL = 8;
CR_CRIT_MELEE = 9;
CR_CRIT_RANGED = 10;
CR_CRIT_SPELL = 11;
CR_HIT_TAKEN_MELEE = 12;
CR_HIT_TAKEN_RANGED = 13;
CR_HIT_TAKEN_SPELL = 14;
COMBAT_RATING_RESILIENCE_CRIT_TAKEN = 15; --renamed to resilience in 4.0.1   CR_CRIT_TAKEN_MELEE = 15;
COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN = 16; -- renamed to resilience in 4.0.1   CR_CRIT_TAKEN_RANGED = 16;
CR_CRIT_TAKEN_SPELL = 17;
CR_HASTE_MELEE = 18;
CR_HASTE_RANGED = 19;
CR_HASTE_SPELL = 20;
CR_WEAPON_SKILL_MAINHAND = 21;
CR_WEAPON_SKILL_OFFHAND = 22;
CR_WEAPON_SKILL_RANGED = 23;
CR_EXPERTISE = 24;
CR_ARMOR_PENETRATION = 25;
CR_MASTERY = 26;
CR_PVP_POWER = 27
--]]

-- Level 60 rating base
local RatingBase
RatingBase = {
	[CR_WEAPON_SKILL] = 2.5,
	[CR_DEFENSE_SKILL] = 1.5, --TODO: Remove Defense, since defense was removed
	[CR_DODGE] = 13.8, --Confirmed 20120905 with level 60 paladin. 13.8  (36 -> 1.74%)
	[CR_PARRY] = 13.8, --Confirmed 20120905 with level 60 paladin. 13.8  (36 -> 1.74%), 90->6.52%
	[CR_BLOCK] = 6.9,
	[CR_HIT_MELEE] = 9.37931,  
	[CR_HIT_RANGED] = 9.37931,
	[CR_HIT_SPELL] = 8,
	[CR_CRIT_MELEE] = 14, --confirmed Level 60 DK.   100->7.14%,138->9.86%  20120709  5.0.4  10016  
	[CR_CRIT_RANGED] = 14,
	[CR_CRIT_SPELL] = 14,
	[CR_HIT_TAKEN_MELEE] = 10, -- hit avoidance
	[CR_HIT_TAKEN_RANGED] = 10,
	[CR_HIT_TAKEN_SPELL] = 8,
	[COMBAT_RATING_RESILIENCE_CRIT_TAKEN] = 0, --resilience no longer reduces crits taken
	[COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN] = 7.96417713165283,
	[CR_CRIT_TAKEN_SPELL] = 28.75,
	[CR_HASTE_MELEE] = 10, -- changed in 2.2.   Confirmed 21120905 with level 60 paladin. 10
	[CR_HASTE_RANGED] = 10, -- changed in 2.2
	[CR_HASTE_SPELL] = 10, -- changed in 2.2
	[CR_WEAPON_SKILL_MAINHAND] = 2.5,
	[CR_WEAPON_SKILL_OFFHAND] = 2.5,
	[CR_WEAPON_SKILL_RANGED] = 2.5,
	[CR_EXPERTISE] = 8.00000, --8, to five decimal places.  Verified 20120905  5.0.4  16016. Level 60 Warrior  14 -> 1.75%
	[CR_ARMOR_PENETRATION] = 4.69512176513672 / 1.1, -- still manually calculated cause its still 4.69 in dbc
	[CR_MASTERY] = 14, --Verified 20120905  5.0.4  16016.
	[CR_PVP_POWER] = 14.69, --TODO: Figure out the base "PVP Power" for a level 60. Just because i typed in 14.69 doesn't mean it really is 14.69. It just means i like 69.
}
--if wowBuildNo < 13914 then
	--legacy values that nobody on earth uses anymore
	--RatingBase[COMBAT_RATING_RESILIENCE_CRIT_TAKEN] = 9.58333301544189;
	--RatingBase[COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN] = 9.58333301544189;
--end
local Level34Ratings
Level34Ratings = {
	[CR_DEFENSE_SKILL] = true, --TODO: Remove defense, since defense was removed
	[CR_DODGE] = true,
	[CR_PARRY] = true,
	[CR_BLOCK] = true,
	[COMBAT_RATING_RESILIENCE_CRIT_TAKEN] = true,
	[COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN] = true,
	[CR_CRIT_TAKEN_SPELL] = true,
}

-- 80-90 H data - Combat ratings adjustment factors; we start with level 60 values, then divide by these constants
--TODO: Add combat ratings adjustment factors for level 86-90
local H = {
  [80] = 3.2789989471436,
  [81] = 4.3056015014648,
  [82] = 5.6539749145508,
  [83] = 7.4275451660156,
  [84] = 9.7527236938477,
  [85] = 12.8057159423828, --Verified 20120905 5.0.4 16016   value1=2234/12.460951805115000/14, value2=1977/7.458172798156700/13.8/1.5, value3=2357/8.891711235046400/13.8/1.5
  [86] = 16.8186864597624, --TODO: figure out adjustment factor for level 86
  [87] = 22.0870256249850, --TODO: figure out adjustment factor for level 87
  [88] = 29.0056362086221, --TODO: figure out adjustment factor for level 88
  [89] = 38.0914545105255, --TODO: figure out adjustment factor for level 89
  [90] = 50.0233436112710, --TODO: figure out adjustment factor for level 90
  [91] = 1, --TODO: figure out adjustment factor
  [92] = 1, --TODO: figure out adjustment factor
  [93] = 1, --TODO: figure out adjustment factor
  [94] = 1, --TODO: figure out adjustment factor
  [95] = 1, --TODO: figure out adjustment factor
  [96] = 1, --TODO: figure out adjustment factor
  [97] = 1, --TODO: figure out adjustment factor
  [98] = 1, --TODO: figure out adjustment factor
  [99] = 1, --TODO: figure out adjustment factor
  [100] = 1, --TODO: figure out adjustment factor
}
-- 80-90 H data for resilience - Combat ratings adjustment factors for resiliance; we start with level 60 combat rating values, then divide by these constants
--TODO: Add combat rating resilience adjustment factors for level 86-90
local H_Resilience = {
	[80] = 3.278999106,
	[81] = 4.092896174,
	[82] = 5.108814708,
	[83] = 6.376899732,
	[84] = 7.959742271,
	[85] = 9.935470247,
	[86] = 12.40160395, --TODO: figure out adjustment factor for level 86
	[87] = 15.47986911, --TODO: figure out adjustment factor for level 87
	[88] = 19.32220610, --TODO: figure out adjustment factor for level 88
	[89] = 24.11826910, --TODO: figure out adjustment factor for level 89
	[90] = 30.10478726, --TODO: figure out adjustment factor for level 90
    [91] = 1, --TODO: figure out adjustment factor
	[92] = 1, --TODO: figure out adjustment factor
	[93] = 1, --TODO: figure out adjustment factor
	[94] = 1, --TODO: figure out adjustment factor
	[95] = 1, --TODO: figure out adjustment factor
	[96] = 1, --TODO: figure out adjustment factor
	[97] = 1, --TODO: figure out adjustment factor
	[98] = 1, --TODO: figure out adjustment factor
	[99] = 1, --TODO: figure out adjustment factor
	[100] = 1, --TODO: figure out adjustment factor
}

local combatRatingBonus = {
	[85] =
		{
			[CR_MASTERY] =   179.280045786420, --179.2 Rating / Mastery %   (i.e. 0.00557786559911594 Mastery% per Rating)
			[CR_EXPERTISE] = 102.445737393460, --102.4 Rating / Expertise % (i.e. 0.00976126508962818 Expertise% per Rating)
			--Dodge and Parry from a single regression of 232 data points. Intercept===0 (0 rating gives 0 bonus)
			[CR_PARRY] =     265.078336595056, --265.1 Rating / Parry %     (i.e. 0.00377246979174761 Parry% per Rating)
			[CR_DODGE] =     265.078336595056, --265.1 Rating / Dodge %     (i.e. 0.00377246972365043 Dodge% per Rating)
		},
}

--[[
3.1.0
- Armor Penetration Rating: All classes now receive 25% more benefit from Armor Penetration Rating.
--]]

-- Formula reverse engineered by Whitetooth (hotdogee [at] gmail [dot] com)
-- Percentage = Rating / RatingBase / H
--
-- Level 1 to 10:  H = 2/52
-- Level 10 to 60: H = (level-8)/52
-- Level 60 to 70: H = 82/(262-3*level)
-- Level 70 to 80: H = (82/52)*(131/63)^((level-70)/10)
--
--  Parry Rating, Defense Rating, Block Rating and Resilience: Low-level players
--   will now convert these ratings into their corresponding defensive
--   stats at the same rate as level 34 players.
--  Dodge Rating too

--- Calculates a combat rating's effect on its corresponding stat.
-- For a given combat rating value, this function will calculate the resulting effect on the
-- corresponding combat stat. 
-- @usage GetEffectFromRating(rating, id, [level, [class]])
-- @param rating (number) Combat rating value, e.g. 871
-- @param id (number) The Combat Rating ID to calculate the stat effect of. e.g. CR_PARRY
--    Can be one of the following values (defined in FrameXML\PaperDollFrame.lua):
--       CR_WEAPON_SKILL = 1;
--       CR_DEFENSE_SKILL = 2;
--       CR_DODGE = 3;
--       CR_PARRY = 4;
--       CR_BLOCK = 5;
--       CR_HIT_MELEE = 6;
--       CR_HIT_RANGED = 7;
--       CR_HIT_SPELL = 8;
--       CR_CRIT_MELEE = 9;
--       CR_CRIT_RANGED = 10;
--       CR_CRIT_SPELL = 11;
--       CR_HIT_TAKEN_MELEE = 12;
--       CR_HIT_TAKEN_RANGED = 13;
--       CR_HIT_TAKEN_SPELL = 14;
--       COMBAT_RATING_RESILIENCE_CRIT_TAKEN = 15;
--       COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN = 16;
--       CR_CRIT_TAKEN_SPELL = 17;
--       CR_HASTE_MELEE = 18;
--       CR_HASTE_RANGED = 19;
--       CR_HASTE_SPELL = 20;
--       CR_WEAPON_SKILL_MAINHAND = 21;
--       CR_WEAPON_SKILL_OFFHAND = 22;
--       CR_WEAPON_SKILL_RANGED = 23;
--       CR_EXPERTISE = 24;
--       CR_ARMOR_PENETRATION = 25;
--       CR_MASTERY = 26;
--       CR_PVP_POWER = 27;
-- @param level [optional] number - The level of the player being calculated. e.g. CR_PARRY
--        If omitted, the player's current level will be used, i.e. UnitLevel("player")
-- @param class [optional] number - A constant indicating the class of the player being calculated.
--        If omitted, the player's current class will be used. Can be one of the following values:
--            WARRIOR = 1
--            PALADIN = 2
--            HUNTER = 3
--            ROGUE = 4
--            PRIEST = 5
--            DEATHKNIGHT = 6
--            SHAMAN = 7
--            MAGE = 8
--            WARLOCK = 9
--            MONK = 10
--            DRUID = 11
-- @return Returns effect, effectName
--     effect : number - Effect value as a percentage e.g. 6.47162208
--     effectName : string - Stat ID of converted effect, e.g. "PARRY"
--
-- Example - Get the effect of 871 Parry Rating on Parry for a level 84 Paladin:
--        6.47162208, "PARRY" = GetEffectFromRating(871, CR_PARRY, 84, 2)
function StatLogic:GetEffectFromRating(rating, id, level, class)
	self:argCheck(rating, 1, "number", "string");
	self:argCheck(id, 2, "number", "string");
	self:argCheck(level, 3, "nil", "number");
	self:argCheck(class, 4, "nil", "number", "string");

	-- if id is stringID then convert to numberID
	if type(id) == "string" and RatingNameToID[id] then
		id = RatingNameToID[id]
	end
  
	-- check for invalid input
	if type(rating) ~= "number" or (id < 1) or (id > CR_MAX) then return 0 end
  
	-- defaults to player level if not given
	level = level or UnitLevel("player")
  
	-- argCheck for invalid input
	self:argCheck(class, 5, "nil", "string", "number")
  
	-- if class is a class string, convert to class id
	if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
	    class = ClassNameToID[strupper(class)]
	-- if class is invalid input, default to player class
	elseif type(class) ~= "number" or class < 1 or class > 10 then
	    class = ClassNameToID[playerClass]
	end
  
	--2.4.3  Parry Rating and Block Rating: Low-level players
	--   will now convert these ratings into their corresponding defensive
	--   stats at the same rate as level 34 players.
	if level < 34 and Level34Ratings[id] then
	    level = 34
	end
  
	local combatEffect;
  
	--20120804 - i don't know the combat ratings above 85 yet; so i'll use the values *AT* 85 instead of crashing
	--TODO: Update H[] and H_Resilience for 86-90
	--20120820 - i made up numbers that nobody would ever be able to distinguish from reality
	--			So we'll use them until Torhal finds something else to complain about
	--if (level > 85) then
		--print("TODO: LibStatLogic:GetEffectFromRating() being called with level "..level..". Assuming level 85 instead");
		--level = 85;
	--end
  
	--debugPrint(string.format("rating=%d, id=%d", rating, id));
	--debugPrint(string.format("RatingBase=%d", RatingBase[id]));
	--debugPrint(string.format("H[level %d]=%d", H[level]));
  
	local combatUnitRatingBase = RatingBase[id];
	
	--Resilience no longer affects crits taken; so the rating base is zero.
	--But i don't want the division by zero.
	if ((combatUnitRatingBase == 0) and (id == COMBAT_RATING_RESILIENCE_CRIT_TAKEN)) then
		return 0, RatingIDToConvertedStat[id]; --e.g. 0, "adslfkjasd;lfjas"
	end;
	
	
	assert(combatUnitRatingBase ~= nil, string.format('RatingBase[id=%d] is nil', id));
	assert(combatUnitRatingBase ~= 0, string.format('RatingBase[id=%d] is 0', id));	
	
	if (level >= 80 and level <= 85) then
		local hvalue;
	    if id == COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN then --16 = COMBAT_RATING_RESILIENCE_PLAYER_DAMAGE_TAKEN
			hvalue = H_Resilience[level];
			assert(hvalue ~= nil, string.format("H_Resilience[level=%d] is nil", level));
			assert(hvalue ~= 0, string.format("H_Resilience[level=%d] is zero", level));
		else
			hvalue = H[level];
			assert(hvalue ~= nil, string.format("H[level=%d] is nil", level));
			assert(hvalue ~= 0, string.format("H[level=%d] is zero", level));
		end
		
		if (id == CR_DODGE) or (id == CR_PARRY) then
			--Dodge and parry use a higher scaling factor.
			--Level 85.  12.80571 --> 19.20857  (for dodge and parry only), a factor of 1.5 larger
			hvalue = hvalue*1.5; 
		end;

		combatEffect = rating/combatUnitRatingBase/hvalue;
	elseif level >= 70 then
	    combatEffect = rating/combatUnitRatingBase/((82/52)*(131/63)^((level-70)/10));
	elseif level >= 60 then
	    combatEffect = rating/combatUnitRatingBase/(82/(262-3*level));
	elseif level >= 10 then
	    combatEffect = rating/combatUnitRatingBase/((level-8)/52);
	else
	    combatEffect = rating/combatUnitRatingBase/(2/52);
	end
	
	return combatEffect, RatingIDToConvertedStat[id]; --e.g. 6.47162208, "PARRY"
end


--[[---------------------------------
  :GetAPPerStr([class])
-------------------------------------
Notes:
  * Returns the attack power per strength for given class.
  * Player level does not effect attack power per strength.
Arguments:
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
Returns:
  ; ap : number - Attack power per strength
  ; statid : string - "AP"
Example:
  local ap, statid = StatLogic:GetAPPerStr()
  local ap, statid = StatLogic:GetAPPerStr("WARRIOR")
-----------------------------------]]

local APPerStr = {
	2, --WARRIOR
	2, --PALADIN
	1, --HUNTER
	1, --ROGUE
	2, --PRIEST
	2, --DEATHKNIGHT
	1, --SHAMAN
	2, --MAGE
	2, --WARLOCK
	1, --MONK   todo: 20120811: Figure out Monk's Attack Power per Strength. (Does AP/Str even vary by class anymore?)
	1, --DRUID
}
if wowBuildNo < 14333 then --4.2.0
	APPerStr[ClassNameToID["DRUID"]] = 1; -- pre 4.2.0 druid's were 2 Attack Power per Strength
end

function StatLogic:GetAPPerStr(class)
	-- argCheck for invalid input
	self:argCheck(class, 2, "nil", "string", "number")
	
	-- if class is a class string, convert to class id
	if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
		class = ClassNameToID[strupper(class)]
	elseif type(class) ~= "number" or class < 1 or class > 10 then
		-- if class is invalid input, default to player class
		class = ClassNameToID[playerClass]
	end
	
	return APPerStr[class], "AP"
end


--[[---------------------------------
  :GetAPFromStr(str, [class])
-------------------------------------
Description:
  * Calculates the attack power from strength for given class.
Arguments:
  number - Strength
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
Returns:
  ; ap : number - Attack power
  ; statid : string - "AP"
Examples:
  local ap = StatLogic:GetAPFromStr(1) -- GetAPPerStr
  local ap = StatLogic:GetAPFromStr(10)
  local ap = StatLogic:GetAPFromStr(10, "WARRIOR")
-----------------------------------]]
function StatLogic:GetAPFromStr(str, class)
	-- argCheck for invalid input
	self:argCheck(str, 2, "number")
	self:argCheck(class, 3, "nil", "string", "number")
	-- if class is a class string, convert to class id
	if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
		class = ClassNameToID[strupper(class)]
	-- if class is invalid input, default to player class
	elseif type(class) ~= "number" or class < 1 or class > 11 then
		class = ClassNameToID[playerClass]
	end
	
	-- Calculate
	return str * APPerStr[class], "AP"
end


--[[---------------------------------
  :GetAPPerAgi([class])
-------------------------------------
Notes:
  * Gets the attack power per agility for given class.
  * Player level does not effect attack power per agility.
  * Will check for Cat Form.
Arguments:
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
Returns:
  ; ap : number - Attack power per agility
  ; statid : string - "AP"
Example:
  local apPerAgi = StatLogic:GetAPPerAgi()
  local apPerAgi = StatLogic:GetAPPerAgi("ROGUE")
-----------------------------------]]

local APPerAgi = {
	0, --WARRIOR
	0, --PALADIN
	1, --HUNTER
	2, --ROGUE
	0, --PRIEST
	0, --DEATHKNIGHT
	2, --SHAMAN
	0, --MAGE
	0, --WARLOCK
	0, --MONK  TODO: 20120811: Figure out Monk's AP per AGI (do they even have any? Does AP/Agi ever vary by class anymore?)
	0, --DRUID
}

function StatLogic:GetAPPerAgi(class)
	-- argCheck for invalid input
	self:argCheck(class, 2, "nil", "string", "number")
	-- if class is a class string, convert to class id
	if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
		class = ClassNameToID[strupper(class)]
		-- if class is invalid input, default to player class
	elseif type(class) ~= "number" or class < 1 or class > 10 then
		class = ClassNameToID[playerClass]
	end
  
	-- Check druid cat/bear form
	--11 = DRUID = ClassNameToID["DRUID"]
	if (class == 11) and (PlayerHasAura((GetSpellInfo(768))) or PlayerHasAura((GetSpellInfo(5487)))) then
		if toc >= 40000 then
			return 2
		else
			return 1
		end
	end
	return APPerAgi[class], "AP"
end


--[[---------------------------------
  :GetAPFromAgi(agi, [class])
-------------------------------------
Notes:
  * Calculates the attack power from agility for given class.
Arguments:
  number - Agility
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
Returns:
  ; ap : number - Attack power
  ; statid : string - "AP"
Example:
  local ap = StatLogic:GetAPFromAgi(1) -- GetAPPerAgi
  local ap = StatLogic:GetAPFromAgi(10)
  local ap = StatLogic:GetAPFromAgi(10, "WARRIOR")
-----------------------------------]]

function StatLogic:GetAPFromAgi(agi, class)
	-- argCheck for invalid input
	self:argCheck(agi, 2, "number")
	self:argCheck(class, 3, "nil", "string", "number")
	-- if class is a class string, convert to class id
	if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
		class = ClassNameToID[strupper(class)]
		-- if class is invalid input, default to player class
	elseif type(class) ~= "number" or class < 1 or class > 10 then
		class = ClassNameToID[playerClass]
	end
	-- Calculate
	return agi * APPerAgi[class], "AP"
end


--[[---------------------------------
  :GetRAPPerAgi([class])
-------------------------------------
Notes:
  * Gets the ranged attack power per agility for given class.
  * Player level does not effect ranged attack power per agility.
Arguments:
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
Returns:
  ; rap : number - Ranged attack power per agility
  ; statid : string - "RANGED_AP"
Example:
  local rapPerAgi = StatLogic:GetRAPPerAgi()
  local rapPerAgi = StatLogic:GetRAPPerAgi("HUNTER")
-----------------------------------]]

local RAPPerAgi = {
	1,	--Warrior
	0,	--Paladin
	2,	--Hunter
	1,	--Rogue
	0,	--Priest
	0,	--Death Knight
	0,	--Shaman
	0,	--Mage
	0,	--Warlock
	0,	--Monk
	0,	--Druid
}

function StatLogic:GetRAPPerAgi(class)
  -- argCheck for invalid input
  self:argCheck(class, 2, "nil", "string", "number")
  -- if class is a class string, convert to class id
  if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
    class = ClassNameToID[strupper(class)]
  -- if class is invalid input, default to player class
  elseif type(class) ~= "number" or class < 1 or class > 10 then
    class = ClassNameToID[playerClass]
  end
  return RAPPerAgi[class], "RANGED_AP"
end


--[[---------------------------------
  :GetRAPFromAgi(agi, [class])
-------------------------------------
Notes:
  * Calculates the ranged attack power from agility for given class.
Arguments:
  number - Agility
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
Returns:
  ; rap : number - Ranged attack power
  ; statid : string - "RANGED_AP"
Example:
  local rap = StatLogic:GetRAPFromAgi(1) -- GetRAPPerAgi
  local rap = StatLogic:GetRAPFromAgi(10)
  local rap = StatLogic:GetRAPFromAgi(10, "WARRIOR")
-----------------------------------]]

function StatLogic:GetRAPFromAgi(agi, class)
  -- argCheck for invalid input
  self:argCheck(agi, 2, "number")
  self:argCheck(class, 3, "nil", "string", "number")
  -- if class is a class string, convert to class id
  if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
    class = ClassNameToID[strupper(class)]
  -- if class is invalid input, default to player class
  elseif type(class) ~= "number" or class < 1 or class > 10 then
    class = ClassNameToID[playerClass]
  end
  -- Calculate
  return agi * RAPPerAgi[class], "RANGED_AP"
end


--[[---------------------------------
  :GetBaseDodge([class])
-------------------------------------
Notes:
  * BaseDodge values derived by Whitetooth (hotdogee [at] gmail [dot] com)
  * Gets the base dodge percentage for given class.
  * Base dodge is the amount of dodge you have with 0 Agility, independent of level.
Arguments:
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
Returns:
  ; dodge : number - Base dodge in percentages
  ; statid : string - "DODGE"
Example:
  local baseDodge = StatLogic:GetBaseDodge()
  local baseDodge = StatLogic:GetBaseDodge("WARRIOR")
-----------------------------------]]
-- local BaseDodge declared at StatLogic:GetDodgeChanceBeforeDR()
-- Numbers derived by Whitetooth (hotdogee [at] gmail [dot] com)
BaseDodge = {
	5,					--Warrior
	5,					--Paladin
	-5.4499989748001,	--Hunter
	-0.590,				--Rogue
	3.1830,				--Priest
	5,					--Death Knight
	1.6750,				--Shaman
	3.4575,				--Mage
	2.0350,				--Warlock
	0,					--Monk  TODO: Get real base value for Monk (i only put zero because it's not unreasonable)
	4.9510,				--Druid
}
if wowBuildNo < 14333 then -- 4.2.0
	BaseDodge = {
		3.7580,				--Warrior
		3.6520,				--Paladin
		-5.4499989748001,	--Hunter
		-0.590,				--Rogue
		3.1830,				--Priest
		3.6640,				--Death Knight
		1.6750,				--Shaman
		3.4575,				--Mage
		2.0350,				--Warlock
		4.9510,				--Druid
	}
end

function StatLogic:GetBaseDodge(class)
	-- argCheck for invalid input
	self:argCheck(class, 2, "nil", "string", "number")
	-- if class is a class string, convert to class id
	if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
		class = ClassNameToID[strupper(class)]
		-- if class is invalid input, default to player class
	elseif type(class) ~= "number" or class < 1 or class > 10 then
		class = ClassNameToID[playerClass]
	end
	return BaseDodge[class], "DODGE"
end

--[[---------------------------------
  :GetDodgePerAgi()
-------------------------------------
Arguments:
  None
Returns:
  ; dodge : number - Dodge percentage per agility
  ; statid : string - "DODGE"
Notes:
  * Formula by Whitetooth (hotdogee [at] gmail [dot] com)
  * Calculates the dodge percentage per agility for your current class and level.
  * Only works for your currect class and current level, does not support class and level args.
  * Calculations got a bit more complicated with the introduction of the avoidance DR in WotLK, these are the values we know or can be calculated easily:
  ** D'=Total Dodge% after DR
  ** D_r=Dodge from Defense and Dodge Rating before DR
  ** D_b=Dodge unaffected by DR (BaseDodge + Dodge from talent/buffs + Lower then normal defense correction)
  ** A=Total Agility
  ** A_b=Base Agility (This is what you have with no gear on)
  ** A_g=Total Agility - Base Agility
  ** Let d be the Dodge/Agi value we are going to calculate.

  #  1     1     k
  # --- = --- + ---
  #  x'    c     x

  # x'=D'-D_b-A_b*d
  # x=A_g*d+D_r

  # 1/(D'-D_b-A_b*d)=1/C_d+k/(A_g*d+D_r)=(A_g*d+D_r+C_d*k)/(C_d*A_g*d+C_d*D_r)

  # C_d*A_g*d+C_d*D_r=[(D'-D_b)-A_b*d]*[Ag*d+(D_r+C_d*k)]

  # After rearranging the terms, we get an equation of type a*d^2+b*d+c where
  # a=-A_g*A_b
  # b=A_g(D'-D_b)-A_b(D_r+C_d*k)-C_dA_g
  # c=(D'-D_b)(D_r+C_d*k)-C_d*D_r
  ** Dodge/Agi=(-b-(b^2-4ac)^0.5)/(2a)
Example:
  local dodge, statid = StatLogic:GetDodgePerAgi()
-----------------------------------]]

local DodgePerAgiStatic = {
	0,			--"WARRIOR" 1
	0,			--"PALADIN" 2
	0.0133266,	--"HUNTER" 3
	0.0240537,	--"ROGUE" 4
	0.0192366,	--"PRIEST" 5
	0,			--"DEATHKNIGHT" 6
	0.0192366,	--"SHAMAN" 7
	0.0195253,	--"MAGE" 8
	0.0192366,	--"WARLOCK" 9
	0,			--"MONK" 10: TODO: Get *real* dodge/agi static (i just put in zero because it's not unreasonable)
	0.0240458,	--"DRUID" 11
}


local ZeroDodgePerAgiClasses = {
	["WARRIOR"] = true,
	["PALADIN"] = true,
	["DEATHKNIGHT"] = true,
}

if wowBuildNo < 14333 then --4.2.0 Rage of the Firelands
	DodgePerAgiStatic = {
		0.0135962, 0.0192366, 0.0133266, 0.0240537, 0.0192366, 0.0135962, 0.0192366, 0.0195253, 0.0192366, 0, 0.0240458,
		--["WARRIOR"] =     0.0135962,
		--["PALADIN"] =     0.0192366,
		--["HUNTER"] =      0.0133266,
		--["ROGUE"] =       0.0240537,
		--["PRIEST"] =      0.0192366,
		--["DEATHKNIGHT"] = 0.0135962,
		--["SHAMAN"] =      0.0192366,
		--["MAGE"] =        0.0195253,
		--["WARLOCK"] =     0.0192366,
		--["DRUID"] =       0.0240458,
	}
	ZeroDodgePerAgiClasses = {};
end

local ModAgiClasses = {
	["DRUID"] = true,
	["HUNTER"] = true,
	["ROGUE"] = true,
	["SHAMAN"] = true,
}
local BoK = GetSpellInfo(20217);
function StatLogic:GetDodgePerAgi()
	local level = UnitLevel("player");
	local class = ClassNameToID[playerClass];
	if level >= 80 and DodgePerAgiStatic[class] then
		return DodgePerAgiStatic[class], "DODGE";
	end
	if ZeroDodgePerAgiClasses[playerClass] then
		return 0, "DODGE";
	end

	-- Collect data
	local D_dr = GetDodgeChance();
	local dodgeFromDodgeRating = self:GetEffectFromRating(GetCombatRating(CR_DODGE), CR_DODGE, level);
	--local baseDefense, modDefense = level * 5, 0
	--local dodgeFromModDefense = modDefense * 0.04
	local D_r = dodgeFromDodgeRating; -- + dodgeFromModDefense
	local D_b = BaseDodge[class] + self:GetStatMod("ADD_DODGE"); -- + (baseDefense - level * 5) * 0.04
	local stat, effectiveStat, posBuff, negBuff = UnitStat("player", 2) -- 2 = Agility

	-- Talents that modify AGI will not add to posBuff, so we need to calculate baseAgi
	-- But Kings added Agi will add to posBuff, so we need to check for Kings
	local modAgi = 1
	if ModAgiClasses[playerClass] then
		modAgi = self:GetStatMod("MOD_AGI", nil, nil, "CLASS")
	end
	local A = effectiveStat
	local A_b = ceil((stat - posBuff - negBuff) / modAgi)
	local A_g = A - A_b
	local C = C_d[class]
	local k = K[class]
	-- Solve a*x^2+b*x+c
	local a = -A_g*A_b
	local b = A_g*(D_dr-D_b)-A_b*(D_r+C*k)-C*A_g
	local c = (D_dr-D_b)*(D_r+C*k)-C*D_r
	--RatingBuster:Print(a, b, c, D_b, D_r, A_b, A_g, C, k)
	local dodgePerAgi = (-b-(b^2-4*a*c)^0.5)/(2*a)
	if a == 0 then
		dodgePerAgi = -c / b
	end
	--return dodgePerAgi
	--return floor(dodgePerAgi*10000+0.5)/10000, "DODGE"
	return dodgePerAgi, "DODGE"
end

--[[---------------------------------
  :GetDodgeFromAgi(agi)
-------------------------------------
Notes:
  * Calculates the dodge chance from agility for your current class and level.
  * Only works for your currect class and current level, does not support class and level args.
Arguments:
  number - Agility
Returns:
  ; dodge : number - Dodge percentage
  ; statid : string - "DODGE"
Example:
  local dodge = StatLogic:GetDodgeFromAgi(1) -- GetDodgePerAgi
  local dodge = StatLogic:GetDodgeFromAgi(10)
-----------------------------------]]

function StatLogic:GetDodgeFromAgi(agi)
  -- argCheck for invalid input
  self:argCheck(agi, 2, "number")
  -- Calculate
  return agi * self:GetDodgePerAgi(), "DODGE"
end


--[[---------------------------------
  :GetCritFromAgi(agi, [class], [level])
-------------------------------------
Notes:
  * CritPerAgi values reverse engineered by Whitetooth (hotdogee [at] gmail [dot] com)
  * Calculates the melee/ranged crit chance from agility for given class and level.
Arguments:
  number - Agility
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
  [optional] number - Level used in calculations. Default: player's level
Returns:
  ; crit : number - Melee/ranged crit percentage
  ; statid : string - "MELEE_CRIT"
Example:
  local crit = StatLogic:GetCritFromAgi(1) -- GetCritPerAgi
  local crit = StatLogic:GetCritFromAgi(10)
  local crit = StatLogic:GetCritFromAgi(10, "WARRIOR")
  local crit = StatLogic:GetCritFromAgi(10, nil, 70)
  local crit = StatLogic:GetCritFromAgi(10, "WARRIOR", 70)
-----------------------------------]]

-- Numbers reverse engineered by Whitetooth (hotdogee [at] gmail [dot] com)
local CritPerAgi = {
	--TODO: Add monk values; since i'm too sure they're not zero
	--WARRIOR, PALADIN, HUNTER, ROGUE, PRIEST, DEATHKNIGHT, SHAMAN, MAGE, WARLOCK, MONK, DRUID
	[1] =   {0.24999999, 0.21740000, 0.28400000, 0.44760900, 0.09117500, 0.24999999, 0.10390000, 0.07730000, 0.11890000, 0, 0.12622500, },
	[2] =   {0.23809499, 0.20704800, 0.28343401, 0.42895800, 0.09117500, 0.23809499, 0.10390000, 0.07730000, 0.11890000, 0, 0.12622500, },
	[3] =   {0.23809499, 0.20704800, 0.27107401, 0.41180002, 0.09117500, 0.23809499, 0.09895239, 0.07730000, 0.11323800, 0, 0.12021400, },
	[4] =   {0.22727300, 0.19763601, 0.25299001, 0.38129601, 0.08683330, 0.22727300, 0.09895239, 0.07361900, 0.11323800, 0, 0.12021400, },
	[5] =   {0.21739099, 0.19763601, 0.24300499, 0.36767900, 0.08683330, 0.21739099, 0.09445450, 0.07361900, 0.11323800, 0, 0.11475000, },
	[6] =   {0.20833299, 0.18904300, 0.23373601, 0.35500000, 0.08683330, 0.20833299, 0.09445450, 0.07361900, 0.10809100, 0, 0.11475000, },
	[7] =   {0.20833299, 0.18904300, 0.22510700, 0.33209701, 0.08683330, 0.20833299, 0.09445450, 0.07361900, 0.10809100, 0, 0.10976100, },
	[8] =   {0.20000001, 0.18116700, 0.21705499, 0.32171900, 0.08288640, 0.20000001, 0.09034780, 0.07361900, 0.10809100, 0, 0.10976100, },
	[9] =   {0.19230800, 0.18116700, 0.20513700, 0.31196999, 0.08288640, 0.19230800, 0.09034780, 0.07361900, 0.10339100, 0, 0.10518699, },
	[10] =  {0.19230800, 0.17392000, 0.19836400, 0.29414301, 0.08288640, 0.19230800, 0.08658330, 0.07027270, 0.10339100, 0, 0.09709620, },
	[11] =  {0.18518500, 0.17392000, 0.18475300, 0.26397400, 0.08288640, 0.18518500, 0.08658330, 0.07027270, 0.09908330, 0, 0.09350000, },
	[12] =  {0.17857100, 0.16723101, 0.16697299, 0.23941901, 0.07928260, 0.17857100, 0.08312000, 0.07027270, 0.09908330, 0, 0.09350000, },
	[13] =  {0.16666700, 0.15528600, 0.15469900, 0.21447900, 0.07928260, 0.16666700, 0.08312000, 0.07027270, 0.09908330, 0, 0.09016070, },
	[14] =  {0.16129001, 0.15528600, 0.14408800, 0.19798100, 0.07928260, 0.16129001, 0.07992310, 0.07027270, 0.09588690, 0, 0.09016070, },
	[15] =  {0.15625000, 0.14493300, 0.13299300, 0.17750000, 0.07928260, 0.15625000, 0.07696300, 0.06721740, 0.09435100, 0, 0.08415000, },
	[16] =  {0.15151500, 0.14493300, 0.12666400, 0.16604800, 0.07597920, 0.15151500, 0.07421430, 0.06721740, 0.09284450, 0, 0.08415000, },
	[17] =  {0.14705900, 0.14025799, 0.11942200, 0.15598499, 0.07597920, 0.14705900, 0.07421430, 0.06721740, 0.09136670, 0, 0.08143550, },
	[18] =  {0.13888900, 0.13175800, 0.11166499, 0.14500000, 0.07597920, 0.13888900, 0.07165520, 0.06721740, 0.08991690, 0, 0.07889060, },
	[19] =  {0.13513500, 0.13175800, 0.10597800, 0.13546100, 0.07294000, 0.13513500, 0.07165520, 0.06721740, 0.08849440, 0, 0.07889060, },
	[20] =  {0.12820500, 0.12422900, 0.09980580, 0.12709900, 0.07294000, 0.12820500, 0.06703230, 0.06441670, 0.08709860, 0, 0.07012500, },
	[21] =  {0.12820500, 0.12077800, 0.09615530, 0.11970900, 0.07294000, 0.12820500, 0.06703230, 0.06441670, 0.08572890, 0, 0.07012500, },
	[22] =  {0.12500000, 0.12077800, 0.09103050, 0.11438900, 0.07294000, 0.12500000, 0.06493750, 0.06441670, 0.08438450, 0, 0.06822970, },
	[23] =  {0.11904800, 0.11442100, 0.08718610, 0.10836800, 0.07013460, 0.11904800, 0.06493750, 0.06441670, 0.08306500, 0, 0.06643420, },
	[24] =  {0.11627900, 0.11148700, 0.08293830, 0.10399000, 0.07013460, 0.11627900, 0.06296970, 0.06184000, 0.08176980, 0, 0.06643420, },
	[25] =  {0.11111100, 0.10870000, 0.07972230, 0.09804760, 0.07013460, 0.11111100, 0.06111760, 0.06184000, 0.08049830, 0, 0.06311250, },
	[26] =  {0.10869600, 0.10604900, 0.07674040, 0.09359090, 0.06753700, 0.10869600, 0.05937140, 0.06184000, 0.07924990, 0, 0.06311250, },
	[27] =  {0.10638300, 0.10352400, 0.07341370, 0.09030700, 0.06753700, 0.10638300, 0.05937140, 0.06184000, 0.07802420, 0, 0.06157320, },
	[28] =  {0.10204100, 0.10111600, 0.07086710, 0.08651260, 0.06753700, 0.10204100, 0.05772220, 0.06184000, 0.07682060, 0, 0.06010710, },
	[29] =  {0.10000000, 0.09881820, 0.06801100, 0.08302420, 0.06512500, 0.10000000, 0.05772220, 0.05946150, 0.07563860, 0, 0.06010710, },
	[30] =  {0.09615380, 0.09452170, 0.06537160, 0.07919230, 0.06512500, 0.09615380, 0.05468420, 0.05946150, 0.07447780, 0, 0.05488040, },
	[31] =  {0.09433960, 0.09251060, 0.06374390, 0.07682840, 0.06512500, 0.09433960, 0.05468420, 0.05946150, 0.07333770, 0, 0.05371280, },
	[32] =  {0.09259260, 0.09251060, 0.06141140, 0.07406470, 0.06287930, 0.09259260, 0.05328210, 0.05946150, 0.07221780, 0, 0.05371280, },
	[33] =  {0.08928570, 0.08873470, 0.05923970, 0.07149310, 0.06287930, 0.08928570, 0.05195000, 0.05725930, 0.07111760, 0, 0.05259370, },
	[34] =  {0.08771930, 0.08696000, 0.05754870, 0.06909400, 0.06287930, 0.08771930, 0.05195000, 0.05725930, 0.07003690, 0, 0.05152040, },
	[35] =  {0.08474580, 0.08361540, 0.05563030, 0.06641940, 0.06078330, 0.08474580, 0.04947620, 0.05725930, 0.06897500, 0, 0.05049000, },
	[36] =  {0.08333330, 0.08203770, 0.05412970, 0.06434370, 0.06078330, 0.08333330, 0.04832560, 0.05521430, 0.06793170, 0, 0.04950000, },
	[37] =  {0.08196720, 0.08203770, 0.05242280, 0.06277440, 0.06078330, 0.08196720, 0.04832560, 0.05521430, 0.06690650, 0, 0.04854810, },
	[38] =  {0.07936510, 0.07905450, 0.05081710, 0.06091720, 0.05882260, 0.07936510, 0.04722730, 0.05521430, 0.06589910, 0, 0.04854810, },
	[39] =  {0.07812500, 0.07764290, 0.04930400, 0.05916670, 0.05882260, 0.07812500, 0.04722730, 0.05521430, 0.06490910, 0, 0.04763210, },
	[40] =  {0.07575760, 0.07496550, 0.04811060, 0.05719440, 0.05882260, 0.07575760, 0.04517390, 0.05331030, 0.06393600, 0, 0.04428950, },
	[41] =  {0.07352940, 0.07369490, 0.04697090, 0.05564860, 0.05698440, 0.07352940, 0.04421280, 0.05331030, 0.06297960, 0, 0.04352590, },
	[42] =  {0.07246380, 0.07369490, 0.04566750, 0.05418420, 0.05698440, 0.07246380, 0.04421280, 0.05331030, 0.06203950, 0, 0.04352590, },
	[43] =  {0.07042250, 0.07127870, 0.04443200, 0.05279490, 0.05525760, 0.07042250, 0.04329170, 0.05331030, 0.06111540, 0, 0.04278810, },
	[44] =  {0.06944440, 0.07012900, 0.04325920, 0.05121890, 0.05525760, 0.06944440, 0.04240820, 0.05153330, 0.06020690, 0, 0.04207500, },
	[45] =  {0.06756760, 0.06793750, 0.04214440, 0.04973430, 0.05525760, 0.06756760, 0.04156000, 0.05153330, 0.05931370, 0, 0.04071770, },
	[46] =  {0.06666670, 0.06689230, 0.04125640, 0.04856130, 0.05363240, 0.06666670, 0.04074510, 0.05153330, 0.05843550, 0, 0.04007140, },
	[47] =  {0.06493510, 0.06587880, 0.04023690, 0.04744240, 0.05363240, 0.06493510, 0.03996150, 0.04987100, 0.05757210, 0, 0.04007140, },
	[48] =  {0.06329110, 0.06394120, 0.03910780, 0.04637390, 0.05210000, 0.06329110, 0.03920750, 0.04987100, 0.05672300, 0, 0.03944530, },
	[49] =  {0.06250000, 0.06301450, 0.03818660, 0.04535240, 0.05210000, 0.06250000, 0.03920750, 0.04987100, 0.05588810, 0, 0.03883850, },
	[50] =  {0.06097560, 0.06123940, 0.03730590, 0.04399570, 0.05210000, 0.06097560, 0.03778180, 0.04831250, 0.05506700, 0, 0.03658700, },
	[51] =  {0.05952380, 0.06038890, 0.03659920, 0.04307530, 0.05065280, 0.05952380, 0.03710710, 0.04831250, 0.05425950, 0, 0.03606430, },
	[52] =  {0.05882350, 0.05956160, 0.03578580, 0.04219260, 0.05065280, 0.05882350, 0.03645610, 0.04831250, 0.05346520, 0, 0.03555630, },
	[53] =  {0.05747130, 0.05797330, 0.03500600, 0.04118000, 0.04928380, 0.05747130, 0.03645610, 0.04684850, 0.05268400, 0, 0.03506250, },
	[54] =  {0.05617980, 0.05721050, 0.03413840, 0.04037250, 0.04928380, 0.05617980, 0.03582760, 0.04684850, 0.05191560, 0, 0.03506250, },
	[55] =  {0.05494510, 0.05574360, 0.03342470, 0.03944440, 0.04798680, 0.05494510, 0.03463330, 0.04684850, 0.05115970, 0, 0.03411490, },
	[56] =  {0.05434780, 0.05503800, 0.03284840, 0.03855810, 0.04798680, 0.05434780, 0.03406560, 0.04547060, 0.05041610, 0, 0.03366000, },
	[57] =  {0.05319150, 0.05435000, 0.03207870, 0.03784930, 0.04675640, 0.05319150, 0.03351610, 0.04547060, 0.04968460, 0, 0.03321710, },
	[58] =  {0.05208330, 0.05302440, 0.03144320, 0.03703240, 0.04675640, 0.05208330, 0.03351610, 0.04547060, 0.04896490, 0, 0.03278570, },
	[59] =  {0.05102040, 0.05238550, 0.03073430, 0.03637810, 0.04558750, 0.05102040, 0.03298410, 0.04417140, 0.04825680, 0, 0.03236540, },
	[60] =  {0.05000000, 0.05115290, 0.03014770, 0.03550000, 0.04558750, 0.05000000, 0.03196920, 0.04417140, 0.04756000, 0, 0.03078660, },
	[61] =  {0.04761900, 0.04940910, 0.02967120, 0.03343390, 0.04449120, 0.04761900, 0.03102490, 0.04417140, 0.04687440, 0, 0.02988460, },
	[62] =  {0.04545450, 0.04831110, 0.02903510, 0.03217750, 0.04458550, 0.04545450, 0.03040300, 0.04417140, 0.04619980, 0, 0.02950110, },
	[63] =  {0.04347830, 0.04726090, 0.02842440, 0.03063990, 0.04426580, 0.04347830, 0.02935380, 0.04294440, 0.04553590, 0, 0.02849640, },
	[64] =  {0.04166670, 0.04576840, 0.02791680, 0.02959690, 0.04344670, 0.04166670, 0.02848270, 0.04294440, 0.04488250, 0, 0.02791070, },
	[65] =  {0.04000000, 0.04482470, 0.02734930, 0.02861530, 0.04271900, 0.04000000, 0.02807010, 0.04294440, 0.04423950, 0, 0.02738060, },
	[66] =  {0.03846150, 0.04391920, 0.02695050, 0.02767470, 0.04207770, 0.03846150, 0.02734040, 0.04178380, 0.04360670, 0, 0.02690250, },
	[67] =  {0.03703700, 0.04262750, 0.02641880, 0.02680990, 0.04151880, 0.03703700, 0.02668210, 0.04178380, 0.04298380, 0, 0.02647340, },
	[68] =  {0.03571430, 0.04180770, 0.02590650, 0.02618690, 0.04133470, 0.03571430, 0.02608150, 0.04178380, 0.04237070, 0, 0.02577280, },
	[69] =  {0.03448280, 0.04101890, 0.02541250, 0.02560950, 0.04117830, 0.03448280, 0.02552060, 0.04068420, 0.04176730, 0, 0.02537310, },
	[70] =  {0.03333330, 0.04025930, 0.02499810, 0.02499880, 0.04014610, 0.03333330, 0.02499630, 0.04068420, 0.04117320, 0, 0.02499810, },
	[71] =  {0.03105590, 0.03716240, 0.02323340, 0.02323930, 0.03721430, 0.03105590, 0.02323170, 0.03770730, 0.03835480, 0, 0.02323340, },
	[72] =  {0.02873560, 0.03450790, 0.02159330, 0.02158280, 0.03440570, 0.02873560, 0.02159180, 0.03513640, 0.03549250, 0, 0.02159330, },
	[73] =  {0.02673800, 0.03220740, 0.02006910, 0.02006820, 0.03199120, 0.02673800, 0.02006760, 0.03289360, 0.03302780, 0, 0.02006910, },
	[74] =  {0.02487560, 0.02978080, 0.01865240, 0.01865040, 0.02989340, 0.02487560, 0.01865100, 0.03031370, 0.03088310, 0, 0.01865240, },
	[75] =  {0.02314810, 0.02769430, 0.01733570, 0.01733160, 0.02762880, 0.02314810, 0.01733440, 0.02810910, 0.02865060, 0, 0.01733570, },
	[76] =  {0.02145920, 0.02572780, 0.01611190, 0.01611110, 0.02568310, 0.02145920, 0.01611080, 0.02620340, 0.02642220, 0, 0.01611190, },
	[77] =  {0.01992030, 0.02389010, 0.01497460, 0.01496370, 0.02399340, 0.01992030, 0.01497350, 0.02415630, 0.02451550, 0, 0.01497460, },
	[78] =  {0.01851850, 0.02229740, 0.01391750, 0.01391220, 0.02223780, 0.01851850, 0.01391650, 0.02273530, 0.02286540, 0, 0.01391750, },
	[79] =  {0.01724140, 0.02070480, 0.01293510, 0.01293340, 0.02072160, 0.01724140, 0.01293410, 0.02089190, 0.02123210, 0, 0.01293510, },
	[80] =  {0.01602560, 0.01923890, 0.01202200, 0.01202690, 0.01919470, 0.01602560, 0.01202110, 0.01956960, 0.01981670, 0, 0.01202200, },
	[81] =  {0.01219510, 0.01463970, 0.00915552, 0.00915925, 0.01461810, 0.01220460, 0.00915486, 0.01486540, 0.01509170, 0, 0.00915552, },
	[82] =  {0.00929368, 0.01114870, 0.00697209, 0.00697493, 0.01113190, 0.00929400, 0.00697159, 0.01136760, 0.01149260, 0, 0.00697209, },
	[83] =  {0.00707214, 0.00849219, 0.00530727, 0.00530944, 0.00847380, 0.00707475, 0.00530689, 0.00863687, 0.00874836, 0, 0.00530727, },
	[84] =  {0.00538793, 0.00647024, 0.00404195, 0.00404360, 0.00645353, 0.00538804, 0.00404166, 0.00657872, 0.00666263, 0, 0.00404195, },
	[85] =  {0.00410509, 0.00492412, 0.00307831, 0.00307957, 0.00491496, 0.00410348, 0.00307809, 0.00500324, 0.00507420, 0, 0.00307831, },
	[86] =  {0.00192160, 0.00230663, 0.00144129, 0.00144188, 0.00230122, 0.00192160, 0.00144118, 0.00234598, 0.00237578, 0, 0.00144129, },
	[87] =  {0.00134916, 0.00161937, 0.00101208, 0.00101249, 0.00161592, 0.00134916, 0.00101200, 0.00164819, 0.00166828, 0, 0.00101208, },
	[88] =  {0.00094733, 0.00113733, 0.00071069, 0.00071098, 0.00113472, 0.00094733, 0.00071064, 0.00115719, 0.00117148, 0, 0.00071069, },
	[89] =  {0.00066525, 0.00079868, 0.01154730, 0.00049925, 0.03508770, 0.00066525, 0.03300000, 0.00081240, 0.03296700, 0, 0.03388430, },
	[90] =  {0.00046711, 0.00056081, 0.01136360, 0.00035058, 0.03448280, 0.00046711, 0.03235290, 0.00057048, 0.03225810, 0, 0.03333330, },
	[91] =  {0.00032802, 0.00039381, 0.01123600, 0.00024617, 0.03448280, 0.00032802, 0.03203880, 0.00040052, 0.03191490, 0, 0.03280000, },
	[92] =  {0.00023034, 0.00027652, 0.01106190, 0.00017286, 0.03389830, 0.00023034, 0.03142860, 0.00028130, 0.03157890, 0, 0.03253970, },
	[93] =  {0.00016174, 0.00019418, 0.01089320, 0.00012139, 0.03333330, 0.00016174, 0.03113210, 0.00019752, 0.03125000, 0, 0.03203130, },
	[94] =  {0.00011358, 0.00013635, 0.01075270, 0.00008524, 0.03278690, 0.00011358, 0.03084110, 0.00013869, 0.03092780, 0, 0.03178290, },
	[95] =  {0.00007976, 0.00009575, 0.01059320, 0.00005985, 0.03278690, 0.00007976, 0.03000000, 0.00009739, 0.03030300, 0, 0.03106060, },
	[96] =  {0.00005600, 0.00006723, 0.01046030, 0.00004203, 0.03225810, 0.00005600, 0.02972970, 0.00006839, 0.03000000, 0, 0.03082710, },
	[97] =  {0.00003933, 0.00004721, 0.01030930, 0.00002951, 0.03174600, 0.00003933, 0.02946430, 0.00004802, 0.02970300, 0, 0.03037040, },
	[98] =  {0.00002762, 0.00003315, 0.01016260, 0.00002072, 0.03125000, 0.00002762, 0.02894740, 0.00003372, 0.02912620, 0, 0.02992700, },
	[99] =  {0.00001939, 0.00002328, 0.01002000, 0.00001455, 0.03076920, 0.00001939, 0.02869570, 0.00002368, 0.02884620, 0, 0.02971010, },
	[100] = {0.00001362, 0.00001635, 0.00988142, 0.00001022, 0.03076920, 0.00001362, 0.02820510, 0.00001663, 0.02830190, 0, 0.02907800, },
}
function StatLogic:GetCritFromAgi(agi, class, level)
  -- argCheck for invalid input
  self:argCheck(agi, 2, "number")
  self:argCheck(class, 3, "nil", "string", "number")
  self:argCheck(level, 4, "nil", "number")
  -- if class is a class string, convert to class id
  if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
    class = ClassNameToID[strupper(class)]
  -- if class is invalid input, default to player class
  elseif type(class) ~= "number" or class < 1 or class > 10 then
    class = ClassNameToID[playerClass]
  end
  -- if level is invalid input, default to player level
  if type(level) ~= "number" or level < 1 or level > 100 then
    level = UnitLevel("player")
  end
  -- Calculate
  return agi * CritPerAgi[level][class], "MELEE_CRIT"
end

--[[---------------------------------
  :GetHealthFromSta(sta, [level])
-------------------------------------
Notes:
  * For each level beyond 80, stamina grants an additional 8% health per point. 
Arguments:
  number - Stamina
  [optional] number - Level used in calculations. Default: player's level
Returns:
  ; health : number - Health
  ; statid : string - "HEALTH"
Example:
  local health = StatLogic:GetHealthFromSta(1) -- GetHealthPerSta
  local health = StatLogic:GetHealthFromSta(10)
  local health = StatLogic:GetCritFromAgi(10, 85)
-----------------------------------]]
function StatLogic:GetHealthFromSta(sta, level)
  -- argCheck for invalid input
  self:argCheck(sta, 2, "number")
  self:argCheck(level, 3, "nil", "number")
  -- if level is invalid input, default to player level
  if type(level) ~= "number" or level < 1 or level > 100 then
    level = UnitLevel("player")
  end
  local mod = 10
  if level > 80 then
    mod = mod + (level - 80) * 0.8
  end
  -- Calculate
  return sta * mod, "HEALTH"
end


--[[---------------------------------
  :GetSpellCritFromInt(int, [class], [level])
-------------------------------------
Notes:
  * SpellCritPerInt values reverse engineered by Whitetooth (hotdogee [at] gmail [dot] com)
  * Calculates the spell crit chance from intellect for given class and level.
Arguments:
  number - Intellect
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
  [optional] number - Level used in calculations. Default: player's level
Returns:
  ; spellcrit : number - Spell crit percentage
  ; statid : string - "SPELL_CRIT"
Example:
  local spellCrit = StatLogic:GetSpellCritFromInt(1) -- GetSpellCritPerInt
  local spellCrit = StatLogic:GetSpellCritFromInt(10)
  local spellCrit = StatLogic:GetSpellCritFromInt(10, "MAGE")
  local spellCrit = StatLogic:GetSpellCritFromInt(10, nil, 70)
  local spellCrit = StatLogic:GetSpellCritFromInt(10, "MAGE", 70)
-----------------------------------]]

-- Numbers reverse engineered by Whitetooth (hotdogee [at] gmail [dot] com)
local SpellCritPerInt

SpellCritPerInt = {
  --WARRIOR, PALADIN, HUNTER, ROGUE, PRIEST, DEATHKNIGHT, SHAMAN, MAGE, WARLOCK, MONK, DRUID
  [1] =  {0.00000000, 0.08322500, 0.06990000, 0.00000000, 0.17102300, 0.00000000, 0.13328600, 0.16370000, 0.15000000, 0, 0.14311400, },
  [2] =  {0.00000000, 0.07926190, 0.06657140, 0.00000000, 0.16358700, 0.00000000, 0.12722700, 0.15740400, 0.14347800, 0, 0.13689100, },
  [3] =  {0.00000000, 0.07926190, 0.06657140, 0.00000000, 0.15677100, 0.00000000, 0.12169600, 0.15157399, 0.13750000, 0, 0.13118699, },
  [4] =  {0.00000000, 0.07565910, 0.06354550, 0.00000000, 0.15050001, 0.00000000, 0.12169600, 0.14112100, 0.13200000, 0, 0.12594000, },
  [5] =  {0.00000000, 0.07565910, 0.06354550, 0.00000000, 0.13935200, 0.00000000, 0.11662500, 0.13641700, 0.12692300, 0, 0.12109600, },
  [6] =  {0.00000000, 0.07236960, 0.06078260, 0.00000000, 0.13437500, 0.00000000, 0.11196000, 0.13201600, 0.12222200, 0, 0.11661100, },
  [7] =  {0.00000000, 0.06935420, 0.06078260, 0.00000000, 0.12974100, 0.00000000, 0.10765400, 0.12789100, 0.11785700, 0, 0.11244600, },
  [8] =  {0.00000000, 0.06935420, 0.05825000, 0.00000000, 0.12541700, 0.00000000, 0.10366699, 0.12401500, 0.11379300, 0, 0.11244600, },
  [9] =  {0.00000000, 0.06658000, 0.05825000, 0.00000000, 0.12137101, 0.00000000, 0.09996430, 0.11692900, 0.11000000, 0, 0.10856900, },
  [10] = {0.00000000, 0.06658000, 0.05592000, 0.00000000, 0.11401500, 0.00000000, 0.09996430, 0.11368101, 0.10645200, 0, 0.09839060, },
  [11] = {0.00000000, 0.06401920, 0.05592000, 0.00000000, 0.10451400, 0.00000000, 0.09330000, 0.10493600, 0.09705880, 0, 0.09260290, },
  [12] = {0.00000000, 0.06164810, 0.05376920, 0.00000000, 0.09406250, 0.00000000, 0.08746880, 0.09301140, 0.08918920, 0, 0.08509460, },
  [13] = {0.00000000, 0.05944640, 0.04992860, 0.00000000, 0.08750000, 0.00000000, 0.07997140, 0.08707450, 0.08250000, 0, 0.08073080, },
  [14] = {0.00000000, 0.05739660, 0.04992860, 0.00000000, 0.07838540, 0.00000000, 0.07564860, 0.07308040, 0.07674420, 0, 0.07496430, },
  [15] = {0.00000000, 0.05369350, 0.04660000, 0.00000000, 0.07235580, 0.00000000, 0.06997500, 0.06709020, 0.07173910, 0, 0.06844570, },
  [16] = {0.00000000, 0.05369350, 0.04660000, 0.00000000, 0.06840910, 0.00000000, 0.06664290, 0.06394530, 0.06875000, 0, 0.06559370, },
  [17] = {0.00000000, 0.05201560, 0.04509680, 0.00000000, 0.06270830, 0.00000000, 0.06361360, 0.06018380, 0.06346150, 0, 0.06173530, },
  [18] = {0.00000000, 0.04895590, 0.04236360, 0.00000000, 0.05972220, 0.00000000, 0.05955320, 0.05684030, 0.06000000, 0, 0.05940570, },
  [19] = {0.00000000, 0.04895590, 0.04236360, 0.00000000, 0.05615670, 0.00000000, 0.05712240, 0.05384870, 0.05689660, 0, 0.05622320, },
  [20] = {0.00000000, 0.04623610, 0.03994290, 0.00000000, 0.05225690, 0.00000000, 0.05382690, 0.05052470, 0.05409840, 0, 0.05161480, },
  [21] = {0.00000000, 0.04498650, 0.03883330, 0.00000000, 0.05016670, 0.00000000, 0.05183330, 0.04872020, 0.05156250, 0, 0.04997620, },
  [22] = {0.00000000, 0.04380260, 0.03883330, 0.00000000, 0.04703120, 0.00000000, 0.04998210, 0.04598310, 0.04925370, 0, 0.04770450, },
  [23] = {0.00000000, 0.04267950, 0.03678950, 0.00000000, 0.04533130, 0.00000000, 0.04744070, 0.04448370, 0.04714290, 0, 0.04630150, },
  [24] = {0.00000000, 0.04161250, 0.03584620, 0.00000000, 0.04275570, 0.00000000, 0.04588520, 0.04219070, 0.04459460, 0, 0.04372920, },
  [25] = {0.00000000, 0.03963100, 0.03495000, 0.00000000, 0.04089670, 0.00000000, 0.04373440, 0.04051980, 0.04285710, 0, 0.04198000, },
  [26] = {0.00000000, 0.03870930, 0.03409760, 0.00000000, 0.03919270, 0.00000000, 0.04240910, 0.03897620, 0.04177220, 0, 0.04088960, },
  [27] = {0.00000000, 0.03870930, 0.03328570, 0.00000000, 0.03762500, 0.00000000, 0.04116180, 0.03720450, 0.03975900, 0, 0.03935630, },
  [28] = {0.00000000, 0.03698890, 0.03251160, 0.00000000, 0.03617790, 0.00000000, 0.03942250, 0.03382230, 0.03837210, 0, 0.03839630, },
  [29] = {0.00000000, 0.03618480, 0.03177270, 0.00000000, 0.03483800, 0.00000000, 0.03834250, 0.03248020, 0.03666670, 0, 0.03661050, },
  [30] = {0.00000000, 0.03467710, 0.03039130, 0.00000000, 0.03329650, 0.00000000, 0.03682890, 0.03124050, 0.03548390, 0, 0.03459890, },
  [31] = {0.00000000, 0.03396940, 0.02974470, 0.00000000, 0.03215810, 0.00000000, 0.03543040, 0.03054100, 0.03473680, 0, 0.03385480, },
  [32] = {0.00000000, 0.03329000, 0.02974470, 0.00000000, 0.03109500, 0.00000000, 0.03455560, 0.02944240, 0.03333330, 0, 0.03245880, },
  [33] = {0.00000000, 0.03263730, 0.02853060, 0.00000000, 0.03010000, 0.00000000, 0.03332140, 0.02861890, 0.03235290, 0, 0.03180300, },
  [34] = {0.00000000, 0.03200960, 0.02796000, 0.00000000, 0.02894230, 0.00000000, 0.03254650, 0.02784010, 0.03113210, 0, 0.03086760, },
  [35] = {0.00000000, 0.03082410, 0.02688460, 0.00000000, 0.02807840, 0.00000000, 0.03144940, 0.02692430, 0.03027520, 0, 0.02970280, },
  [36] = {0.00000000, 0.03026360, 0.02637740, 0.00000000, 0.02726450, 0.00000000, 0.03042390, 0.02623400, 0.02946430, 0, 0.02915280, },
  [37] = {0.00000000, 0.02972320, 0.02637740, 0.00000000, 0.02631120, 0.00000000, 0.02977660, 0.02541930, 0.02844830, 0, 0.02836490, },
  [38] = {0.00000000, 0.02869830, 0.02541820, 0.00000000, 0.02559520, 0.00000000, 0.02885570, 0.02480300, 0.02773110, 0, 0.02761840, },
  [39] = {0.00000000, 0.02821190, 0.02496430, 0.00000000, 0.02491720, 0.00000000, 0.02827270, 0.02407350, 0.02682930, 0, 0.02691030, },
  [40] = {0.00000000, 0.02728690, 0.02410340, 0.00000000, 0.02411860, 0.00000000, 0.02717480, 0.02352010, 0.02619050, 0, 0.02559760, },
  [41] = {0.00000000, 0.02684680, 0.02369490, 0.00000000, 0.02351560, 0.00000000, 0.02665710, 0.02299160, 0.02558140, 0, 0.02518800, },
  [42] = {0.00000000, 0.02642060, 0.02369490, 0.00000000, 0.02280300, 0.00000000, 0.02615890, 0.02153950, 0.02481200, 0, 0.02440700, },
  [43] = {0.00000000, 0.02560770, 0.02291800, 0.00000000, 0.02226330, 0.00000000, 0.02544550, 0.02109540, 0.02426470, 0, 0.02403440, },
  [44] = {0.00000000, 0.02560770, 0.02254840, 0.00000000, 0.02162360, 0.00000000, 0.02476990, 0.02056530, 0.02357140, 0, 0.02332220, },
  [45] = {0.00000000, 0.02484330, 0.02184380, 0.00000000, 0.02101960, 0.00000000, 0.02412930, 0.02006130, 0.02291670, 0, 0.02281520, },
  [46] = {0.00000000, 0.02447790, 0.02150770, 0.00000000, 0.02056010, 0.00000000, 0.02352100, 0.01967550, 0.02244900, 0, 0.02232980, },
  [47] = {0.00000000, 0.02377860, 0.02118180, 0.00000000, 0.02001330, 0.00000000, 0.02313220, 0.01921360, 0.02200000, 0, 0.02186460, },
  [48] = {0.00000000, 0.02311810, 0.02055880, 0.00000000, 0.01959640, 0.00000000, 0.02257260, 0.01877290, 0.02142860, 0, 0.02141840, },
  [49] = {0.00000000, 0.02280140, 0.02026090, 0.00000000, 0.01909900, 0.00000000, 0.02203940, 0.01835200, 0.02088610, 0, 0.02085100, },
  [50] = {0.00000000, 0.02219330, 0.01969010, 0.00000000, 0.01862620, 0.00000000, 0.02153080, 0.01794960, 0.02037040, 0, 0.02018270, },
  [51] = {0.00000000, 0.02190130, 0.01941670, 0.00000000, 0.01826460, 0.00000000, 0.02104510, 0.01764010, 0.02000000, 0, 0.01980190, },
  [52] = {0.00000000, 0.02161690, 0.01915070, 0.00000000, 0.01783180, 0.00000000, 0.02073330, 0.01726790, 0.01952660, 0, 0.01931600, },
  [53] = {0.00000000, 0.02106960, 0.01864000, 0.00000000, 0.01750000, 0.00000000, 0.02013670, 0.01698130, 0.01907510, 0, 0.01908180, },
  [54] = {0.00000000, 0.02080620, 0.01839470, 0.00000000, 0.01710230, 0.00000000, 0.01985110, 0.01656880, 0.01864410, 0, 0.01863020, },
  [55] = {0.00000000, 0.02029880, 0.01792310, 0.00000000, 0.01664820, 0.00000000, 0.01930340, 0.01624010, 0.01823200, 0, 0.01819940, },
  [56] = {0.00000000, 0.02005420, 0.01769620, 0.00000000, 0.01635870, 0.00000000, 0.01904080, 0.01544340, 0.01793480, 0, 0.01788920, },
  [57] = {0.00000000, 0.01981550, 0.01747500, 0.00000000, 0.01601060, 0.00000000, 0.01866000, 0.01510150, 0.01755320, 0, 0.01758940, },
  [58] = {0.00000000, 0.01913220, 0.01704880, 0.00000000, 0.01574270, 0.00000000, 0.01817530, 0.01488180, 0.01718750, 0, 0.01729950, },
  [59] = {0.00000000, 0.01891480, 0.01684340, 0.00000000, 0.01535710, 0.00000000, 0.01794230, 0.01456410, 0.01683670, 0, 0.01692740, },
  [60] = {0.00000000, 0.01849440, 0.01644710, 0.00000000, 0.01505000, 0.00000000, 0.01749370, 0.01430940, 0.01650000, 0, 0.01639840, },
  [61] = {0.00000000, 0.01585240, 0.01571290, 0.00000000, 0.01481300, 0.00000000, 0.01638980, 0.01430940, 0.01590570, 0, 0.01615360, },
  [62] = {0.00000000, 0.01541200, 0.01539680, 0.00000000, 0.01447120, 0.00000000, 0.01585240, 0.01430940, 0.01540280, 0, 0.01565850, },
  [63] = {0.00000000, 0.01486160, 0.01496970, 0.00000000, 0.01425190, 0.00000000, 0.01519880, 0.01430940, 0.01476210, 0, 0.01502430, },
  [64] = {0.00000000, 0.01447390, 0.01442100, 0.00000000, 0.01393520, 0.00000000, 0.01467380, 0.01425960, 0.01428430, 0, 0.01457140, },
  [65] = {0.00000000, 0.01398740, 0.01406840, 0.00000000, 0.01368180, 0.00000000, 0.01422450, 0.01421010, 0.01384890, 0, 0.01415760, },
  [66] = {0.00000000, 0.01364340, 0.01375000, 0.00000000, 0.01343750, 0.00000000, 0.01378730, 0.01377950, 0.01345140, 0, 0.01373170, },
  [67] = {0.00000000, 0.01342340, 0.01333460, 0.00000000, 0.01320180, 0.00000000, 0.01338800, 0.01337420, 0.01302730, 0, 0.01334290, },
  [68] = {0.00000000, 0.01310630, 0.01303730, 0.00000000, 0.01297410, 0.00000000, 0.01307380, 0.01311700, 0.01273350, 0, 0.01305880, },
  [69] = {0.00000000, 0.01280380, 0.01275940, 0.00000000, 0.01271110, 0.00000000, 0.01277850, 0.01282920, 0.01261620, 0, 0.01277080, },
  [70] = {0.00000000, 0.01251500, 0.01250000, 0.00000000, 0.01250000, 0.00000000, 0.01250000, 0.01251530, 0.01250000, 0, 0.01250000, },
  [71] = {0.00000000, 0.01162640, 0.01162640, 0.00000000, 0.01162640, 0.00000000, 0.01162640, 0.01162640, 0.01162640, 0, 0.01162640, },
  [72] = {0.00000000, 0.01079820, 0.01079820, 0.00000000, 0.01079820, 0.00000000, 0.01079820, 0.01079820, 0.01079820, 0, 0.01079820, },
  [73] = {0.00000000, 0.01005530, 0.01005530, 0.00000000, 0.01005530, 0.00000000, 0.01005530, 0.01005530, 0.01005530, 0, 0.01005530, },
  [74] = {0.00000000, 0.00934361, 0.00934361, 0.00000000, 0.00934361, 0.00000000, 0.00934361, 0.00934361, 0.00934361, 0, 0.00934361, },
  [75] = {0.00000000, 0.00868896, 0.00868896, 0.00000000, 0.00868896, 0.00000000, 0.00868896, 0.00868896, 0.00868896, 0, 0.00868896, },
  [76] = {0.00000000, 0.00807199, 0.00807199, 0.00000000, 0.00807199, 0.00000000, 0.00807199, 0.00807199, 0.00807199, 0, 0.00807199, },
  [77] = {0.00000000, 0.00749542, 0.00749542, 0.00000000, 0.00749542, 0.00000000, 0.00749542, 0.00749542, 0.00749542, 0, 0.00749542, },
  [78] = {0.00000000, 0.00697189, 0.00697189, 0.00000000, 0.00697189, 0.00000000, 0.00697189, 0.00697189, 0.00697189, 0, 0.00697189, },
  [79] = {0.00000000, 0.00647547, 0.00647547, 0.00000000, 0.00647547, 0.00000000, 0.00647547, 0.00647547, 0.00647547, 0, 0.00647547, },
  [80] = {0.00000000, 0.00601838, 0.00601838, 0.00000000, 0.00601838, 0.00000000, 0.00601838, 0.00601838, 0.00601838, 0, 0.00601838, },
  [81] = {0.00000000, 0.00458339, 0.00458339, 0.00000000, 0.00458339, 0.00000000, 0.00458339, 0.00458339, 0.00458339, 0, 0.00458339, },
  [82] = {0.00000000, 0.00349034, 0.00349034, 0.00000000, 0.00349034, 0.00000000, 0.00349034, 0.00349034, 0.00349034, 0, 0.00349034, },
  [83] = {0.00000000, 0.00265690, 0.00265690, 0.00000000, 0.00265690, 0.00000000, 0.00265690, 0.00265690, 0.00265690, 0, 0.00265690, },
  [84] = {0.00000000, 0.00202346, 0.00202346, 0.00000000, 0.00202346, 0.00000000, 0.00202346, 0.00202346, 0.00202346, 0, 0.00202346, },
  [85] = {0.00000000, 0.00154105, 0.00154105, 0.00000000, 0.00154105, 0.00000000, 0.00154105, 0.00154105, 0.00154105, 0, 0.00154105, },
  [86] = {0.00000000, 0.00072153, 0.00072153, 0.00000000, 0.00072153, 0.00000000, 0.00072153, 0.00072153, 0.00072153, 0, 0.00072153, },
  [87] = {0.00000000, 0.00050666, 0.00050666, 0.00000000, 0.00050666, 0.00000000, 0.00050666, 0.00050666, 0.00050666, 0, 0.00050666, },
  [88] = {0.00000000, 0.00035578, 0.00035578, 0.00000000, 0.00035578, 0.00000000, 0.00035578, 0.00035578, 0.00035578, 0, 0.00035578, },
  [89] = {0.00000000, 0.00024983, 0.00024983, 0.00000000, 0.00024983, 0.00000000, 0.00024983, 0.00024983, 0.00024983, 0, 0.00024983, },
  [90] = {0.00000000, 0.00017543, 0.00017543, 0.00000000, 0.00017543, 0.00000000, 0.00017543, 0.00017543, 0.00017543, 0, 0.00017543, },
  [91] = {0.00000000, 0.00012319, 0.00012319, 0.00000000, 0.00012319, 0.00000000, 0.00012319, 0.00012319, 0.00012319, 0, 0.00012319, },
  [92] = {0.00000000, 0.00008650, 0.00008650, 0.00000000, 0.00008650, 0.00000000, 0.00008650, 0.00008650, 0.00008650, 0, 0.00008650, },
  [93] = {0.00000000, 0.00006074, 0.00006074, 0.00000000, 0.00006074, 0.00000000, 0.00006074, 0.00006074, 0.00006074, 0, 0.00006074, },
  [94] = {0.00000000, 0.00004265, 0.00004265, 0.00000000, 0.00004265, 0.00000000, 0.00004265, 0.00004265, 0.00004265, 0, 0.00004265, },
  [95] = {0.00000000, 0.00002995, 0.00002995, 0.00000000, 0.00002995, 0.00000000, 0.00002995, 0.00002995, 0.00002995, 0, 0.00002995, },
  [96] = {0.00000000, 0.00002103, 0.00002103, 0.00000000, 0.00002103, 0.00000000, 0.00002103, 0.00002103, 0.00002103, 0, 0.00002103, },
  [97] = {0.00000000, 0.00001477, 0.00001477, 0.00000000, 0.00001477, 0.00000000, 0.00001477, 0.00001477, 0.00001477, 0, 0.00001477, },
  [98] = {0.00000000, 0.00001037, 0.00001037, 0.00000000, 0.00001037, 0.00000000, 0.00001037, 0.00001037, 0.00001037, 0, 0.00001037, },
  [99] = {0.00000000, 0.00000728, 0.00000728, 0.00000000, 0.00000728, 0.00000000, 0.00000728, 0.00000728, 0.00000728, 0, 0.00000728, },
  [100] = {0.00000000, 0.00000511, 0.00000511, 0.00000000, 0.00000511, 0.00000000, 0.00000511, 0.00000511, 0.00000511, 0, 0.00000511, },
}

function StatLogic:GetSpellCritFromInt(int, class, level)
  -- argCheck for invalid input
  self:argCheck(int, 2, "number")
  self:argCheck(class, 3, "nil", "string", "number")
  self:argCheck(level, 4, "nil", "number")
  -- if class is a class string, convert to class id
  if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
    class = ClassNameToID[strupper(class)]
  -- if class is invalid input, default to player class
  elseif type(class) ~= "number" or class < 1 or class > 10 then
    class = ClassNameToID[playerClass]
  end
  -- if level is invalid input, default to player level
  if type(level) ~= "number" or level < 1 or level > 100 then
    level = UnitLevel("player")
  end
  -- Calculate
  return int * SpellCritPerInt[level][class], "SPELL_CRIT"
end


local BaseManaRegenPerSpi
--[[---------------------------------
  :GetNormalManaRegenFromSpi(spi, [int], [level])
-------------------------------------
Notes:
  * Formula and BASE_REGEN values derived by Whitetooth (hotdogee [at] gmail [dot] com)
  * Calculates the mana regen per 5 seconds from spirit when out of 5 second rule for given intellect and level.
  * Player class is no longer a parameter
  * ManaRegen(SPI, INT, LEVEL) = (0.001+SPI*BASE_REGEN[LEVEL]*(INT^0.5))*5
Arguments:
  number - Spirit
  [optional] number - Intellect. Default: player's intellect
  [optional] number - Level used in calculations. Default: player's level
Returns:
  ; mp5o5sr : number - Mana regen per 5 seconds when out of 5 second rule
  ; statid : string - "MANA_REGEN"
Example:
  local mp5o5sr = StatLogic:GetNormalManaRegenFromSpi(1) -- GetNormalManaRegenPerSpi
  local mp5o5sr = StatLogic:GetNormalManaRegenFromSpi(10, 15)
  local mp5o5sr = StatLogic:GetNormalManaRegenFromSpi(10, 15, 70)
-----------------------------------]]

-- Numbers reverse engineered by Whitetooth (hotdogee [at] gmail [dot] com)
local BaseManaRegenPerSpi = {
	[1] =  0.020979,
	[2] =  0.020515,
	[3] =  0.020079,
	[4] =  0.019516,
	[5] =  0.018997,
	[6] =  0.018646,
	[7] =  0.018314,
	[8] =  0.017997,
	[9] =  0.017584,
	[10] = 0.017197,
	[11] = 0.016551,
	[12] = 0.015729,
	[13] = 0.015229,
	[14] = 0.014580,
	[15] = 0.014008,
	[16] = 0.013650,
	[17] = 0.013175,
	[18] = 0.012832,
	[19] = 0.012475,
	[20] = 0.012073,
	[21] = 0.011840,
	[22] = 0.011494,
	[23] = 0.011292,
	[24] = 0.010990,
	[25] = 0.010761,
	[26] = 0.010546,
	[27] = 0.010321,
	[28] = 0.010151,
	[29] = 0.009949,
	[30] = 0.009740,
	[31] = 0.009597,
	[32] = 0.009425,
	[33] = 0.009278,
	[34] = 0.009123,
	[35] = 0.008974,
	[36] = 0.008847,
	[37] = 0.008698,
	[38] = 0.008581,
	[39] = 0.008457,
	[40] = 0.008338,
	[41] = 0.008235,
	[42] = 0.008113,
	[43] = 0.008018,
	[44] = 0.007906,
	[45] = 0.007798,
	[46] = 0.007713,
	[47] = 0.007612,
	[48] = 0.007524,
	[49] = 0.007430,
	[50] = 0.007340,
	[51] = 0.007268,
	[52] = 0.007184,
	[53] = 0.007116,
	[54] = 0.007029,
	[55] = 0.006945,
	[56] = 0.006884,
	[57] = 0.006805,
	[58] = 0.006747,
	[59] = 0.006667,
	[60] = 0.006600,
	[61] = 0.006421,
	[62] = 0.006314,
	[63] = 0.006175,
	[64] = 0.006072,
	[65] = 0.005981,
	[66] = 0.005885,
	[67] = 0.005791,
	[68] = 0.005732,
	[69] = 0.005668,
	[70] = 0.005596,
	[71] = 0.005316,
	[72] = 0.005049,
	[73] = 0.004796,
	[74] = 0.004555,
	[75] = 0.004327,
	[76] = 0.004110,
	[77] = 0.003903,
	[78] = 0.003708,
	[79] = 0.003522,
	[80] = 0.003345,
	[81] = 0.003345,
	[82] = 0.003345,
	[83] = 0.003345,
	[84] = 0.003345,
	[85] = 0.003345,
	[86] = 0.003, --TODO: Mop: Get base mana regen for level 86
	[87] = 0.003, --TODO: Mop: Get base mana regen for level 87
	[88] = 0.003, --TODO: Mop: Get base mana regen for level 88
	[89] = 0.003, --TODO: Mop: Get base mana regen for level 89
	[90] = 0.003, --TODO: Mop: Get base mana regen for level 90
	[91] = 0.002, --FUTURE
	[92] = 0.002, --FUTURE
	[93] = 0.002, --FUTURE
	[94] = 0.002, --FUTURE
	[95] = 0.002, --FUTURE
	[96] = 0.002, --FUTURE
	[97] = 0.002, --FUTURE
	[98] = 0.002, --FUTURE
	[99] = 0.002, --FUTURE
	[100] = 0.002, --FUTURE
}

function StatLogic:GetNormalManaRegenFromSpi(spi, int, level)
	-- argCheck for invalid input
	self:argCheck(spi, 2, "number")
	self:argCheck(int, 3, "nil", "number")
	self:argCheck(level, 4, "nil", "number")

	-- if level is invalid input, default to player level
	if type(level) ~= "number" or level < 1 or level > 85 then
		level = UnitLevel("player")
	end

	-- if int is invalid input, default to player int
	if type(int) ~= "number" then
		local _
		_, int = UnitStat("player",4)
	end
	-- Calculate
	return (0.001 + spi * BaseManaRegenPerSpi[level] * (int ^ 0.5)) * 5, "MANA_REGEN"
end

-- local Level80BaseMana = {
  -- 0, 23422, 0, 0, 20590, 0, 23430, 17418, 20553, 18635,
  -- --["WARRIOR"] = 0,
  -- --["PALADIN"] = 23422,
  -- --["HUNTER"] = 0,
  -- --["ROGUE"] = 0,
  -- --["PRIEST"] = 20590,
  -- --["DEATHKNIGHT"] = 0,
  -- --["SHAMAN"] = 23430,
  -- --["MAGE"] = 17418,
  -- --["WARLOCK"] = 20553,
  -- --["DRUID"] = 18635,
-- }

-- -- StatLogic:GetNormalManaRegenFromSpi2()
-- function StatLogic:GetNormalManaRegenFromSpi2(spi, int, level, class)
  -- -- argCheck for invalid input
  -- self:argCheck(spi, 2, "number")
  -- self:argCheck(int, 3, "nil", "number")
  -- self:argCheck(level, 4, "nil", "number")

  -- -- if level is invalid input, default to player level
  -- if type(level) ~= "number" or level < 1 or level > 80 then
    -- level = UnitLevel("player")
  -- end

  -- -- if int is invalid input, default to player int
  -- if type(int) ~= "number" then
    -- local _
    -- _, int = UnitStat("player",4)
  -- end
  
  -- -- if class is a class string, convert to class id
  -- if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
    -- class = ClassNameToID[strupper(class)]
  -- -- if class is invalid input, default to player class
  -- elseif type(class) ~= "number" or class < 1 or class > 10 then
    -- class = ClassNameToID[playerClass]
  -- end
  
  -- -- Calculate
  -- return Level80BaseMana[class] * 0.05 + (0.001 + spi * BaseManaRegenPerSpi[level] * (int ^ 0.5)) * 5, "MANA_REGEN"
-- end

--[[---------------------------------
  :GetHealthRegenFromSpi(spi, [class])
-------------------------------------
Notes:
  * HealthRegenPerSpi values derived by Whitetooth (hotdogee [at] gmail [dot] com)
  * Calculates the health regen per 5 seconds when out of combat from spirit for given class.
  * Player level does not effect health regen per spirit.
Arguments:
  number - Spirit
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
Returns:
  ; hp5oc : number - Health regen per 5 seconds when out of combat
  ; statid : string - "HEALTH_REGEN"
Example:
  local hp5oc = StatLogic:GetHealthRegenFromSpi(1) -- GetHealthRegenPerSpi
  local hp5oc = StatLogic:GetHealthRegenFromSpi(10)
  local hp5oc = StatLogic:GetHealthRegenFromSpi(10, "MAGE")
-----------------------------------]]

-- Numbers reverse engineered by Whitetooth (hotdogee [at] gmail [dot] com)
-- 4.0.1.12803: Removed Health Regen from Spirit
local HealthRegenPerSpi = {
	0, --Warrior
	0, --Paladin
	0, --Hunter
	0, --Rogue
	0, --Priest
	0, --Death Knight
	0, --Shaman
	0, --Mage
	0, --Warlock
	0, --Monk
	0, --Druid
	--["WARRIOR"] = 0.5,
	--["PALADIN"] = 0.125,
	--["HUNTER"] = 0.125,
	--["ROGUE"] = 0.333333,
	--["PRIEST"] = 0.041667,
	--["DEATHKNIGHT"] = 0.5,
	--["SHAMAN"] = 0.071429,
	--["MAGE"] = 0.041667,
	--["WARLOCK"] = 0.045455,
	--["DRUID"] = 0.0625,
}

function StatLogic:GetHealthRegenFromSpi(spi, class)
	-- argCheck for invalid input
	self:argCheck(spi, 2, "number")
	self:argCheck(class, 3, "nil", "string", "number")
	-- if class is a class string, convert to class id
	if type(class) == "string" and ClassNameToID[strupper(class)] ~= nil then
		class = ClassNameToID[strupper(class)]
	-- if class is invalid input, default to player class
	elseif type(class) ~= "number" or class < 1 or class > 10 then
		class = ClassNameToID[playerClass]
	end
	-- Calculate
	return spi * HealthRegenPerSpi[class] * 5, "HEALTH_REGEN"
end


----------
-- Gems --
----------

--[[---------------------------------
  :RemoveEnchant(link)
-------------------------------------
Notes:
  * Remove item's enchants.
Arguments:
  string - "itemlink"
Returns:
  ; link : number - The modified link
Example:
  local link = StatLogic:RemoveEnchant("Hitem:31052:425:525:525:525:525:0:0")
  "|cffa335ee|Hitem:47468:0:2968:0:0:0:0:-2055313024:80:147|h[Cry of the Val'kyr]|h|r",
-----------------------------------]]
function StatLogic:RemoveEnchant(link)
	-- check link
	if not strfind(link, "item:%d+:%d+:%d+:%d+:%d+:%d+:%-?%d+:%-?%d+") then
		return link
	end
	local linkType, itemId, enchantId, rest = strsplit(":", link, 4)
	return strjoin(":", linkType, itemId, 0, rest)
end

--[[---------------------------------
  :RemoveGem(link)
-------------------------------------
Notes:
  * Remove item's gems.
Arguments:
  string - "itemlink"
Returns:
  ; link : number - The modified link
Example:
  local link = StatLogic:RemoveGem("Hitem:31052:425:525:525:525:525:0:0")
-----------------------------------]]
function StatLogic:RemoveGem(link)
	-- check link
	if not strfind(link, "item:%d+:%d+:%d+:%d+:%d+:%d+:%-?%d+:%-?%d+") then
		return link
	end
	local linkType, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, rest = strsplit(":", link, 8)
	if reforging then
	else
	end
	return strjoin(":", linkType, itemId, enchantId, 0, 0, 0, 0, rest)
end

--[[---------------------------------
  :RemoveExtraSocketGem(link)
-------------------------------------
Notes:
  * Remove gems socketed in Prismatic Sockets, this includes Eternal Belt Buckles and Blacksmith only Bracer Socket and Glove Socket.
Arguments:
  string - "itemlink"
Returns:
  ; link : number - The modified link
Example:
  local link = StatLogic:RemoveExtraSocketGem("Hitem:31052:425:525:525:525:525:0:0")
-----------------------------------]]
local itemStatsTable = {}
local GetExtraSocketGemLoc = setmetatable({}, { __index = function(self, n)
	-- We are here because what we want is not in cache
	-- Get last gem location
	local lastGemLoc = 0
	local _, _, _, jewelId1, jewelId2, jewelId3, jewelId4 = strsplit(":", n)
	if jewelId4 ~= "0" then
		lastGemLoc = 4
	elseif jewelId3 ~= "0" then
		lastGemLoc = 3
	elseif jewelId2 ~= "0" then
		lastGemLoc = 2
	elseif jewelId1 ~= "0" then
		lastGemLoc = 1
	end
	if lastGemLoc == 0 then
		self[n] = 0
		return 0
	end
	-- Get number of sockets
	wipe(itemStatsTable)
	GetItemStats(n, itemStatsTable)
	--RatingBuster:Print(itemStatsTable)
	local numSockets = (itemStatsTable["EMPTY_SOCKET_RED"] or 0) + (itemStatsTable["EMPTY_SOCKET_YELLOW"] or 0) + (itemStatsTable["EMPTY_SOCKET_BLUE"] or 0)
	if numSockets < lastGemLoc then
		self[n] = lastGemLoc
		return lastGemLoc
	else
		self[n] = 0
		return 0
	end
end })

local extraSocketLoc = {
	["INVTYPE_WAIST"] = true,
	["INVTYPE_WRIST"] = true,
	["INVTYPE_HAND"] = true,
}
function StatLogic:RemoveExtraSocketGem(link)
	-- check link
	if not strfind(link, "item:%d+:%d+:%d+:%d+:%d+:%d+:%-?%d+:%-?%d+") then
		return link
	end
	-- check only belt, bracer and gloves
	local _, _, _, _, _, _, _, _, itemType = GetItemInfo(link)
	if not extraSocketLoc[itemType] then return link end
	-- Get current gem count
	local extraGemLoc = GetExtraSocketGemLoc[link]
	if extraGemLoc == 0 then return link end
	local linkType, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, rest = strsplit(":", link, 8)
	if extraGemLoc == 1 then
		jewelId1 = "0"
	elseif extraGemLoc == 2 then
		jewelId2 = "0"
	elseif extraGemLoc == 3 then
		jewelId3 = "0"
	elseif extraGemLoc == 4 then
		jewelId4 = "0"
	end
	return strjoin(":", linkType, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, rest)
end


--[[---------------------------------
  :RemoveEnchantGem(link)
-------------------------------------
Notes:
  * Remove item's gems and enchants.
Arguments:
  string - "itemlink"
Returns:
  ; link : number - The modified link
Example:
  local link = StatLogic:RemoveEnchantGem("Hitem:31052:425:525:525:525:525:0:0")
-----------------------------------]]
function StatLogic:RemoveEnchantGem(link)
  -- check link
  if not strfind(link, "item:%d+:%d+:%d+:%d+:%d+:%d+:%-?%d+:%-?%d+") then
    return link
  end
  local linkType, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, rest = strsplit(":", link, 8)
  return strjoin(":", linkType, itemId, 0, 0, 0, 0, 0, rest)
end

--[[---------------------------------
  :ModEnchantGem(link, enc, gem1, gem2, gem3, gem4)
-------------------------------------
Notes:
  * Add/Replace item's enchants or gems with given enchants or gems.
Arguments:
  string - "itemlink"
  [optional] number or string - enchantID to replace the current enchant. Default: no change
  [optional] number or string - gemID to replace the first gem. Default: no change
  [optional] number or string - gemID to replace the second gem. Default: no change
  [optional] number or string - gemID to replace the third gem. Default: no change
  [optional] number or string - gemID to replace the fourth gem. Default: no change
Returns:
  ; link : number - The modified link
Example:
  local link = StatLogic:ModEnchantGem("Hitem:31052:0:0:0:0:0:0:0", 1394)
-----------------------------------]]
function StatLogic:ModEnchantGem(link, enc, gem1, gem2, gem3, gem4)
  -- check link
  if not strfind(link, "item:%d+") then
    return
  end
  local linkType, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, rest = strsplit(":", link, 8)
  return strjoin(":", linkType, itemId, enc or enchantId or 0, gem1 or jewelId1 or 0, gem2 or jewelId2 or 0, gem3 or jewelId3 or 0, gem4 or jewelId4 or 0, rest)
end

--[[---------------------------------
  :BuildGemmedTooltip(item, red, yellow, blue, meta)
-------------------------------------
Notes:
  * Returns a modified link with all empty sockets replaced with the specified gems, sockets already gemmed will remain.
  * item:
  :;tooltip : table - The tooltip showing the item
  :;itemId : number - The numeric ID of the item. ie. 12345
  :;"itemString" : string - The full item ID in string format, e.g. "item:12345:0:0:0:0:0:0:0".
  :::Also supports partial itemStrings, by filling up any missing ":x" value with ":0", e.g. "item:12345:0:0:0"
  :;"itemName" : string - The Name of the Item, ex: "Hearthstone"
  :::The item must have been equiped, in your bags or in your bank once in this session for this to work.
  :;"itemLink" : string - The itemLink, when Shift-Clicking items.
Arguments:
  number or string or table - tooltip or itemId or "itemString" or "itemName" or "itemLink"
  number or string - gemID to replace a red socket
  number or string - gemID to replace a yellow socket
  number or string - gemID to replace a blue socket
  number or string - gemID to replace a meta socket
Returns:
  ; link : string - modified item link
Example:
  local link = StatLogic:BuildGemmedTooltip(28619, 3119, 3119, 3119, 3119)
  StatLogic:SetTip("item:28619")
  StatLogic:SetTip(StatLogic:BuildGemmedTooltip(28619, 3119, 3119, 3119, 3119))
-----------------------------------]]
local EmptySocketLookup = {
	[EMPTY_SOCKET_RED] = 0, -- EMPTY_SOCKET_RED = "Red Socket";
	[EMPTY_SOCKET_YELLOW] = 0, -- EMPTY_SOCKET_YELLOW = "Yellow Socket";
	[EMPTY_SOCKET_BLUE] = 0, -- EMPTY_SOCKET_BLUE = "Blue Socket";
	[EMPTY_SOCKET_META] = 0, -- EMPTY_SOCKET_META = "Meta Socket";
}
function StatLogic:BuildGemmedTooltip(item, red, yellow, blue, meta)
  local _
  -- Check item
  if (type(item) == "string") or (type(item) == "number") then
  elseif type(item) == "table" and type(item.GetItem) == "function" then
    -- Get the link
    _, item = item:GetItem()
    if type(item) ~= "string" then return item end
  else
    return item
  end
  -- Check if item is in local cache
  local name, link, _, _, reqLv, _, _, _, itemType = GetItemInfo(item)
  if not name then return item end

  -- Check gemID
  if not red or not tonumber(red) then red = 0 end
  if not yellow or not tonumber(yellow) then yellow = 0 end
  if not blue or not tonumber(blue) then blue = 0 end
  if not meta or not tonumber(meta) then meta = 0 end
  if red == 0 and yellow == 0 and blue == 0 and meta == 0 then return link end -- nothing to modify
  -- Fill EmptySocketLookup
  EmptySocketLookup[EMPTY_SOCKET_RED] = red
  EmptySocketLookup[EMPTY_SOCKET_YELLOW] = yellow
  EmptySocketLookup[EMPTY_SOCKET_BLUE] = blue
  EmptySocketLookup[EMPTY_SOCKET_META] = meta

  -- Build socket list
  local socketList = {}
  -- Get a link without any socketed gems
  local cleanLink = link:match("(item:%d+)")
  -- Start parsing
  tip:ClearLines() -- this is required or SetX won't work the second time its called
  tip:SetHyperlink(link)
  for i = 2, tip:NumLines() do
    local text = tip[i]:GetText()
    -- DEBUG
    if not text then
      print(i, tip:NumLines(), link)
    end
    -- Trim spaces
    text = strtrim(text)
    -- Strip color codes
    if strsub(text, -2) == "|r" then
      text = strsub(text, 1, -3)
    end
    if strfind(strsub(text, 1, 10), "|c%x%x%x%x%x%x%x%x") then
      text = strsub(text, 11)
    end
    local socketFound = EmptySocketLookup[text]
    if socketFound then
      socketList[#socketList+1] = socketFound
    end
  end
  -- If there are no sockets
  if #socketList == 0 then return link end
  -- link breakdown
  local linkType, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, rest = strsplit(":", link, 8)
  if socketList[1] and (not jewelId1 or jewelId1 == "0") then jewelId1 = socketList[1] end
  if socketList[2] and (not jewelId2 or jewelId2 == "0") then jewelId2 = socketList[2] end
  if socketList[3] and (not jewelId3 or jewelId3 == "0") then jewelId3 = socketList[3] end
  if socketList[4] and (not jewelId4 or jewelId4 == "0") then jewelId4 = socketList[4] end
  return strjoin(":", linkType, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, rest)
end

--[[---------------------------------
  :GetGemID(item)
-------------------------------------
Notes:
  * Returns the gemID and gemText of a gem for use in links
  * item:
  :;tooltip : table - The tooltip showing the item
  :;itemId : number - The numeric ID of the item. ie. 12345
  :;"itemString" : string - The full item ID in string format, e.g. "item:12345:0:0:0:0:0:0:0".
  :::Also supports partial itemStrings, by filling up any missing ":x" value with ":0", e.g. "item:12345:0:0:0"
  :;"itemName" : string - The Name of the Item, ex: "Hearthstone"
  :::The item must have been equiped, in your bags or in your bank once in this session for this to work.
  :;"itemLink" : string - The itemLink, when Shift-Clicking items.
Arguments:
  number or string or table - Gem, tooltip or itemId or "itemString" or "itemName" or "itemLink"
Returns:
  ; gemID or false : number or bool - The gemID of this gem, false if invalid input
  ; gemText : string - The text shown in the tooltip when socketed in an item
Example:
  local gemID, gemText = StatLogic:GetGemID(28363)
-----------------------------------]]
-- SetTip("item:3185:0:2946")
function StatLogic:GetGemID(item)
  local t = GetTime()
  -- Check item
  if (type(item) == "string") or (type(item) == "number") then
	-- We have link
  elseif type(item) == "table" and type(item.GetItem) == "function" then
    -- We were given a tooltip - get the link
    _, item = item:GetItem()
    if type(item) ~= "string" then return false end
  else
    return false
  end
  -- Check if item is in local cache
  local name, link = GetItemInfo(item)
  if not name then
    if tonumber(item) then
      -- Query server for item
      tipMiner:ClearLines()
      tipMiner:SetHyperlink("item:"..item)
    else
      item = item:match("item:(%d+)")
      if item then
        -- Query server for item
        tipMiner:ClearLines()
        tipMiner:SetHyperlink("item:"..item)
      else
        return false
      end
    end
    return
  end
  local itemID = strmatch(link, "item:(%d+)")
  local itemIDPattern=format("item:%d:", itemID)    -- for testing against gem itemlinks that we find
  local gemScanLink = "item:6948:0:0:0:%d:%d"
  

  -- Method 1: Try to find the gem already in our gear. Provides a layer of safety for future expansions (and might avoid unnecessary disconnects from scanning)
  for i=INVSLOT_FIRST_EQUIPPED, INVSLOT_LAST_EQUIPPED do
    local eqLink = GetInventoryItemLink("player", i)
    local gemIDs = { strmatch(eqLink or "", "item:[-0-9]+:[-0-9]+:([-0-9]+):([-0-9]+):([-0-9]+):") }    -- yah garbage. this is a manual operation; it matters not.
    for i=1, #gemIDs do
		local gemID = gemIDs[i]
        local gemName, gemLink = GetItemGem(eqLink, i)
        if gemLink and gemLink:match(itemIDPattern) then
          tipMiner:ClearLines() -- this is required or SetX won't work the second time its called
          tipMiner:SetHyperlink(gemScanLink:format(gemID, gemID))
          if GetCVarBool("colorblindMode") then
            return gemID, StatLogicMinerTooltipTextLeft6:GetText(), GetTime()-t
          else
            return gemID, StatLogicMinerTooltipTextLeft5:GetText(), GetTime()-t
          end
        end
    end
  end
  
  
  -- Method 2: Fallback scanner if we didn't find the gem in our gear. This will fail if gemIDs go higher than our assumed maximum
  if not GetItemInfo(6948) then -- Hearthstone
    -- Query server for Hearthstone
    tipMiner:ClearLines()
    tipMiner:SetHyperlink("item:6948")
    return
  end
  local gemID
  -- Start GemID scan
  for gemID = 5000, 1, -1 do    -- THIS NUMBER MAY NEED TO BE INCREASED IN NEW EXPANSIONS
    local itemLink = gemScanLink:format(gemID, gemID)
    local _, gem1Link = GetItemGem(itemLink, 3)
    if gem1Link and gem1Link:match(itemIDPattern) then
      tipMiner:ClearLines() -- this is required or SetX won't work the second time its called
      tipMiner:SetHyperlink(itemLink)
      if GetCVarBool("colorblindMode") then
        return gemID, StatLogicMinerTooltipTextLeft6:GetText(), GetTime()-t
      else
        return gemID, StatLogicMinerTooltipTextLeft5:GetText(), GetTime()-t
      end
    end
  end
end

-- will sometimes disconnect
-- StatLogic:GetEnchantID("+10 All Stats")
--[[
function StatLogic:GetEnchantID(spell)
  -- Check item
  if not ((type(spell) == "string") or (type(spell) == "number")) then
    return
  end
  local spellName = spell
  if type(spell) == "number" then
    spellName = GetSpellInfo(spell)
  end
  if not GetItemInfo(6948) then -- Hearthstone
    -- Query server for Hearthstone
    tipMiner:ClearLines()
    tipMiner:SetHyperlink("item:6948")
    return
  end
  local scanLink = "item:6948:%d:%d:%d:%d:%d"
  local id
  -- Start EnchantID scan
  for id = 4300, 5, -5 do
    local itemLink = scanLink:format(id, id-1, id-2, id-3, id-4)
    tipMiner:ClearLines() -- this is required or SetX won't work the second time its called
    tipMiner:SetHyperlink(itemLink)
    if StatLogicMinerTooltipTextLeft4:GetText() == spellName then
      return id, StatLogicMinerTooltipTextLeft4:GetText()
    elseif StatLogicMinerTooltipTextLeft5:GetText() and strfind(StatLogicMinerTooltipTextLeft5:GetText(), spellName) then
      return id, StatLogicMinerTooltipTextLeft5:GetText()
    elseif StatLogicMinerTooltipTextLeft6:GetText() and strfind(StatLogicMinerTooltipTextLeft6:GetText(), spellName) then
      return id, StatLogicMinerTooltipTextLeft6:GetText()
    elseif StatLogicMinerTooltipTextLeft7:GetText() and strfind(StatLogicMinerTooltipTextLeft7:GetText(), spellName) then
      return id, StatLogicMinerTooltipTextLeft7:GetText()
    elseif StatLogicMinerTooltipTextLeft8:GetText() == spellName then
      return id, StatLogicMinerTooltipTextLeft8:GetText()
    end
  end
end
--]]

-- ================== --
-- Stat Summarization --
-- ================== --
--[[---------------------------------
  :GetSum(item, [table])
-------------------------------------
Notes:
  * Calculates the sum of all stats for a specified item.
  * item:
  :;tooltip : table - The tooltip showing the item
  :;itemId : number - The numeric ID of the item. ie. 12345
  :;"itemString" : string - The full item ID in string format, e.g. "item:12345:0:0:0:0:0:0:0".
  :::Also supports partial itemStrings, by filling up any missing ":x" value with ":0", e.g. "item:12345:0:0:0"
  :;"itemName" : string - The Name of the Item, ex: "Hearthstone"
  :::The item must have been equiped, in your bags or in your bank once in this session for this to work.
  :;"itemLink" : string - The itemLink, when Shift-Clicking items.
Arguments:
  number or string or table - tooltip or itemId or "itemString" or "itemName" or "itemLink"
  table - The sum of stat values are writen to this table if provided
Returns:
  ; sumTable : table - The table with stat sum values
  :{
  ::    ["itemType"] = itemType,
  ::    ["STAT_ID1"] = value,
  ::    ["STAT_ID2"] = value,
  :}
Example:
  StatLogic:GetSum(21417) -- [Ring of Unspoken Names]
  StatLogic:GetSum("item:28040:2717")
  StatLogic:GetSum("item:19019:117") -- TF
  StatLogic:GetSum("item:3185:0:0:0:0:0:1957") -- Acrobatic Staff of Frozen Wrath ID:3185:0:0:0:0:0:1957
  StatLogic:GetSum(24396)
  SetTip("item:3185:0:0:0:0:0:1957")
  -- [Deadly Fire Opal] ID:30582 - Attack Power +8 and Critical Rating +5
  -- [Gnomeregan Auto-Blocker 600] ID:29387
  StatLogic:GetSum("item:30538:3011:2739:2739:2739:0") -- [Midnight Legguards] with enchant and gems
  StatLogic:GetSum("item:30538:3011:2739:2739:2739:0") -- [Midnight Legguards] with enchant and gems
-----------------------------------]]
--[[
0 = Poor
1 = Common
2 = Uncommon
3 = Rare
4 = Epic
5 = Legendary
6 = Artifact
7 = Heirloom
--]]
-- baseArmorTable[rarity][equipLoc][armorType][ilvl] = armorValue
-- Not interested in lower level items
local baseArmorTable = {
  [4] = {
    ["INVTYPE_CLOAK"] = {
      [L["Cloth"]] = {
        [200] = 404, -- Cloak of Armed Strife ID:39225
        [213] = 428, -- Cloak of the Shadowed Sun ID:40252
        [264] = 556, -- Sentinel's Winter Cloak ID:50466
      },
    },
    ["INVTYPE_LEGS"] = {
      [L["Plate"]] = {
        [226] = 2088, -- Saronite Plated Legguards ID:45267
      },
    },
  },
  [3] = {
    ["INVTYPE_CLOAK"] = {
      [L["Cloth"]] = {
        [167] = 312, -- Cloak of Tormented Skies ID:41238
        [187] = 364, -- Flowing Cloak of Command ID:37084
      },
    },
  },
}
local bonusArmorItemEquipLoc = {
	["INVTYPE_WEAPON"] = true,
	["INVTYPE_2HWEAPON"] = true,
	["INVTYPE_WEAPONMAINHAND"] = true,
	["INVTYPE_WEAPONOFFHAND"] = true,
	["INVTYPE_HOLDABLE"] = true,
	["INVTYPE_RANGED"] = true,
	["INVTYPE_THROWN"] = true,
	["INVTYPE_RANGEDRIGHT"] = true,
	["INVTYPE_NECK"] = true,
	["INVTYPE_FINGER"] = true,
	["INVTYPE_TRINKET"] = true,
}

function StatLogic:GetSum(item, table)
	-- Locale check
	if noPatternLocale then return end
	-- Clear table values
	clearTable(table)
	local _
	
	-- Check item
	if (type(item) == "string") or (type(item) == "number") then -- common case first
	elseif type(item) == "table" and type(item.GetItem) == "function" then
		-- Get the link
		_, item = item:GetItem()
		if type(item) ~= "string" then return end
	else
		return table
	end
	-- Check if item is in local cache
	local name, link, rarity , ilvl, reqLv, _, armorType, _, itemType = GetItemInfo(item)
	if not name then return table end

	-- Initialize table
	table = table or new()
	setmetatable(table, statTableMetatable)

	-- Get data from cache if available
	if cache[link] then
		copyTable(table, cache[link])
		return table
	end

	-- Set metadata
	table.itemType = itemType
	table.link = link

	-- Start parsing
	tip:ClearLines() -- this is required or SetX won't work the second time its called
	tip:SetHyperlink(link)
	debugPrint("Item being analysed: "..link)
	for i = 2, tip:NumLines() do
		local text = tip[i]:GetText()
		-- Trim spaces
		text = strtrim(text)
		-- Strip color codes
		if strsub(text, -2) == "|r" then
			text = strsub(text, 1, -3)
		end
		if strfind(strsub(text, 1, 10), "|c%x%x%x%x%x%x%x%x") then
			text = strsub(text, 11)
		end

		local r, g, b = tip[i]:GetTextColor()
	
		-----------------------
		-- Whole Text Lookup --
		-----------------------
		-- Mainly used for enchants or stuff without numbers:
		-- "Mithril Spurs"
		local found
		local idTable = L.WholeTextLookup[text]
		if idTable == false then
			found = true
			debugPrint("|cffadadad".."   WholeText Exclude: "..text)
		elseif idTable then
			found = true
			for id, value in pairs(L.WholeTextLookup[text]) do
				-- sum stat
				table[id] = (table[id] or 0) + CNumber(value);
				debugPrint("|cffff5959".."   WholeText: ".."|cffffc259"..text..", ".."|cffffff59"..tostring(id).."="..tostring(value))
			end
		end
		
		-- Fast Exclude --
		-- Exclude obvious strings that do not need to be checked, also exclude lines that are not white and green and normal (normal for Frozen Wrath bonus)
		-- Note: LUA has no "continue" statement; so we have to indent everything one extra needless level
		if not (found or L.Exclude[text] or L.Exclude[strutf8sub(text, 1, L.ExcludeLen)] or strsub(text, 1, 1) == '"' or g < 0.8 or (b < 0.99 and b > 0.1)) then
			--debugPrint(text.." = ")
			-- Strip enchant time
			-- ITEM_ENCHANT_TIME_LEFT_DAYS = "%s (%d day)";
			-- ITEM_ENCHANT_TIME_LEFT_DAYS_P1 = "%s (%d days)";
			-- ITEM_ENCHANT_TIME_LEFT_HOURS = "%s (%d hour)";
			-- ITEM_ENCHANT_TIME_LEFT_HOURS_P1 = "%s (%d hrs)";
			-- ITEM_ENCHANT_TIME_LEFT_MIN = "%s (%d min)"; -- Enchantment name, followed by the time left in minutes
			-- ITEM_ENCHANT_TIME_LEFT_SEC = "%s (%d sec)"; -- Enchantment name, followed by the time left in seconds
			--[[ 
			Seems temp enchants such as mana oil can't be seen from item links, so commented out
			if strfind(text, "%)") then
				debugPrint("test")
				text = gsub(text, gsub(gsub(ITEM_ENCHANT_TIME_LEFT_DAYS, "%%s ", ""), "%%", "%%%%"), "")
				text = gsub(text, gsub(gsub(ITEM_ENCHANT_TIME_LEFT_DAYS_P1, "%%s ", ""), "%%", "%%%%"), "")
				text = gsub(text, gsub(gsub(ITEM_ENCHANT_TIME_LEFT_HOURS, "%%s ", ""), "%%", "%%%%"), "")
				text = gsub(text, gsub(gsub(ITEM_ENCHANT_TIME_LEFT_HOURS_P1, "%%s ", ""), "%%", "%%%%"), "")
				text = gsub(text, gsub(gsub(ITEM_ENCHANT_TIME_LEFT_MIN, "%%s ", ""), "%%", "%%%%"), "")
				text = gsub(text, gsub(gsub(ITEM_ENCHANT_TIME_LEFT_SEC, "%%s ", ""), "%%", "%%%%"), "")
			end
			--]]

			----------------------------
			-- Single Plus Stat Check --
			----------------------------
			-- depending on locale, L.SinglePlusStatCheck may be
			-- +19 Stamina = "^%+(%d+) ([%a ]+%a)$"
			-- Stamina +19 = "^([%a ]+%a) %+(%d+)$"
			-- +19 耐力 = "^%+(%d+) (.-)$"
			if not found then
				local _, _, value, statText = strfind(strutf8lower(text), L.SinglePlusStatCheck)
				if value then
					if tonumber(statText) then
						value, statText = statText, value
					end
					local idTable = L.StatIDLookup[statText]
					if idTable == false then
						found = true
						debugPrint("|cffadadad".."   SinglePlus Exclude: "..text)
					elseif idTable then
						found = true
						local debugText = "|cffff5959".."   SinglePlus: ".."|cffffc259"..text
						local nValue = CNumber(value);
						for _, id in ipairs(idTable) do
							--debugPrint("  '"..value.."', '"..id.."'")
							-- sum stat
							table[id] = (table[id] or 0) + nValue;
							debugText = debugText..", ".."|cffffff59"..tostring(id).."="..tostring(value)
						end
						debugPrint(debugText)
					else
						-- pattern match but not found in L.StatIDLookup, keep looking
						debugPrint("  SinglePlusStatCheck Lookup Fail: |cffffd4d4'"..statText.."'")
					end
				end
			end
			
			-----------------------------
			-- Single Equip Stat Check --
			-----------------------------
			-- depending on locale, L.SingleEquipStatCheck may be
			-- "^Equip: (.-) by u?p? ?t?o? ?(%d+) ?(.-)%.$"
			if not found then
				local _, _, statText1, value, statText2 = strfind(text, L.SingleEquipStatCheck)
				if value then
					local statText = statText1..statText2
					local idTable = L.StatIDLookup[strutf8lower(statText)]
					if idTable == false then
						--We're excluding it because the value is *false*
						found = true
						debugPrint("|cffadadad".."   SingleEquip Exclude: "..text)
					elseif idTable then
						found = true
						local debugText = "|cffff5959".."   SingleEquip: ".."|cffffc259"..text
						
						--Check that value is a valid number (e.g. 5,50 Schaden pro Sekunde)
						local nValue = CNumber(value);
						if nValue == nil then
							error(string.format("SingleEquip: Cannot convert value \"%s\" to a number. Line \"%s\"", value, text));
						end;
						
						for _, id in ipairs(idTable) do
							--debugPrint("  '"..value.."', '"..id.."'")
							-- sum stat
							table[id] = (table[id] or 0) + nValue;
							debugText = debugText..", ".."|cffffff59"..tostring(id).."="..tostring(value)
						end
						debugPrint(debugText)
					else
						-- pattern match but not found in L.StatIDLookup, keep looking
						debugPrint("   SingleEquipStatCheck Lookup Fail: |cffffd4d4'"..statText.."'")
					end
				end
			end
			
			-- PreScan for special cases, that will fit wrongly into DeepScan
			-- PreScan also has exclude patterns
			-- This is where base armor gets scanned, check text color for bonus armor
			if not found then
				for pattern, id in pairs(L.PreScanPatterns) do
					local value
					found, _, value = strfind(text, pattern)
					if found then
						--found = true
						if id ~= false then
							local debugText = "|cffff5959".."   PreScan: ".."|cffffc259"..text
							--debugPrint("  '"..value.."' = '"..id.."'")
							-- check text color for bonus armor
							if id == "ARMOR" and r == 0 and b == 0 and
									baseArmorTable[rarity] and baseArmorTable[rarity][itemType] and
									baseArmorTable[rarity][itemType][armorType] and baseArmorTable[rarity][itemType][armorType][ilvl] and
									tonumber(value) > baseArmorTable[rarity][itemType][armorType][ilvl] then
								table["ARMOR"] = (table["ARMOR"] or 0) + baseArmorTable[rarity][itemType][armorType][ilvl]
								table["ARMOR_BONUS"] = (table["ARMOR_BONUS"] or 0) + CNumber(value) - baseArmorTable[rarity][itemType][armorType][ilvl]
								debugText = debugText..", ".."|cffffff59ARMOR="..baseArmorTable[rarity][itemType][armorType][ilvl]..", ARMOR_BONUS="..(CNumber(value) - baseArmorTable[rarity][itemType][armorType][ilvl])
							else
								-- sum stat
								table[id] = (table[id] or 0) + CNumber(value)
								debugText = debugText..", ".."|cffffff59"..tostring(id).."="..tostring(value)
							end
							debugPrint(debugText)
						else
							debugPrint("|cffadadad".."   PreScan Exclude: "..text.."  ("..pattern..")")
						end
						break
					end
				end
				if found then

				end
			end
		
			--------------
			-- DeepScan --
			--------------
			--[[
			-- Strip trailing "."
			["."] = ".",
			["DeepScanSeparators"] = {
				"/", -- "+10 Defense Rating/+10 Stamina/+15 Block Value": ZG Enchant
				" & ", -- "+26 Healing Spells & 2% Reduced Threat": Bracing Earthstorm Diamond ID:25897
				", ", -- "+6 Spell Damage, +5 Spell Crit Rating": Potent Ornate Topaz ID: 28123
				"%. ", -- "Equip: Increases attack power by 81 when fighting Undead. It also allows the acquisition of Scourgestones on behalf of the Argent Dawn.": Seal of the Dawn
			},
			["DeepScanWordSeparators"] = {
				" and ", -- "Critical Rating +6 and Dodge Rating +5": Assassin's Fire Opal ID:30565
			},
			["DeepScanPatterns"] = {
				"^(.-) by u?p? ?t?o? ?(%d+) ?(.-)$", -- "xxx by up to 22 xxx" (scan first)
				"^(.-) ?%+(%d+) ?(.-)$", -- "xxx xxx +22" or "+22 xxx xxx" or "xxx +22 xxx" (scan 2ed)
				"^(.-) ?([%d%.]+) ?(.-)$", -- 22.22 xxx xxx (scan last)
			},
			--]]
			if not found then
				-- Get a local copy
				local text = text
				-- Strip leading "Equip: ", "Socket Bonus: "
				text = gsub(text, _G.ITEM_SPELL_TRIGGER_ONEQUIP, "") -- ITEM_SPELL_TRIGGER_ONEQUIP = "Equip:";
				text = gsub(text, StripGlobalStrings(_G.ITEM_SOCKET_BONUS), "") -- ITEM_SOCKET_BONUS = "Socket Bonus: %s"; -- Tooltip tag for socketed item matched socket bonuses
				-- Trim spaces
				text = strtrim(text)
				-- Strip trailing "."
				if strutf8sub(text, -1) == L["."] then
					text = strutf8sub(text, 1, -2)
				end
				-- Replace separators with @
				for _, sep in ipairs(L.DeepScanSeparators) do
					if strfind(text, sep) then
						-- if there is a capture, for deDE
						if strsub(sep, 1, 1) == "(" then
							text = gsub(text, sep, "%1@")
						else
							text = gsub(text, sep, "@")
						end
					end
				end
				
				-- Split text using @
				text = {strsplit("@", text)}
				for i, text in ipairs(text) do
					-- Trim spaces
					text = strtrim(text)
					-- Strip color codes
					if strsub(text, -2) == "|r" then
						text = strsub(text, 1, -3)
					end
					if strfind(strsub(text, 1, 10), "|c%x%x%x%x%x%x%x%x") then
						text = strsub(text, 11)
					end
					-- Strip trailing "."
					if strutf8sub(text, -1) == L["."] then
						text = strutf8sub(text, 1, -2)
					end
					debugPrint("|cff008080".."S"..i..": ".."'"..text.."'")
					
					-- Whole Text Lookup
					local foundWholeText = false
					local idTable = L.WholeTextLookup[text]
					if idTable == false then
						foundWholeText = true
						found = true
						debugPrint("|cffadadad".."  DeepScan WholeText Exclude: "..text)
					elseif idTable then
						foundWholeText = true
						found = true
						for id, value in pairs(L.WholeTextLookup[text]) do
							-- sum stat
							table[id] = (table[id] or 0) + CNumber(value);
							debugPrint("|cffff5959".."  DeepScan WholeText: ".."|cffffc259"..text..", ".."|cffffff59"..tostring(id).."="..tostring(value))
						end
						elseif L.Exclude[text] or L.Exclude[strutf8sub(text, 1, L.ExcludeLen)] then
							foundWholeText = true
							found = true
							debugPrint("|cffadadad".."  DeepScan Exclude: "..text)
							-- pattern match but not found in L.WholeTextLookup, keep looking
						end
						-- Scan DualStatPatterns
						if not foundWholeText then
							for pattern, dualStat in pairs(L.DualStatPatterns) do
								local lowered = strutf8lower(text)
								local _, dEnd, value1, value2 = strfind(lowered, pattern)
								if value1 and value2 then
									foundWholeText = true
									found = true
									local nValue1 = CNumber(value1);
									local nValue2 = CNumber(value2);
									local debugText = "|cffff5959".."  DeepScan DualStat: ".."|cffffc259"..text
									for _, id in ipairs(dualStat[1]) do
										--debugPrint("  '"..value.."', '"..id.."'")
										-- sum stat
										table[id] = (table[id] or 0) + nValue1;
										debugText = debugText..", ".."|cffffff59"..tostring(id).."="..tostring(value1)
									end
									for _, id in ipairs(dualStat[2]) do
										--debugPrint("  '"..value.."', '"..id.."'")
										-- sum stat
										table[id] = (table[id] or 0) + nValue2;
										debugText = debugText..", ".."|cffffff59"..tostring(id).."="..tostring(value2)
									end
									debugPrint(debugText)
									if dEnd ~= strlen(lowered) then
										foundWholeText = false
										text = strsub(text, dEnd + 1)
									end
									break
								end
							end
						end
						local foundDeepScan1 = false
						if not foundWholeText then
							local lowered = strutf8lower(text)
							-- Pattern scan
							for _, pattern in ipairs(L.DeepScanPatterns) do -- try all patterns in order
								local _, _, statText1, value, statText2 = strfind(lowered, pattern)
								if value then
									local statText = statText1..statText2
									local idTable = L.StatIDLookup[statText]
									if idTable == false then
										foundDeepScan1 = true
										found = true
										debugPrint("|cffadadad".."  DeepScan Exclude: "..text)
										break -- break out of pattern loop and go to the next separated text
									elseif idTable then
										foundDeepScan1 = true
										found = true
										local nValue = CNumber(value);
										local debugText = "|cffff5959".."  DeepScan: ".."|cffffc259"..text
										for _, id in ipairs(idTable) do
											--debugPrint("  '"..value.."', '"..id.."'")
											-- sum stat
											table[id] = (table[id] or 0) + nValue;
											debugText = debugText..", ".."|cffffff59"..tostring(id).."="..tostring(value)
										end
										debugPrint(debugText)
										break -- break out of pattern loop and go to the next separated text
									else
										-- Not matching pattern
									end
								end
							end
						end
						-- If still not found, use the word separators to split the text
						if not foundWholeText and not foundDeepScan1 then
						-- Replace separators with @
						for _, sep in ipairs(L.DeepScanWordSeparators) do
							if strfind(text, sep) then
								text = gsub(text, sep, "@")
							end
						end
						-- Split text using @
						text = {strsplit("@", text)}
						for j, text in ipairs(text) do
							-- Trim spaces
							text = strtrim(text)
							-- Strip color codes
							if strsub(text, -2) == "|r" then
								text = strsub(text, 1, -3)
							end
							if strfind(strsub(text, 1, 10), "|c%x%x%x%x%x%x%x%x") then
								text = strsub(text, 11)
							end
							-- Strip trailing "."
							if strutf8sub(text, -1) == L["."] then
								text = strutf8sub(text, 1, -2)
							end
							debugPrint("|cff008080".."S"..i.."-"..j..": ".."'"..text.."'")
							-- Whole Text Lookup
							local foundWholeText = false
							local idTable = L.WholeTextLookup[text]
							if idTable == false then
								foundWholeText = true
								found = true
								debugPrint("|cffadadad".."  DeepScan2 WholeText Exclude: "..text)
							elseif idTable then
								foundWholeText = true
								found = true
								for id, value in pairs(L.WholeTextLookup[text]) do
									-- sum stat
									table[id] = (table[id] or 0) + value
									debugPrint("|cffff5959".."  DeepScan2 WholeText: ".."|cffffc259"..text..", ".."|cffffff59"..tostring(id).."="..tostring(value))
								end
							else
								-- pattern match but not found in L.WholeTextLookup, keep looking
							end
							
							-- Scan DualStatPatterns
							if not foundWholeText then
								for pattern, dualStat in pairs(L.DualStatPatterns) do
									local lowered = strutf8lower(text)
									local _, _, value1, value2 = strfind(lowered, pattern)
									if value1 and value2 then
										foundWholeText = true
										found = true
										local debugText = "|cffff5959".."  DeepScan2 DualStat: ".."|cffffc259"..text
										
										local nValue1 = CNumber(value1);
										local nValue2 = CNumber(value2);
										
										for _, id in ipairs(dualStat[1]) do
											--debugPrint("  '"..value.."', '"..id.."'")
											-- sum stat
											table[id] = (table[id] or 0) + nValue1;
											debugText = debugText..", ".."|cffffff59"..tostring(id).."="..tostring(value1)
										end
										for _, id in ipairs(dualStat[2]) do
											--debugPrint("  '"..value.."', '"..id.."'")
											-- sum stat
											table[id] = (table[id] or 0) + nValue2;
											debugText = debugText..", ".."|cffffff59"..tostring(id).."="..tostring(value2)
										end
										debugPrint(debugText)
										break
									end
								end
							end
							local foundDeepScan2 = false
							if not foundWholeText then
								local lowered = strutf8lower(text)
								-- Pattern scan
								for _, pattern in ipairs(L.DeepScanPatterns) do
									local _, _, statText1, value, statText2 = strfind(lowered, pattern)
									if value then
										local statText = statText1..statText2
										local idTable = L.StatIDLookup[statText]
										if idTable == false then
											foundDeepScan2 = true
											found = true
											debugPrint("|cffadadad".."  DeepScan2 Exclude: "..text.." (Pattern: "..pattern..")")
											break
										elseif idTable then
											foundDeepScan2 = true
											found = true
											local nValue = CNumber(value);
											
											local debugText = "|cffff5959".."  DeepScan2: ".."|cffffc259"..text
											for _, id in ipairs(idTable) do
												--debugPrint("  '"..value.."', '"..id.."'")
												-- sum stat
												table[id] = (table[id] or 0) + nValue;
												debugText = debugText..", ".."|cffffff59"..tostring(id).."="..tostring(value)
											end
											debugPrint(debugText.." (Pattern: "..pattern..")");
											break
										else
											-- pattern match but not found in L.StatIDLookup, keep looking
											debugPrint("  DeepScan2 Lookup Fail: |cffffd4d4'"..statText.."'|r, pattern = |cff72ff59'"..pattern.."'")
										end
									end
								end -- for
							end
							if not foundWholeText and not foundDeepScan2 then
								debugPrint("  DeepScan2 Fail: |cffff0000'"..text.."'")
							end
						end
					end -- if not foundWholeText and not foundDeepScan1 then
				end
			end
	
			if not found then
				local notFoundCount = (StatLogic.ItemsNotRecognized[text] or 0) +1;
				StatLogic.ItemsNotRecognized[text] = notFoundCount;
	
				if ((notFoundCount == 1) and (_itemNotRecognizedWarning)) or DEBUG then
					local s = ORANGE_FONT_COLOR_CODE.."LibStatLogic"..FONT_COLOR_CODE_CLOSE..": No Match for \""..RED_FONT_COLOR_CODE..text..FONT_COLOR_CODE_CLOSE.."\"";
					
					--If warnings were enabled by default (i.e. they have no idea how to turn them off), tell them how to turn them off
					if ItemNotRecognizedWarningDefault == true then
						s = s.." Warnings can be disabled by typing"..YELLOW_FONT_COLOR_CODE.." /slwarning"..FONT_COLOR_CODE_CLOSE;
					end;
					
					print(s);
				end;
				
				--print("LibStatLogic: No match for |cffff0000'"..text.."' on "..link);
				-- if DEBUG and RatingBuster then
				--		RatingBuster.db.profile.test = text
				--	end
			end
		else
			--line was fast excluded
			--debugPrint("   Excluded: "..text); --it's helpful when debugging to see if an item's property was ignored - even if it is spammy
		end
	end --for each line in the tooltip

	-- Tooltip scanning done, do post processing
	--[[ 3.0.8
		Bonus Armor: The mechanics for items with bonus armor on them has
		changed (any cloth, leather, mail, or plate items with extra armor,
		or any other items with any armor). Bonus armor beyond the base
		armor of an item will no longer be multiplied by any talents or by
		the bonuses of Bear Form, Dire Bear Form, or Frost Presence.
	--]]
	if bonusArmorItemEquipLoc[itemType] and table["ARMOR"] then
		-- Convert "ARMOR" to "ARMOR_BONUS"
		table["ARMOR_BONUS"] = (table["ARMOR_BONUS"] or 0) + table["ARMOR"]
		table["ARMOR"] = nil
	end
	cache[link] = copy(table)
	return table
end

function StatLogic:GetFinalArmor(item, text)
	-- Locale check
	if noPatternLocale then return end
	local _
	-- Check item
	if (type(item) == "string") or (type(item) == "number") then -- common case first
	elseif type(item) == "table" and type(item.GetItem) == "function" then
		-- Get the link
		_, item = item:GetItem()
		if type(item) ~= "string" then return end
	else
		return
	end
	-- Check if item is in local cache
	local name, _, rarity , ilvl, _, _, armorType, _, itemType = GetItemInfo(item)
	if not name then return end

	for pattern, id in pairs(L.PreScanPatterns) do
		if id == "ARMOR" or id == "ARMOR_BONUS" then
			local found, _, value = strfind(text, pattern)
			if found then
				local armor = 0
				local bonus_armor = 0
				if id == "ARMOR" and baseArmorTable[rarity] and baseArmorTable[rarity][itemType] and
						baseArmorTable[rarity][itemType][armorType] and baseArmorTable[rarity][itemType][armorType][ilvl] and
						tonumber(value) > baseArmorTable[rarity][itemType][armorType][ilvl] then
					armor = baseArmorTable[rarity][itemType][armorType][ilvl]
					bonus_armor = tonumber(value) - baseArmorTable[rarity][itemType][armorType][ilvl]
				else
					armor = tonumber(value)
				end
				if bonusArmorItemEquipLoc[itemType] then
					bonus_armor = bonus_armor + armor
					armor = 0
				end
				return armor * self:GetStatMod("MOD_ARMOR") + bonus_armor
			end
		end
	end
end

--[[---------------------------------
  :GetDiffID(item, [ignoreEnchant], [ignoreGem], [red], [yellow], [blue], [meta], [ignorePris])
-------------------------------------
Notes:
  * Returns a unique identification string of the diff calculation, the identification string is made up of links concatenated together, can be used for cache indexing
  * item:
  :;tooltip : table - The tooltip showing the item
  :;itemId : number - The numeric ID of the item. e.g. 12345
  :;"itemString" : string - The full item ID in string format, e.g. "item:12345:0:0:0:0:0:0:0".
  :::Also supports partial itemStrings, by filling up any missing ":x" value with ":0", e.g. "item:12345:0:0:0"
  :;"itemName" : string - The Name of the Item, ex: "Hearthstone"
  :::The item must have been equiped, in your bags or in your bank once in this session for this to work.
  :;"itemLink" : string - The itemLink, when Shift-Clicking items.
Arguments:
  number or string or table - tooltip or itemId or "itemString" or "itemName" or "itemLink"
  boolean - Ignore enchants when calculating the id if true
  boolean - Ignore gems when calculating the id if true
  number or string - gemID to replace a red socket
  number or string - gemID to replace a yellow socket
  number or string - gemID to replace a blue socket
  number or string - gemID to replace a meta socket
  boolean - Ignore prismatic sockets when calculating the id if true
Returns:
  ; id : string - a unique identification string of the diff calculation, for use as cache key
  ; link : string - link of main item
  ; linkDiff1 : string - link of compare item 1
  ; linkDiff2 : string - link of compare item 2
Example:
  StatLogic:GetDiffID(21417) -- Ring of Unspoken Names
  StatLogic:GetDiffID("item:18832:2564:0:0:0:0:0:0", true, true) -- Brutality Blade with +15 agi enchant
  http://www.wowwiki.com/EnchantId
-----------------------------------]]

local getSlotID = {
	INVTYPE_AMMO           = 0,
	INVTYPE_GUNPROJECTILE  = 0,
	INVTYPE_BOWPROJECTILE  = 0,
	INVTYPE_HEAD           = 1,
	INVTYPE_NECK           = 2,
	INVTYPE_SHOULDER       = 3,
	INVTYPE_BODY           = 4,
	INVTYPE_CHEST          = 5,
	INVTYPE_ROBE           = 5,
	INVTYPE_WAIST          = 6,
	INVTYPE_LEGS           = 7,
	INVTYPE_FEET           = 8,
	INVTYPE_WRIST          = 9,
	INVTYPE_HAND           = 10,
	INVTYPE_FINGER         = {11,12},
	INVTYPE_TRINKET        = {13,14},
	INVTYPE_CLOAK          = 15,
	INVTYPE_WEAPON         = {16,17},
	INVTYPE_2HWEAPON       = 16+17,
	INVTYPE_WEAPONMAINHAND = 16,
	INVTYPE_WEAPONOFFHAND  = 17,
	INVTYPE_SHIELD         = 17,
	INVTYPE_HOLDABLE       = 17,
	INVTYPE_RANGED         = 18,
	INVTYPE_RANGEDRIGHT    = 18,
	INVTYPE_RELIC          = 18,
	INVTYPE_GUN            = 18,
	INVTYPE_CROSSBOW       = 18,
	INVTYPE_WAND           = 18,
	INVTYPE_THROWN         = 18,
	INVTYPE_TABARD         = 19,
}

function HasTitanGrip(talentGroup)
	--20120831: Mists of Panderia (5.0.4) removed all (useful) talents from the game. Titan Grip
	-- no longer exists as a talent (i don't know if it exists at all anymore).
	return false;
	
--[[
	if playerClass == "WARRIOR" then
		local _, _, _, _, r = GetTalentInfo(2, 20, nil, nil, talentGroup);
		return r > 0;
	end;
--]]
end

function StatLogic:GetDiffID(item, ignoreEnchant, ignoreGem, red, yellow, blue, meta, ignorePris)
	local _, name, itemType, link, linkDiff1, linkDiff2
	-- Check item
	if (type(item) == "string") or (type(item) == "number") then
	elseif type(item) == "table" and type(item.GetItem) == "function" then
		-- Get the link
		_, item = item:GetItem()
		if type(item) ~= "string" then return end
	else
		return
	end
	-- Check if item is in local cache
	name, link, _, _, _, _, _, _, itemType = GetItemInfo(item)
	if not name then return end
	-- Get equip location slot id for use in GetInventoryItemLink
	local slotID = getSlotID[itemType]
	-- Don't do bags
	if not slotID then return end

	-- 1h weapon, check if player can dual wield, check for 2h equipped
	if itemType == "INVTYPE_WEAPON" then
		linkDiff1 = GetInventoryItemLink("player", 16) or "NOITEM"
		-- If player can Dual Wield, calculate offhand difference
		if IsUsableSpell(GetSpellInfo(674)) then        -- ["Dual Wield"]
			local _, _, _, _, _, _, _, _, eqItemType = GetItemInfo(linkDiff1)
			-- If 2h is equipped, copy diff1 to diff2
			if eqItemType == "INVTYPE_2HWEAPON" and not HasTitanGrip() then
				linkDiff2 = linkDiff1
			else
				linkDiff2 = GetInventoryItemLink("player", 17) or "NOITEM"
			end
		end
	-- Ring or trinket
	elseif type(slotID) == "table" then
		-- Get slot link
		linkDiff1 = GetInventoryItemLink("player", slotID[1]) or "NOITEM"
		linkDiff2 = GetInventoryItemLink("player", slotID[2]) or "NOITEM"
		-- 2h weapon, so we calculate the difference with equipped main hand and off hand
	elseif itemType == "INVTYPE_2HWEAPON" then
		linkDiff1 = GetInventoryItemLink("player", 16) or "NOITEM"
		linkDiff2= GetInventoryItemLink("player", 17) or "NOITEM"
		-- Off hand slot, check if we have 2h equipped
	elseif slotID == 17 then
		linkDiff1 = GetInventoryItemLink("player", 16) or "NOITEM"
		-- If 2h is equipped
		local _, _, _, _, _, _, _, _, eqItemType = GetItemInfo(linkDiff1)
		if eqItemType ~= "INVTYPE_2HWEAPON" then
			linkDiff1 = GetInventoryItemLink("player", 17) or "NOITEM"
		end
		-- Single slot item
	else
		linkDiff1 = GetInventoryItemLink("player", slotID) or "NOITEM"
	end

	-- Ignore Enchants
	if ignoreEnchant then
		link = self:RemoveEnchant(link)
		linkDiff1 = self:RemoveEnchant(linkDiff1)
		if linkDiff2 then
			linkDiff2 = self:RemoveEnchant(linkDiff2)
		end
	end
	if ignorePris then
		link = self:RemoveExtraSocketGem(link)
		linkDiff1 = self:RemoveExtraSocketGem(linkDiff1)
		if linkDiff2 then
			linkDiff2 = self:RemoveExtraSocketGem(linkDiff2)
		end
	end
	
	-- Ignore Gems
	if ignoreGem then
		link = self:RemoveGem(link)
		linkDiff1 = self:RemoveGem(linkDiff1)
		if linkDiff2 then
			linkDiff2 = self:RemoveGem(linkDiff2)
		end
	else
		link = self:BuildGemmedTooltip(link, red, yellow, blue, meta)
		linkDiff1 = self:BuildGemmedTooltip(linkDiff1, red, yellow, blue, meta)
		if linkDiff2 then
			linkDiff2 = self:BuildGemmedTooltip(linkDiff2, red, yellow, blue, meta)
		end
	end

	-- Build ID string
	local id = link..linkDiff1
	if linkDiff2 then
		id = id..linkDiff2
	end

	return id, link, linkDiff1, linkDiff2
end


--[[---------------------------------
  :GetDiff(item, [diff1], [diff2], [ignoreEnchant], [ignoreGem], [red], [yellow], [blue], [meta], [ignorePris])
-------------------------------------
Notes:
  * Calculates the stat diffrence from the specified item and your currently equipped items.
  * item:
  :;tooltip : table - The tooltip showing the item
  :;itemId : number - The numeric ID of the item. ie. 12345
  :;"itemString" : string - The full item ID in string format, e.g. "item:12345:0:0:0:0:0:0:0".
  :::Also supports partial itemStrings, by filling up any missing ":x" value with ":0", e.g. "item:12345:0:0:0"
  :;"itemName" : string - The Name of the Item, ex: "Hearthstone"
  :::The item must have been equiped, in your bags or in your bank once in this session for this to work.
  :;"itemLink" : string - The itemLink, when Shift-Clicking items.
Arguments:
  number or string or table - tooltip or itemId or "itemString" or "itemName" or "itemLink"
  table - Stat difference of item and equipped item 1 are writen to this table if provided
  table - Stat difference of item and equipped item 2 are writen to this table if provided
  boolean - Ignore enchants when calculating stat diffrences
  boolean - Ignore gems when calculating stat diffrences
  number or string - gemID to replace a red socket
  number or string - gemID to replace a yellow socket
  number or string - gemID to replace a blue socket
  number or string - gemID to replace a meta socket
  boolean - Ignore prismatic sockets when calculating the id if true
Returns:
  ; diff1 : table - The table with stat diff values for item 1
  :{
  ::    ["STAT_ID1"] = value,
  ::    ["STAT_ID2"] = value,
  :}
  ; diff2 : table - The table with stat diff values for item 2
  :{
  ::    ["STAT_ID1"] = value,
  ::    ["STAT_ID2"] = value,
  :}
Example:
  StatLogic:GetDiff(21417, {}) -- Ring of Unspoken Names
  StatLogic:GetDiff(21452) -- Staff of the Ruins
-----------------------------------]]

-- TODO 2.1.0: Use SetHyperlinkCompareItem in StatLogic:GetDiff
function StatLogic:GetDiff(item, diff1, diff2, ignoreEnchant, ignoreGem, red, yellow, blue, meta, ignorePris)
    debugPrint("StatLogic:GetDiff");
    --debugPrint("    item="..item)

	-- Locale check
	if noPatternLocale then return end

	-- Get DiffID
	local id, link, linkDiff1, linkDiff2 = self:GetDiffID(item, ignoreEnchant, ignoreGem, red, yellow, blue, meta, ignorePris)
	if not id then return end

	-- Clear Tables
	clearTable(diff1)
	clearTable(diff2)

	-- Get diff data from cache if available
	if cache[id..1] then
		copyTable(diff1, cache[id..1])
		if cache[id..2] then
			copyTable(diff2, cache[id..2])
		end
		return diff1, diff2
	end

	-- Get item sum, results are written into diff1 table
	local itemSum = self:GetSum(link)
	if not itemSum then return end
	local itemType = itemSum.itemType

	if itemType == "INVTYPE_2HWEAPON" and not HasTitanGrip() then
		local equippedSum1, equippedSum2
		-- Get main hand item sum
		if linkDiff1 == "NOITEM" then
			equippedSum1 = newStatTable()
		else
			equippedSum1 = self:GetSum(linkDiff1)
		end
		-- Get off hand item sum
		if linkDiff2 == "NOITEM" then
			equippedSum2 = newStatTable()
		else
			equippedSum2 = self:GetSum(linkDiff2)
		end
		-- Calculate diff
		diff1 = copyTable(diff1, itemSum) - equippedSum1 - equippedSum2
		-- Return table to pool
		del(equippedSum1)
		del(equippedSum2)
	else
		local equippedSum
		-- Get equipped item 1 sum
		if linkDiff1 == "NOITEM" then
			equippedSum = newStatTable()
		else
			equippedSum = self:GetSum(linkDiff1)
		end
		-- Calculate item 1 diff
		diff1 = copyTable(diff1, itemSum) - equippedSum
		-- Clean up
		del(equippedSum)
		equippedSum = nil
		-- Check if item has a second equip slot
		if linkDiff2 then -- If so
			-- Get equipped item 2 sum
			if linkDiff2 == "NOITEM" then
				equippedSum = newStatTable()
			else
				equippedSum = self:GetSum(linkDiff2)
			end
			-- Calculate item 2 diff
			diff2 = copyTable(diff2, itemSum) - equippedSum
			-- Clean up
			del(equippedSum)
		end
	end
	-- Return itemSum table to pool
	del(itemSum)
	-- Write cache
	copyTable(cache[id..1], diff1)
	if diff2 then
		copyTable(cache[id..2], diff2)
	end
	-- return tables
	return diff1, diff2
end


-----------
-- DEBUG --
-----------
-- StatLogic:Bench(1000)
---------
-- self:GetSum(link, table)
-- 1000 times: 0.6 sec without cache
-- 1000 times: 0.012 sec with cache
---------
-- ItemBonusLib:ScanItemLink(link)
-- 1000 times: 1.58 sec
---------
-- LibItemBonus:ScanItem(link, true)
-- 1000 times: 0.72 sec without cache
-- 1000 times: 0.009 sec with cache
---------
--[[
LoadAddOn("LibItemBonus-2.0")
local LibItemBonus = LibStub("LibItemBonus-2.0")
-- #NODOC
function StatLogic:Bench(k)
  DEFAULT_CHAT_FRAME:AddMessage("test")
  local t = GetTime()
  local link = GetInventoryItemLink("player", 12)
  local table = {}
  --local GetItemInfo = _G["GetItemInfo"]
  for i = 1, k do
    ---------------------------------------------------------------------------
    --self:SplitDoJoin("+24 Agility/+4 Stamina, +4 Dodge and +4 Spell Crit/+5 Spirit", {"/", " and ", ","})
    ---------------------------------------------------------------------------
    self:GetSum(link)
    --LibItemBonus:ScanItemLink(link)
    ---------------------------------------------------------------------------
    --ItemRefTooltip:SetScript("OnTooltipSetItem", function(frame, ...) RatingBuster:Print("OnTooltipSetItem") end)
    ---------------------------------------------------------------------------
    --GetItemInfo(link)
  end
  DEFAULT_CHAT_FRAME:AddMessage(GetTime() - t)
  t = GetTime()
  for i = 1, k do
    LibItemBonus:ScanItem(link, true)
  end
  DEFAULT_CHAT_FRAME:AddMessage(GetTime() - t)
  --return GetTime() - t1
end
--]]
--[[
-- #NODOC
function StatLogic:PatternTest()
  patternTable = {
    "(%a[%a ]+%a) ?%d* ?%a* by u?p? ?t?o? ?(%d+) ?a?n?d? ?", -- xxx xxx by 22 (scan first)
    "(%a[%a ]+) %+(%d+) ?a?n?d? ?", -- xxx xxx +22 (scan 2ed)
    "(%d+) ([%a ]+) ?a?n?d? ?", -- 22 xxx xxx (scan last)
  }
  textTable = {
    "Spell Damage +6 and Spell Hit Rating +5",
    "+3 Stamina, +4 Critical Strike Rating",
    "+26 Healing Spells & 2% Reduced Threat",
    "+3 Stamina/+4 Critical Strike Rating",
    "Socket Bonus: 2 mana per 5 sec.",
    "Equip: Increases damage and healing done by magical spells and effects by up to 150.",
    "Equip: Increases the spell critical strike rating of all party members within 30 yards by 28.",
    "Equip: Increases damage and healing done by magical spells and effects of all party members within 30 yards by up to 33.",
    "Equip: Increases healing done by magical spells and effects of all party members within 30 yards by up to 62.",
    "Equip: Increases your spell damage by up to 120 and your healing by up to 300.",
    "Equip: Restores 11 mana per 5 seconds to all party members within 30 yards.",
    "Equip: Increases healing done by spells and effects by up to 300.",
    "Equip: Increases attack power by 420 in Cat, Bear, Dire Bear, and Moonkin forms only.",
  }
  for _, text in ipairs(textTable) do
    DEFAULT_CHAT_FRAME:AddMessage(text.." = ")
    for _, pattern in ipairs(patternTable) do
      local found
      for k, v in gmatch(text, pattern) do
        found = true
        DEFAULT_CHAT_FRAME:AddMessage("  '"..k.."', '"..v.."'")
      end
      if found then
        DEFAULT_CHAT_FRAME:AddMessage("  using: "..pattern)
        DEFAULT_CHAT_FRAME:AddMessage("----------------------------")
        break
      end
    end
  end
end
--]]

-- for debugging
_G.StatLogic = StatLogic

----------------------
-- API doc template --
----------------------
--[[---------------------------------
  :SetTip(item)
-------------------------------------
Notes:
  * This is a debugging tool for localizers
  * Displays item in ItemRefTooltip
  * item:
  :;itemId : number - The numeric ID of the item. ie. 12345
  :;"itemString" : string - The full item ID in string format, e.g. "item:12345:0:0:0:0:0:0:0".
  :::Also supports partial itemStrings, by filling up any missing ":x" value with ":0", e.g. "item:12345:0:0:0"
  :;"itemName" : string - The Name of the Item, ex: "Hearthstone"
  :::The item must have been equiped, in your bags or in your bank once in this session for this to work.
  :;"itemLink" : string - The itemLink, when Shift-Clicking items.
  * Converts ClassID to and from "ClassName"
  * class:
  :{| class="wikitable"
  !ClassID!!"ClassName"
  |-
  |1||"WARRIOR"
  |-
  |2||"PALADIN"
  |-
  |3||"HUNTER"
  |-
  |4||"ROGUE"
  |-
  |5||"PRIEST"
  |-
  |6||"DEATHKNIGHT"
  |-
  |7||"SHAMAN"
  |-
  |8||"MAGE"
  |-
  |9||"WARLOCK"
  |-
  |10||"DRUID"
  |}
Arguments:
  number or string - itemId or "itemString" or "itemName" or "itemLink"
  [optional] string - Armor value. Default: player's armor value
  [optional] number - Attacker level. Default: player's level
  [optional] string or number - ClassID or "ClassName". Default: PlayerClass<br>See :GetClassIdOrName(class) for valid class values.
Returns:
  ; modParry : number - The part that is affected by diminishing returns.
  ; drFreeParry : number - The part that isn't affected by diminishing returns.
Example:
  StatLogic:SetTip("item:3185:0:0:0:0:0:1957")
-----------------------------------]]


------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
if (WoWUnit) then

local function checkEquals(expected, actual, message)
	--print("LibStatLogic:checkEquals");
	--print("1-->expected: "..(expected or "nil"));
	--print("1-->actual: "..(actual or "nil"));
	--print("1-->message: "..(message or "nil"));
	
	return WoWUnit.CheckEquals(expected, actual, message);
end;

local LibStatLogicTests = {

	mocks = {
		UnitName = function(arg)
			return "Soandso";
		end;
	};
	
	setUp = function()
		return {};
	end;
	tearDown = function()
		-- no tear down required
	end;
	
	testExample = function()
		assert(UnitName("player") == "Soandso", "Expected player name to be 'Soandso'");
	end;
	
	--testFailure = function()
	--	assert(UnitName("player") == "Feithar", "Expected player name to be 'Feithar'");
	--end;
	
	testToNumber = function()
		checkEquals(1, tonumber("1"));
		checkEquals(12, tonumber("12"));
		checkEquals(123, tonumber("123"));
		checkEquals(1234, tonumber("1234"));
		checkEquals(12345, tonumber("12345"));
		checkEquals(123456, tonumber("123456"));
		checkEquals(1234567, tonumber("1234567"));
		checkEquals(12345678, tonumber("12345678"));
		checkEquals(123456789, tonumber("123456789"));
		checkEquals(1234567890, tonumber("1234567890"));
		checkEquals(12345678901, tonumber("12345678901"));
		checkEquals(123456789012, tonumber("123456789012"));
	end;

	testCNumber = function()
		local s = "1234";
		local n = CNumber(s);
		checkEquals(1234, n, s);
		
		s = "1,234";
		n = CNumber(s);
		checkEquals(1234, n, s);
		
		s = "1,234,567";
		n = CNumber(s);
		checkEquals(1234567, n, s);
	end;
	
	testCNumberTrillions = function()
		local s = "1,234,567,890";
		local n = CNumber(s);
		checkEquals(1234567890, n, s);
	end;
	
	testLocalizationToNumber = function()
		local s = "1234";
		local n = L.tonumber(s);
		checkEquals(1234, n, s);
	
		s = "1,234";
		n = L.tonumber(s);
		checkEquals(1234, n, s);
		
		s = "1,234,567,890";
		n = L.tonumber(s);
		checkEquals(1234567890, n, s);
		
		s = "1,234.5678";
		n = L.tonumber(s);
		checkEquals(1234.5678, n, s);
	
	end;
	
	testIntegerPattern = function()
		local _, _, value = strfind("1,234 Armor", "^(%d+[,%d]*) Armor$");
		
		checkEquals("1,234", value);
	end;

	testIntegerNumberPattern = function()
		local s = "Something before 1,234,567,890 something after";
		local pattern = "("..(L.patNumber)..")";
		local _, _, value = strfind(s, pattern);
		--print("value = "..(value or "nil"));
		
		checkEquals("1,234,567,890", value, "string from string: "..s..", Pattern: "..pattern);	
		checkEquals(1234567890, CNumber(value), "Number from string: "..s..", Pattern: "..pattern);
	end;
	
	testDecimalNumberPattern = function()
		local s = "Something before 1,234,567,890.123456 something after";
		local _, _, value = strfind(s, "("..L.patDecimal..")");
		checkEquals("1,234,567,890.123456", value);	
		checkEquals(1234567890.123456, CNumber(value));
	end;		
	
	testSinglePlusStatCheck = function()
		-- depending on locale, it may be
		-- +19 Stamina = "^%+(patNumber) (.-)%.?$"
		-- Stamina +19 = "^(.-) %+(patNumber)%.?$"
		-- +19 耐力 = "^%+(patNumber) (.-)%.?$"
		-- Some have a "." at the end of string like:
		-- Enchant Chest - Restore Mana Prime "+6 mana every 5 sec. "
		local text = "+1,234 Stamina";
		
		local _, _, value, statText = strfind(text, L.SinglePlusStatCheck);
		
		checkEquals("+1,234", value, L.SinglePlusStatCheck);
		checkEquals("Stamina", statText, L.SinglePlusStatCheck);
		checkEquals(1234, CNumber(value), L.SinglePlusStatCheck);

		text = "+6 mana every 5 sec."
		local _, _, value, statText = strfind(strutf8lower(text), L.SinglePlusStatCheck);
		
		checkEquals("+6", value, text);
		checkEquals("mana every 5 sec", statText, text);
		checkEquals(6, CNumber(value), text);
	end;
	
	testSingleEquipStatCheck = function()
		-- "^Equip: (.-) by u?p? ?t?o? ?("..patNumber..") ?(.-)%.?$"
		--Equip: Increases your dodge by 172.
		
		local text = "Equip: Increases your dodge by 172.";
	
		local _, _, stat, value = strfind(text, L.SingleEquipStatCheck);
		checkEquals("Increases your dodge", stat);
		checkEquals("172", value);
		checkEquals(172, CNumber(value));
		
		local text = "Equip: Increases your dodge by 1,172.";
		local _, _, stat, value = strfind(text, L.SingleEquipStatCheck);
		checkEquals("Increases your dodge", stat);
		checkEquals("1,172", value);
		checkEquals(1172, CNumber(value));
	end;
	
	testGetAvoidanceAfterDR = function()
			--[[
				avoidanceAfterDR = GetAvoidanceAfterDR(avoidanceType, avoidanceBeforeDR[, class])		
				
				avoidanceType: 	"DODGE", "PARRY", "MELEE_HIT_AVOID"
				avoidanceBeforeDR: let's use 0, since that's what http://www.wowace.com/addons/libstatlogic-1-2/tickets/16-division-by-zero-on-ptr-4-3/ is experiencing
				class: was "6" a guy said
			--]]
			local n;
			
			n = StatLogic:GetAvoidanceAfterDR("DODGE", 0, CLASSID_DEATHKNIGHT);
			n = StatLogic:GetAvoidanceAfterDR("PARRY", 0, CLASSID_DEATHKNIGHT);
			n = StatLogic:GetAvoidanceAfterDR("MELEE_HIT_AVOID", 0, CLASSID_DEATHKNIGHT);
	end;

	testGetAvoidanceAfterDR_DivisionByZerosForAllClasses = function()
			--[[
				avoidanceAfterDR = GetAvoidanceAfterDR(avoidanceType, avoidanceBeforeDR[, class])		
				
				avoidanceType: 	"DODGE", "PARRY", "MELEE_HIT_AVOID"
				avoidanceBeforeDR: let's use 0, since that's what http://www.wowace.com/addons/libstatlogic-1-2/tickets/16-division-by-zero-on-ptr-4-3/ is experiencing
				class: iterate all classes
			--]]
			local n;
			
			local classID;
			for classID = CLASSID_WARRIOR, CLASSID_MAX do
				StatLogic:GetAvoidanceAfterDR("DODGE", 0, CLASSID_DEATHKNIGHT);
				StatLogic:GetAvoidanceAfterDR("PARRY", 0, CLASSID_DEATHKNIGHT);
				StatLogic:GetAvoidanceAfterDR("MELEE_HIT_AVOID", 0, CLASSID_DEATHKNIGHT);
			end;
	end;
	
	testGetEffectFromRating_DivisionByZero = function()
		--[[function StatLogic:GetEffectFromRating(rating, id, level, class)
		--]]

		local rating = 500; --a nice rount, non-zero, number to test with

		StatLogic:GetEffectFromRating(rating, CR_PARRY, 86, CLASSID_DEATHKNIGHT); 
		
		StatLogic:GetEffectFromRating(rating, COMBAT_RATING_RESILIENCE_CRIT_TAKEN, 86, CLASSID_DEATHKNIGHT); 
		
		local nCombatRatingID;
		for nCombatRatingID = 1, CR_MAX do
			StatLogic:GetEffectFromRating(rating, nCombatRatingID, 86, CLASSID_DEATHKNIGHT); 
		end;
	
	end;
	
	
	testGetReductionFromArmor = function()
		--StatLogic:GetReductionFromArmor(armor, attackerLevel)
		local dr; --damageReduction
		dr = StatLogic:GetReductionFromArmor(0, 85);
		checkEquals(dr, 0);
		
		dr = StatLogic:GetReductionFromArmor(1000000, 90);
		checkEquals(dr, 0.75); --a million armor probably enough to now, and forever, hit the 75% cap
	end;
	
	testGetEffectFromRating_TakesCombatRatingIDorString = function()
		local rating = 600;
		local effectByNumber = StatLogic:GetEffectFromRating(rating, CR_PARRY, 85, CLASSID_DEATHKNIGHT); 
		local effectByName = StatLogic:GetEffectFromRating(rating, "PARRY_RATING", 85, CLASSID_DEATHKNIGHT);

		--print(string.format("Effect by Number: %s", effectByNumber));
		--print(string.format("Effect by Name: %s", effectByName));
		
		assert(math.abs(effectByNumber - effectByName) < 0.0001, "GetEffectFromRating should be same whether by CR_PARRY or \"PARRY_RATING\"");
	end;
	
	testProcessWeaponDamage = function()
		local damagePattern = "^%+?"..L.patDecimal.." %- ("..L.patDecimal..") .-Damage$";
		
		local text = "2,397 - 4,454 Damage";

		--local patNumber = "%d+[,%d]*";
		--/dump strfind("2,397 - 4,454 Damage", "^%+?%d+[,%d]* %- (%d+[,%d]*) .-Damage$")
		
		local _, _, maxDamage = strfind(text, damagePattern);
		checkEquals(4454, CNumber(maxDamage));
		
		--Bloodied Arcanite Reaper  ("763.04 - 1,144 Damage")   Decimal points in first number
		text = "763.04 - 1,144 Damage";
		_, _, maxDamage = strfind(text, damagePattern);
		checkEquals(1144, CNumber(maxDamage));
		
		--Balanced Heartseeker  ("221.34 - 411.06 Damage")  Decimal points in both numbers
		text = "221.34 - 411.06 Damage";
		_, _, maxDamage = strfind(text, damagePattern);
		checkEquals(411.06, CNumber(maxDamage));
	end;
	
	testCNumberAndToNumberAreSame = function()
		local text = "1,234,567.891";
		
		local n = CNumber(text);
		checkEquals(1234567.891, n);
		
		n = L.tonumber(text);
		checkEquals(1234567.891, n);
		
		n = tonumber(text);
		checkEquals(1234567.891, n);
	end;
	
	testMatchTheOnChanceProcEquipItems = function()
		--LibStatLogic cannot handle any "chance of" proc items.
		--PreScan sips these with a regex, which i'm test here.
		local pattern = "^Equip: .- (have a chance to) .-";

		local text;
	
		text = "Equip: Your melee attacks have a chance to cause you to summon a Tentacle of the Old Ones to fight by your side for 12 sec.";
		_, _, value = strfind(text, pattern);
		checkEquals("have a chance to", value);

		text = "Equip: Your melee attacks have a chance to trigger a whirlwind attack dealing 8029 to 12044 physical damage to all targets within 10 yards.";
		_, _, value = strfind(text, pattern);
		checkEquals("have a chance to", value);

		text = "Equip: When you heal you have a chance to gain 2904 haste rating for 20 sec.";
		_, _, value = strfind(text, pattern);
		checkEquals("have a chance to", value);

		text = "Equip: Your spells have a chance to grant you 1,962 haste for 10 sec and 392 haste to up to 3 allies within 20 yards.";
		_, _, value = strfind(text, pattern);
		checkEquals("have a chance to", value);
	end;
	
	testConvertCommaToPeriod = function()
		local LOCALE_SDECIMAL = ",";

		local s = "1234,56";
		
		s = string.gsub(s, LOCALE_SDECIMAL, "%.");
		checkEquals("1234.56", s);
	end;
	
	testCNumberDeDE = function()
		local n;
		local text;
		
		local L = {};
		L.LOCALE_STHOUSAND = ".";
		L.LOCALE_SDECIMAL = ",";

		text = "5.50";
		n = CNumberEx(text, nil, L);
		--checkEquals(nil, n, text)
		checkEquals(5.5, n, text); --CNumberEx first call's Lua's tonumber(), which understands decimal points just fine

		text = "5,50";
		n = CNumberEx(text, nil, L);
		checkEquals(5.50, n, text);

		text = "2,345.50";
		n = CNumberEx(text, nil, L);
		checkEquals(2.3455, n, text); -- 2,345.50  ==>  2,34550  ==>  2.3455

		text = "2.345,50";
		n = CNumberEx(text, nil, L);
		checkEquals(2345.50, n, text);
	end;
	
	testGetSpellCritFromInt = function()
		local int = 1000;
		local classId;
		local level; 
		
		for classId = CLASSID_WARRIOR, CLASSID_MAX do
			for level = 1, 100 do		
				StatLogic:GetSpellCritFromInt(int, classId, level);
			end;
		end;
	end;
	
	testGetParryChance = function()
		local parryChance, parryBeforeDR, freeParry = StatLogic:GetParryChance(1000, 1000);
	end;
	
	testGetDodgeChance = function()
		local dodgeChance, dodgeBeforeDR, freeDodge = StatLogic:GetDodgeChance(1000, 1000);
	end;
};	

WoWUnit:AddTestSuite("lsl", LibStatLogicTests);
print("Registered ListStatLogic unit tests. Run the tests using: /wu lsl");
else
	--no WoWUnit
	--print("LibStatLogic: Could not register unit tests");
end; --if (WoWUnit)
